/* $Id: adjobj.cpp,v 1.4.4.1 2005/12/12 12:27:29 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cassert>
#include "deperr.h"
#include "adjobj.h"

namespace DEP
{

// constructor

AdjObject::AdjObject() : _id(), _inuse(false), _index(-1), _num(0),
	_upari(0) {}


// destructor

AdjObject::~AdjObject()
{
	delete[] _upari;
}


// get string of parameter status

string AdjObject::state_str(size_t i) const
{
	ostringstream oss;

	switch (state(i))
		{
		case obs: oss << "obs"; break;
		case eli: oss << "eli"; break;
		case unk: oss << "unk"; break;
		default:
			assert(false);
		}

	return oss.str();
}


// count number of unknowns by state

size_t AdjObject::num(u_state state) const
{
	assert(_upari != 0);

	size_t n = 0;
	for (size_t i = 0; i < _num; i++)
		{
		if (_upari[i].state == state) n++;
		}

	return n;
}


// print unknown parameters

void AdjObject::print_info(ostream& os) const
{
	Form str5; str5.left().width(5);
	Form str8; str8.left().width(8);
	Form str10; str10.width(10);
	Form int5; int5.width(5);
	Form sci9(2); sci9.scientific().width(9);

	for (size_t i = 0; i < _num; i++)
		os << "  " << int5(_index+i) << "   " << str8(uname(i)) << "   "
			<< str10(_id.c_str()) << "   " << str5(state_str(i)) << "   "
			<< sci9(pii(i)) << endl;
	os << endl;
}


// stochastic weighting of the unknowns to the normal matrix

void AdjObject::stoch_weighting(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	for (size_t i = 0, k = _index; i < _num; i++, k++)
		{
		if (state(i) == obs || state(i) == eli)
			{
			double p = pii(i);
			atpa(k,k) += p;
			// observed - approximated
			double l = ival(i) - par(i);
			atpl(k) += p * l;
			ltpl += p * l * l;
			}
		}
}


// update unknown parameters
// return the number of updated unknowns

size_t AdjObject::update_unknowns(const Vector<double>& dx)
{
	assert(_upari != 0);

	size_t n = 0;

	for (size_t i = 0, k = _index; i < _num; i++, k++)
		{
		//assert(_upari[i].par != 0);
        if (_upari[i].state != eli)
			{
			*(_upari[i].par) += dx(k);
			n++;
			}
		}

	return n;
}


// update std. devs. of parameters
// return the number of updated unknowns

size_t AdjObject::update_sdev(double sigma0, const Matrix<double>& qxx)
{
	assert(_upari != 0);

    size_t n = 0;

	for (size_t i = 0, k = _index; i < _num; i++, k++)
		{
		assert(_upari[i].sdev != 0);
        if (_upari[i].state != eli)
			{
			*(_upari[i].sdev) = sqrt(qxx(k,k)) * sigma0;
			n++;
			}
		}

	return n;
}


// get type name

string AdjObject::type_str() const
{
	ostringstream oss;
	oss << "undefined";
	return oss.str();
}


// get key name

string AdjObject::key_str() const
{
	ostringstream oss;
	oss << "undef";
	return oss.str();
}


// get name of unknown parameter

string AdjObject::uname(size_t i) const
{
	assert(i < _num);

	ostringstream oss;
	oss << "x[" << i << "]";
	return oss.str();
}


// print changes of unknown parameters

void AdjObject::print_changes(ostream& os, const Vector<double>& dx) const
{
	os << "ID    :  " << _id << endl;
	os << "Type  :  " << type_str() << endl;
	os << "Index :  " << _index << endl;
	os << endl;

	os << "  Name       State"
		"       Approx.         Change       Adjusted\n\n";

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);
	
	int k = _index;
	for (size_t i = 0; i < _num; i++, k++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(i))
			<< "   " << sci12(par(i)) << "   " << sci12(dx(k))
			<< "   " << sci12(par(i) + dx(k)) << endl;
   	os << endl;
}


// print unknown parameters

void AdjObject::print_unknowns(ostream& os) const 
{
	os << "ID     :  " << _id << endl;
	os << "Type   :  " << type_str() << endl;
	os << "Index  :  " << _index << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	for (size_t i = 0; i < _num; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< sci12(ival(i)) << "  "
			<< sci12(par(i)) << "  "
			<< sci12(change(i)) << "  "
			<< sci12(sdev(i)) << endl;
		}
	os << endl;
}


// print precision of unknowns

void AdjObject::print_precision(ostream& os) const 
{
	os << "ID     :  " << _id << endl;
	os << "Type   :  " << type_str() << endl;
	os << "Index  :  " << _index << endl;
	os << endl;

	os << "  Param.    State     Initial     Adjusted       Change     "
		"Std.dev.   Ratio\n\n";
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci11(3); sci11.scientific().width(11);
	Form fix3(1); fix3.fixed().width(6);

	for (size_t i = 0; i < _num; i++)
		{
		double t_abs = ratio(i); // test value

		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< sci11(ival(i)) << "  "
			<< sci11(par(i)) << "  "
			<< sci11(change(i)) << "  "
			<< sci11(sdev(i)) << "  " << fix3(t_abs);
			
		// test on significance
		if (state(i) == eli)
			os << "  is eliminated";
		else
			{
			if (chk_sign(t_abs))
				os << "  is significant";
			else
				os << "  not significant";
			}
		os << endl;
		}
	os << endl;
}


// check on convergence
/*
int AdjObject::check_convergence(const Vector<double>& dx) const
{
	for (int i = 0, k = _index; i < _num; i++, k++)
		{
		if (fabs(dx(k)) > lspars.conv_gen()) return i;
		}
	return -1;
}
*/


// sum of absolute changes

void AdjObject::sum_changes(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index;
	for (size_t i = 0; i < _num; i++, k++)
		{
		if (state(i) != eli) sum += fabs(dx(k)), num++;
		}
}


// create table

int AdjObject::create(size_t sz)
{
	assert(_upari == 0);
	assert(sz > 0);

	_upari = new Upari[_num = sz];
	if (_upari == 0) return NOMALLOC;

	return OK;
}


// set table entry

void AdjObject::set(size_t i, double* par, double* sdev)
{
	assert(i < _num);

	if (is_obs(*sdev))
		{
		// observed parameter (real std. devs.)
		_upari[i].state = obs;
		_upari[i].pii = lspars._sigma0 / *sdev;
		_upari[i].pii *= _upari[i].pii;
		}
	else if (is_eli(*sdev))
		{
		// eliminated parameter
		_upari[i].state = eli;
		_upari[i].pii = lspars._unksup_wt;
		}
	else
		{
		// unknown parameter
		_upari[i].state = unk;
		_upari[i].pii = 0.;
		}

	_upari[i].ival = *par;	// save initial value
	_upari[i].par = par;	// pointer to parameter
	_upari[i].sdev = sdev;	// pointer to std. dev.
}


// get c-style string of parameter status from std. dev.

string AdjObject::state_str(double s) const
{
	//const int len = 4;
	//static char str[len];

	//ostrstream oss(str, len);
	ostringstream oss;

	if (is_obs(s))
		oss << "obs"; // observed parameter (real std. devs.)
	else if (is_eli(s))
		oss << "eli"; // eliminated parameter
	else if (is_unk(s))
		oss << "unk"; // unknown parameter
	else
		{
		cout << "ERROR: Invalid standard deviation: " << s << endl;
		assert(false);
		}

	//return str;
	return oss.str();
}


// non-member functions ...

// correlations, output in table form

void corr_table(ostream& os, const Matrix<double>& qxx,
const AdjObject& ao1, const AdjObject& ao2, double lim)
{
	Form str8; str8.left().width(8);
	Form int5; int5.width(5);
	Form fix6(3); fix6.fixed().width(6);

	int k = ao1._index;
	for (size_t i = 0; i < ao1._num; i++, k++)
		{
		int l = ao2._index;
		for (size_t j = 0; j < ao2._num; j++, l++)
			{
			if (k == l) continue; // always r = 1!!!
			double r = qxx(k,l) / sqrt(qxx(k,k)) / sqrt(qxx(l,l));
			
			if (fabs(r) > lim) 
				{
				os << "  " << int5(k) << "  " << int5(l)
					<< "    " << str8(ao1.uname(i))
					<< "  " << str8(ao1.get_id())
					<< "  " << str8(ao2.uname(j))
					<< "  " << str8(ao2.get_id())
					<< "  " << fix6(r) << endl;
				}
			}
		}
}


// correlations, output in matrix form

void corr_matrix(ostream& os, const Matrix<double>& qxx,
const AdjObject& ao1, const AdjObject& ao2, double lim)
{
	//os << "Correlation coeffs. between #" << ao1._id
	//	<< " and #" << ao2._id << "\n\n";

	Form str; str.right().width(8);
	Form fix(3); fix.fixed().width(8);

	os << "         ";
	for (size_t j = 0; j < ao2._num; j++)
		os << " " << str(ao2.uname(j));
	os << "\n\n";

	str.left();

	int k = ao1._index;
	for (size_t i = 0; i < ao1._num; i++, k++)
		{
		os << str(ao1.uname(i)) << "  ";
		int l = ao2._index;
		for (size_t j = 0; j < ao2._num; j++, l++)
			{
			double r = qxx(k,l) / sqrt(qxx(k,k)) / sqrt(qxx(l,l));
			os << fix(r);
			if (k != l && fabs(r) > lim)
				os << '*';
			else
				os << ' ';
			}
		os << endl;
		}
	os << endl;
}

} // namespace DEP
