/* $Id: aolist.cpp,v 1.3.4.1 2005/12/12 12:27:29 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "aolist.h"
#include "extori.h"
#include "objpnt.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"

namespace DEP
{

// constructor

AdjObjectList::AdjObjectList(): _list(), _num_unk(0) { }


// setup object list

void AdjObjectList::setup(AdjObject* obj)
{
	assert(obj != 0);

	obj->set_upari();		// setup upar-info

	obj->set_index(_num_unk); // set index to normal equations
	_list.push_back(obj); // add at end of object vector

	_num_unk += obj->num_unk(); // increase number of unknowns
}


// print table

void AdjObjectList::print(ostream& os) const
{
	Form sci8(2); sci8.scientific().width(8);

	os << "Unknown parameters\n";
	os << "==================\n";
	os << "Std. dev. of unit weight             :  "
		<< sci8(AdjObject::lspars.sigma0()) << endl;
	os << "Limit of std. dev. to indicate that\n";
	os << "  unknown should be eliminated       :  "
		<< sci8(AdjObject::lspars.smin()) << endl;
	os << "  value is an unknown                :  "
		<< sci8(AdjObject::lspars.smax()) << endl;
	os << "Weight for suppression of unknowns   :  "
		<< sci8(AdjObject::lspars.unksup_wt()) << endl;
	os << endl;
	os << "State: unk=unknown, obs=observed, eli=eliminated\n";
	os << endl;
	os << "  Index   Name               ID   State      Weight\n";
	os << endl;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		(*obj)->print_info(os);

	os << endl;
}


// count number of constraints

size_t AdjObjectList::num_con() const
{
	size_t n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		n += (*obj)->num_con();

	return n;
}


// count number of unknowns by state

int AdjObjectList::num(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		n += (*obj)->num(state);

	return n;
}

int AdjObjectList::num_obptc() const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ObjectPoint* p = dynamic_cast<ObjectPoint*>(*obj))
			n += p->num_unk();
		}

	return n;
}

int AdjObjectList::num_obptc(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ObjectPoint* p = dynamic_cast<ObjectPoint*>(*obj))
			n += p->num(state);
		}

	return n;
}

int AdjObjectList::num_eop() const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ImageGeometry* p = dynamic_cast<ImageGeometry*>(*obj))
			n += p->num_unk();
		}

	return n;
}

int AdjObjectList::num_eop(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ImageGeometry* p = dynamic_cast<ImageGeometry*>(*obj))
			n += p->num(state);
		}

	return n;
}


int AdjObjectList::num_aps() const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (APSet* p = dynamic_cast<APSet*>(*obj))
			n += p->num_unk();
		}

	return n;
}

int AdjObjectList::num_aps(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (APSet* p = dynamic_cast<APSet*>(*obj))
			n += p->num(state);
		}

	return n;
}


int AdjObjectList::num_pos_cor() const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (PosCorrection* p = dynamic_cast<PosCorrection*>(*obj))
			n += p->num_unk();
		}

	return n;
}

int AdjObjectList::num_pos_cor(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (PosCorrection* p = dynamic_cast<PosCorrection*>(*obj))
			n += p->num(state);
		}

	return n;
}

int AdjObjectList::num_att_cor() const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (AttCorrection* p = dynamic_cast<AttCorrection*>(*obj))
			n += p->num_unk();
		}

	return n;
}

int AdjObjectList::num_att_cor(u_state state) const
{
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (AttCorrection* p = dynamic_cast<AttCorrection*>(*obj))
			n += p->num(state);
		}

	return n;
}


// add weighted constraints

void AdjObjectList::add_constraints(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		(*obj)->add_constraints(atpa, atpl, ltpl);
}


// stochastic weighting of the unknowns to the normal matrix

void AdjObjectList::stoch_weighting(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		(*obj)->stoch_weighting(atpa, atpl, ltpl);
}


// average change of object coords.

double AdjObjectList::average_change_obj(const Vector<double>& dx) const
{
	double sum = 0.;
	int num = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ObjectPoint* op = dynamic_cast<ObjectPoint*>(*obj))
			op->sum_changes(dx, sum, num);
		}

	double av = (num > 0) ? (sum / double(num)) : (-1.);

	return av;
}


// average change of perspective centre coords.

double AdjObjectList::average_change_pcc(const Vector<double>& dx) const
{
	double sum = 0.;
	size_t num = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ExtOrientation* op = dynamic_cast<ExtOrientation*>(*obj))
			op->sum_changes_pcc(dx, sum, num);
		}

	double av = (num > 0) ? (sum / double(num)) : (-1.);

	return av;
}


// average change of rotation parameters

double AdjObjectList::average_change_rot(const Vector<double>& dx) const
{
	double sum = 0.;
	size_t num = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ExtOrientation* op = dynamic_cast<ExtOrientation*>(*obj))
			op->sum_changes_rot(dx, sum, num);
		}

	double av = (num > 0) ? (sum / double(num)) : (-1.);

	return av;
}


// average change of image geometry parameters

double AdjObjectList::average_change_geo(const Vector<double>& dx) const
{
	double sum = 0.;
	size_t num = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		if (ImageGeometry* op = dynamic_cast<ImageGeometry*>(*obj))
			op->sum_changes(dx, sum, num);
		}

	double av = (num > 0) ? (sum / double(num)) : (-1.);

	return av;
}


// update unknowns
// return the number of updated unknowns

int AdjObjectList::update_unknowns(const Vector<double>& dx)
{
	assert(dx.size() == _num_unk);
	int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		n += (*obj)->update_unknowns(dx);

	return n;
}


// update std. devs.
// return the number of updated unknowns

int AdjObjectList::update_sdev(double sigma0, const Matrix<double>& qxx)
{
	//cout << qxx.rows() << endl;
	//cout << _num_unk << endl;
	assert(qxx.rows() == _num_unk);
    int n = 0;

	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		n += (*obj)->update_sdev(sigma0, qxx);

	return n;
}


// check on convergence

bool AdjObjectList::check_convergence(const Vector<double>& dx) const
{
	vector<AdjObject*>::const_iterator obj;
	for (obj = _list.begin(); obj != _list.end(); ++obj)
		{
		int i = (*obj)->check_convergence(dx);
		if (i != -1)
			{
			cout << "Convergence check failed at\n";
			cout << "  ID        : " << (*obj)->get_id() << endl; 
			cout << "  Type      : " << (*obj)->type_str() << endl;
			cout << "  Parameter : " << (*obj)->uname(i) << endl; 
			cout << endl;
			return false;
			}
		}

	return true;
}


// print table of correlation coeffs.

void AdjObjectList::print_correlations(ostream& os, const Matrix<double>& qxx,
double ccoef_lim) const
{
	vector<AdjObject*>::const_iterator ao1;
	for (ao1 = _list.begin(); ao1 != _list.end(); ++ao1)
		{
		vector<AdjObject*>::const_iterator ao2;
		for (ao2 = ao1; ao2 != _list.end(); ++ao2)
			corr_table(os, qxx, **ao1, **ao2, ccoef_lim);
		}
}

} // namespace DEP
