/* $Id: apset.h,v 1.5.4.1 2005/12/12 12:27:29 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_APSET_INCLUDED
#define DEP_APSET_INCLUDED

#include "adjobj.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{

class Params;
class Camera;
class DerivSub;

/** Base of Additional Parameter Set.
 *  Common parameters and functions of all Additional Parameter (AP) sets.
 *  Functions to compute correction, partials derivatives and
 *  absolute vector contribution of parameters modelling distortions
 *  in the image space.
 *  This is the base class for all AP sets! Don't use this directly. 
 *  ATTENTION:  The derivatives are computed with the image coordinates
 *  and original values for c!!!  This is not entirely correct!  One should
 *  use the collineatity equation image coordinates and the adjusted c!
 *
 *  @author Dirk Stallmann
 */
class APSet : public AdjObject
{
public:
	/// allocate memory for parameters and std. dev.
	void allocate(size_t sz);

	/// free memory for parameters and std. dev.
	void free();

	/// new size for parameters and std. dev.
	void newsize(size_t sz);

	/// constructor with size
	APSet(size_t sz);

	/// virtual destructor
	virtual ~APSet();

	/**@name read/write/print AP set */
	//@{
		/// read AP set
		virtual int read(istream& is);

		/// write AP set
		virtual void write(ostream& os) const;

		/// print AP set
		virtual void print(ostream& os) const;
	//@}

	/**@name overridden base class members */
	//@{
		/// get parameter name
		virtual string uname(size_t i) const;

		/// set unknown parameter information (upari)
		virtual void set_upari();

		// update derivatives
		//virtual void update_derivs(const DPoint3& pho_c) = 0;
	//@}

	/// attach AP set to image or camera
	int attach(const vector<Camera*>& cams);

	/// get derivatives
	virtual void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const = 0;

	/// calculate correction
	void calc_corr(const DPoint3& pho_c, DPoint2& corr) const;

	/// calculate correction
	void calc_corr(const string& flgs, const DPoint3& pho_c,
		DPoint2& corr) const;

	/// calculate correction
	DPoint2 calc_corr(const DPoint3& pho_c) const;

	/// update derivatives
	void update_derivs(const DPoint3& pho_c);

	/// update IO parameters
	//virtual void update(double& xp, double& yp, double& dc);
	virtual void update(double& xp, double& yp, double& dc) {
		cerr << "ERROR: calling empty function!\n";
		assert(false);
	}

	/// print influence of APs on image points
	void print_grid(ostream& os, const string& flags,
		double xo, double yo, double dx, double dy, int nx, int ny,
		double c) const;
	
	/// initialize AP grid
	bool init_grid(ostream& os, const Params& pr,
		const vector<Camera*>& cams,
		double& xo, double& yo, double& dx, double& dy, int& nx, int& ny,
		double& c) const;

	/// print influence of APs on image points
	virtual void print_aps_influence(ostream& os, const Params& pr,
		const vector<Camera*>& cams) const;

protected:
	/// corresponding camera (camera that use this set)
	string		_camera_id;

	/// number of APs
	size_t		_size;

	/// AP values
	double*		_ap;

	/// std. devs. of APs
	double*		_sap;

public:
	/// partial derivatives (maybe part of class AdjObject?)
	DerivSub	_derivs;

};


/** Additional Parameter Set implemented in GAP.
 *  @author Dirk Stallmann
 */
class APSetGAP : public APSet
{
public:
	/// constructor
	APSetGAP() : APSet(10) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;

	/// print influence of APs on image points
	void print_aps_influence(ostream& os, const Params& pr,
		const vector<Camera*>& cams) const;
};


/** Additional Parameter Set implemented in Australis.
 *  @author Dirk Stallmann
 */
class APSetAustralis : public APSet
{
public:
	/// constructor
	APSetAustralis() : APSet(10) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;

	/// print influence of APs on image points
	void print_aps_influence(ostream& os, const Params& pr,
		const vector<Camera*>& cams) const;
};


/** Additional Parameter Set using orthogonal polynoms.
 *  Intermediate class for APSetP12 and APSetP44!
 *  @author Dirk Stallmann
 */
class APSetP : public APSet
{
public:
	/// constructor
	APSetP(size_t sz) : APSet(sz), _b(0.) {}

	/**@name read/write/print AP set */
	//@{
		/// read AP set
		int read(istream& is);

		/// write AP set
		void write(ostream& os) const;

		/// print AP set
		void print(ostream& os) const;
	//@}

protected:
	/// grid width
	double		_b;
};


/** Additional Parameter Set using orthogonal polynoms (by Ebner).
 *  @author Dirk Stallmann
 */
class APSetP12 : public APSetP
{
public:
	/// constructor
	APSetP12() : APSetP(12) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};


/** Additional Parameter Set using orthogonal polynoms (by Gruen).
 *  @author Dirk Stallmann
 */
class APSetP44 : public APSetP
{
public:
	/// constructor
	APSetP44() : APSetP(44) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};


/** Additional Parameter Set modelling IO.
 *  AP set modelling the interior orientation:
 *   - principle point coordinates
 *   - change of focal length
 *
 *  @author Dirk Stallmann
 */
class APSetIO : public APSet
{
public:
	/// constructor
	APSetIO() : APSet(3) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;

	/// print influence of APs on image points
	void print_aps_influence(ostream& os, const Params& pr,
		const vector<Camera*>& cams) const;
};


/** Additional Parameter Set modelling radial symmetric distortions
 *  @author Dirk Stallmann
 */
class APSetRadial : public APSet
{
public:
	/// constructor
	APSetRadial() : APSet(3) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};


/** Additional Parameter Set modelling decentering distortions
 *  @author Dirk Stallmann
 */
class APSetDecentering : public APSet
{
public:
	/// constructor
	APSetDecentering() : APSet(2) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};


/** Additional Parameter Set modelling in-plane distortions
 *  @author Dirk Stallmann
 */
class APSetInPlane : public APSet
{
public:
	/// constructor
	APSetInPlane() : APSet(12) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};


/** Additional Parameter Set modelling out-of-plane distortions
 *  @author Dirk Stallmann
 */
class APSetOutOfPlane : public APSet
{
public:
	/// constructor
	APSetOutOfPlane() : APSet(3) {}

	/**@name overridden base class members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;	

		/// update derivatives
		void update_derivs(const DPoint3& pho_c);
	//@}

	/// get partial derivatives
	void get_derivs(const DPoint3& pho_c, DerivSub& derivs) const;
};

} // namespace DEP

#endif
