/* $Id: bunadj.cpp,v 1.14.4.3.2.5 2007/04/29 19:39:54 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <fstream>

#include "timer.h"

#include "osform.h"

#include "deperr.h"
#include "depvec.h"

#include "bunadj.h"

#include "adjintnat.h"

#ifdef GENLIB2_HAVE_LAPACK
	#include "adjintlap.h"
#endif

#include "fiodep.h"
#include "camera.h"
#include "image.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "imggeo.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"

//#include "octransf.h"
//#include "calib.h"
//#include "scaneo.h"
//#include "cgicfg.h"

namespace DEP
{

// erase all images, cameras, etc.

void BundleAdjustment::clear()
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		delete *image;

	_images.clear();

	vector<Camera*>::const_iterator camera;
	for (camera = _cameras.begin(); camera != _cameras.end(); ++camera)
		delete *camera;

	_cameras.clear();

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		delete *op;

	_obj_points.clear();

	vector<ImageGeometry*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		delete *stat;

	_stations.clear();

	vector<APSet*>::const_iterator ap;
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		delete *ap;

	_ap_sets.clear();

	vector<PosCorrection*>::const_iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		delete *pcor;

	_pos_corrs.clear();

	vector<AttCorrection*>::const_iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		delete *acor;

	_att_corrs.clear();
}


// constructor

BundleAdjustment::BundleAdjustment() : _pars(),
	_images(), _cameras(), _obj_points(), _stations(),
	_ap_sets(), _pos_corrs(), _att_corrs(),
//	_oc_transf(0),
	_list(), _atpa(), _atpl(), _dec_vec(), _dx(), _qxx(),
	_ltpl(0.), _vtpv(0.), _sigma0(0.), _f(0), _adj_interface(0) { }


// destructor

BundleAdjustment::~BundleAdjustment()
{
	clear();
	delete _adj_interface;
}


// read configuration file

int BundleAdjustment::read_config(istream& cfg_in)
{
	string fname, keyword;
	int rtn = ERROR;	// return value

	while (cfg_in >> keyword)
		{
		// check keyword
		fio_type type = fio_test(keyword.c_str());
		if (type == UNKNOWN)
			{
			cout << "ERROR: Unknown keyword: '" << keyword << "'\n";
			return ERROR;
			}
		else if (type == STOP_DEP)
			break;

		// open file
		cfg_in >> fname;
		ifstream in(fname.c_str());
		if (!in)
			{
			cout << "ERROR: Cannot open input file: " << fname << endl;
			return ERROR;
			}
		
		// apply specific read-function
		switch (type)
			{
			case PARAMS:
				if ((rtn = _pars.read(in)) != OK)
					cout << "ERROR: Reading params./options file failed.\n";
				break;

			case LS_PARAMS:
				if ((rtn = AdjObject::lspars.read(in)) != OK)
					cout << "ERROR: Reading LS-parameters file failed.\n";
				break;

			case CAMERA: 
				if ((rtn = read_cameras(in, _cameras)) != OK)
					cout << "ERROR: Reading camera file failed.\n";
				break;

			case IMAGE:
				if ((rtn = read_images(in, _cameras, _images)) != OK)
					cout << "ERROR: Reading image file failed.\n";
				break;

			case IMAGE_COORDS:
				if ((rtn = read_image_coords(in, _images)) != OK)
					cout << "ERROR: Reading image coords. file failed.\n";
				break;

			case OBJECT_COORDS:
				if ((rtn = read_object_coords(in, _obj_points)) != OK)
					cout << "ERROR: Reading object coords. file failed.\n";
				break;

			case CONTROL_SUPP:
				if ((rtn = read_ctrl_supp(in, _obj_points)) != OK)
					cout << "ERROR: Reading CS file failed.\n";
				break;

			case IMAGE_GEOMETRY:
			case EXT_ORI:
				if ((rtn = read_ext_orientation(in, _stations)) != OK)
					cout << "ERROR: Reading EO file failed.\n";
				break;

			case APSET:
				if ((rtn = read_ap_sets(in, _ap_sets)) != OK)
					cout << "ERROR: Reading AP set file failed.\n";
				break;

			case POS_COR:
				if ((rtn = read_pos_correction(in, _pos_corrs)) != OK)
					cout << "ERROR: Reading position correction file failed.\n";
				break;

			case ATT_COR:
				if ((rtn = read_att_correction(in, _att_corrs)) != OK)
					cout << "ERROR: Reading attitude correction file failed.\n";
				break;

			//case OC_TRANSFORM:
				//if ((rtn = AdjObject::lspars.read(in)) != OK)
			//	_oc_transf = new OCTransform;
			//	assert(_oc_transf != 0);
			//	if (_oc_transf->read(in) != true)
			//		{
			//		cout << "ERROR: Reading OC transformation file failed.\n";
			//		rtn = ERROR;
			//		}
			//	else
			//		rtn = OK;
			//	break;

			default:
				assert(false);
			}
		in.close();

		if (rtn != OK)
			{
			cout << "ERROR: Reading input file failed: " << fname << endl;
			return ERROR;
			}
		}

	return OK;
}


// print cameras

void BundleAdjustment::print_cameras(ostream& os) const
{
	os << "Cameras\n";
	os << "=======\n\n";

	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		if (*cam != 0) (*cam)->print(os);

	os << endl;
}


// print images

void BundleAdjustment::print_images(ostream& os) const
{
	os << "Images\n";
	os << "======\n\n";

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->print(os, _pars.pr_ic);

	os << endl;
}


// print object points

void BundleAdjustment::print_object_points(ostream& os) const
{
	int num_nu = 0, num_co = 0, num_ch = 0, num_in = 0 ;

	os << "Object point coordinates\n";
	os << "========================\n";
	os << "Object point type:\n";
	os << "in=intersection, co=control point, ch=check point\n\n";

	os << "Point ID       X/m         Y/m         Z/m"
		"      sX/m      sY/m      sZ/m\n\n";
		
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;

		// print single point
		(*op)->print(os);

		// object point type
		ObjectPoint::op_type type = (*op)->get_type();

		switch (type)
			{
			case ObjectPoint::invalid:	num_nu++; break;
			case ObjectPoint::inter:	num_in++; break;
			case ObjectPoint::ctrl:		num_co++; break;
			case ObjectPoint::check:	num_ch++; break;
			default:
				assert(false); break;
			}
		}
	os << endl;

	os << "Number of object points with type\n";
	os << "  intersection  :  " << num_in << endl;
	os << "  control       :  " << num_co << endl;
	os << "  check         :  " << num_ch << endl;

	int num_tot = num_in + num_co + num_ch;
	os << "  total         :  " << num_tot << endl;
	//os << "  inuse         :  " << num_tot - num_un << endl;
	if (num_nu != 0)
		os << "WARNING: " << num_nu << " invalid points found!\n";
	os << endl;
}


// print control support
/*
void BundleAdjustment::print_ctrl_supp(ostream& os) const
{
	os << "Control support\n";
	os << "===============\n\n";

	//ObjectPoint tmp;
	//tmp.print_cs_info(os);
	os << "Point ID   CS\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) (*op)->print_ctrl_supp(os);

	os << endl << endl;
}
*/


// print exterior orientations

void BundleAdjustment::print_ext_orientations(ostream& os) const
{
	os << "Exterior orientations\n";
	os << "=====================\n\n";
	
	vector<ImageGeometry*>::const_iterator p;
	for (p = _stations.begin(); p != _stations.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print AP sets

void BundleAdjustment::print_ap_sets(ostream& os) const
{
	os << "Additional parameter sets\n";
	os << "=========================\n\n";
	
	vector<APSet*>::const_iterator p;
	for (p = _ap_sets.begin(); p != _ap_sets.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print position corrections

void BundleAdjustment::print_pos_corrs(ostream& os) const
{
	os << "Position corrections\n";
	os << "====================\n\n";
	
	vector<PosCorrection*>::const_iterator p;
	for (p = _pos_corrs.begin(); p != _pos_corrs.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print attitude corrections

void BundleAdjustment::print_att_corrs(ostream& os) const
{
	os << "Attitude corrections\n";
	os << "====================\n\n";
	
	vector<AttCorrection*>::const_iterator p;
	for (p = _att_corrs.begin(); p != _att_corrs.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print object coord. transformation
/*
void BundleAdjustment::print_oc_transf(ostream& os) const
{
	if (_oc_transf == 0) return; // escape
		
	os << "Object coordinate transformation\n";
	os << "================================\n";
	os << endl;

	_oc_transf->print(os);
	os << endl;
}
*/

// initialize bundle block adjustment:
// Check data, EOs and APs, allocate memory, etc.
// Attention: Cameras are attached previously!

int BundleAdjustment::initialize(ostream& os)
{
	Timer timer; timer.start();
	os << "*** Start initialize adjustment" << endl << endl;
	
	// check data (still missing)
	//check_image_points(os);

    // attach image objects: EO
    vector<Image*>::const_iterator image;
    for (image = _images.begin(); image != _images.end(); ++image)
		{
		assert(*image != 0);

		if ((*image)->attach(_stations) != true)
			{
			os << "ERROR: Attach EO for image " << (*image)->get_id()
				<< " failed.\n";
			return ERROR;
			}

		// attach cameras previously!!!
		//if ((*image)->attach(_cameras, os) != OK)
		//	{
		//	os << "ERROR: Attach camera for image " << (*image)->get_id()
		//		<< " failed.\n";
		//	return ERROR;
		//	}

		(*image)->inuse(true); // use this image
		}

	// attach AP set
    vector<APSet*>::const_iterator aps;
    for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps == 0) continue;
		//if ((*aps)->attach(_images, _cameras) != OK)
		if ((*aps)->attach(_cameras) != OK)
			{
			os << "ERROR: Attach AP set " << (*aps)->get_id()
				<< " failed.\n";
			return ERROR;
			}
		}

	// attach position corrections
    vector<PosCorrection*>::const_iterator pcor;
    for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		{
		if (*pcor == 0) continue;
		if ((*pcor)->attach(_stations) != OK)
			{
			os << "ERROR: Attach position correction " << (*pcor)->get_id()
				<< " failed.\n";
			return ERROR;
			}
		}

	// attach attitude corrections
    vector<AttCorrection*>::const_iterator acor;
    for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		{
		if (*acor == 0) continue;
		if ((*acor)->attach(_stations) != OK)
			{
			os << "ERROR: Attach attitude correction " << (*acor)->get_id()
				<< " failed.\n";
			return ERROR;
			}
		}

/*
	// attach camera calibration
    vector<Camera*>::iterator cam;
    for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam == 0) continue;
		if ((*cam)->attach(calibs) != OK)
			{
			os << "ERROR: Attach calibration for camera " << (*cam)->get_id()
				<< " failed." << endl;
			return ERROR;
			}
		}
*/

	// check object point approximations
	//if (check_approx((verb) ? &os : 0) != 0)
	if (check_approx((_pars.pr_chk_appr ? &os : 0)) != OK)
		{
		os << "ERROR: Check of object point approximations failed.\n";
		return ERROR;
		}

	// check number of rays per object point
	check_rays_per_point((_pars.pr_chk_rays ? &os : 0));

	// check number of rays per image
	check_rays_per_image((_pars.pr_chk_rays ? &os : 0));

	// DO THIS HERE? remove unused object points, EOs, AP sets etc.
	//clean_up(&os);
		
	// calculate image geometry approximations
	// add flags for printing and debugging class params./options!!!
	if (calc_approx_image_geometry(&os, true) != true)
		os << "ERROR: Calculation of image geometry approximations failed.\n";

	// calculate EO, including position and attitude correction
    vector<ImageGeometry*>::iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0) (*stat)->calc_orientation();
		
	// calculate object point approximations
	if (calc_object_point_approx((_pars.pr_inter ? &os : 0),
		_pars.db_inter) != OK)
		os << "ERROR: Calculation of object point approximations failed.\n";

	// remove unused object points, EOs, AP sets etc.
	clean_up(&os);

	// prepare data
	// time interval
	{
	Form fix4(4); fix4.fixed().width(12);
	os << "Time interval\n";
	os << "-------------\n";
	double t_min = 0., t_max = 0.;
	time_interval(t_min, t_max);
	os << "t_min        = " << fix4(t_min) << " s\n";
	os << "t_max        = " << fix4(t_max) << " s\n";
	os << "duration dt  = " << fix4(t_max-t_min) << " s\n";
	os << endl;

	// set reference time
	set_reference_time(t_min);
	os << "Reference time t0 (= t_min) = " << fix4(t_min) << " s\n";
	os << endl;
	}

	// count number of image points/observations
	int num_ipts = 0;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) num_ipts += (*image)->count_image_points();

	int num_obs = 2 * num_ipts;

	// create unknown parameter information (upari)
	// and build list of adjustment objects
	// ... for object points
    vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) _list.setup(*op);

	// ... for EOs
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)	
		if (*stat != 0) _list.setup(*stat);

	// ... for AP sets
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)	
		if (*aps != 0) _list.setup(*aps);

	// ... for position correction
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)	
		if (*pcor != 0) _list.setup(*pcor);

	// ... for attitude correction
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)	
		if (*acor != 0) _list.setup(*acor);

	// number of unknowns: observed and eliminated
	int nu = _list.num_unk();
	int nu_obs = _list.num(obs);
	int nu_eli = _list.num(eli);

	// number of contraints
	int num_con = _list.num_con();

	// degree of freedom
	int f = num_obs - nu + nu_obs + nu_eli + num_con;

	// print the number of observations, unknowns etc.
	Form int6; int6.width(6);
	
	os << "Data defined for adjustment\n";
	os << "---------------------------\n";
	os << "Images                      :  " << int6(count(_images)) << endl;
	os << "Cameras                     :  " << int6(count(_cameras)) << endl;
	os << "Exterior orientations       :  " << int6(count(_stations)) << endl;
	os << "Additional parameter sets   :  " << int6(count(_ap_sets)) << endl;
	os << "Position corrections        :  " << int6(count(_pos_corrs)) << endl;
	os << "Attitude corrections        :  " << int6(count(_att_corrs)) << endl;
	os << "Object points (total)       :  " << int6(num_obpts()) << endl;
	os << "Control points              :  " << int6(num_copts()) << endl;
	os << "Check points                :  " << int6(num_chpts()) << endl;
	os << endl;
	os << "Number of observations, unknowns and redundancy\n";
	os << "-----------------------------------------------\n";
	os << "Observations:\n";
	os << "   Image point coordinates      :  " << int6(num_ipts) << endl;
	os << "   Total number of observations :  " << int6(num_obs) << endl;
	os << endl;
	os << "Unknowns:                           total        thereof\n";
	os << "                                              obs.     eli.\n";
	// Hierfuer Template-Funktionen schreiben!!!
	os << "   Object point coordinates     :  "
		<< int6(_list.num_obptc()) << "   "
		<< int6(_list.num_obptc(obs)) << "   "
		<< int6(_list.num_obptc(eli)) << endl;
	os << "   Exterior orientations        :  "
		<< int6(_list.num_eop()) << "   "
		<< int6(_list.num_eop(obs)) << "   "
		<< int6(_list.num_eop(eli)) << endl;
	os << "   Additional parameters        :  "
		<< int6(_list.num_aps()) << "   "
		<< int6(_list.num_aps(obs)) << "   "
		<< int6(_list.num_aps(eli)) << endl;
	os << "   Position corrections         :  "
		<< int6(_list.num_pos_cor()) << "   "
		<< int6(_list.num_pos_cor(obs)) << "   "
		<< int6(_list.num_pos_cor(eli)) << endl;
	os << "   Attitude corrections         :  "
		<< int6(_list.num_att_cor()) << "   "
		<< int6(_list.num_att_cor(obs)) << "   "
		<< int6(_list.num_att_cor(eli)) << endl;
	os << "   Total number of unknowns     :  "
		<< int6(nu) << "   "
		<< int6(nu_obs) << "   "
		<< int6(nu_eli) << endl;
	os << endl;
	os << "Constraints                     :  " << int6(num_con) << endl;
	os << endl;
	os << "Redundancy                      :  " << int6(f) << endl;
	os << endl;

	// DO THIS OUTSIDE!!!
	// print unknown parameter infomation
	//upari.print(os);

	// debug upari-table
	if (_pars.db_upari)
		{
		//_list.print(cout, _ls_pars);
		_list.print(cout);
		cout.flush();
		}

	// create the desired adjustment interface
	switch(AdjObject::lspars.adj_interface())
		{
		case LSParams::NATIVE:
			_adj_interface = new AdjInterfaceNative;
			break;

#ifdef GENLIB2_HAVE_LAPACK
		case LSParams::LAPACK:
			_adj_interface = new AdjInterfaceLapack;
			break;
#endif

		default:
			os << "ERROR: Unkown adjustment interface.\n";
			return ERROR;
			
			break;
		}

	// allocate memory for normal equations
	_atpa.newsize(nu, nu);
	_atpl.newsize(nu);
	_dec_vec.newsize(nu);
	_dx.newsize(nu);

	Form fix2(2); fix2.fixed();
	os << "Elapsed cpu time to initialize adjustment: " << fix2(timer.stop())
		<< " s\n\n";	

	return OK;
}


// perform adjustment

int BundleAdjustment::adjustment(ostream& os)
{
	Form int3; int3.width(3);
	Form fix2(2); fix2.fixed();

	Timer timer; timer.start();
	bool cv = false;	// convergence flag
	int nit = 0;		// iteration counter

	//cout << _atpa.rows() << " " << _atpa.cols() << endl;
	os << "*** Start iteration loop" << endl << endl;

	// iteration loop
	while (nit++ < AdjObject::lspars.max_iter() && !cv)
		{
		Timer timer1, timer2; timer1.start();
		int num_obs = 0;	// number of observations

		os << "Iteration number " << int3(nit) << "\n";
		os << "--------------------\n\n";

		// initialize normal equation matrices, etc.
		_atpa.clear();
		_atpl.clear();
		_dx.clear();
		_ltpl = 0.;

		// add contributions to normal equations for all images 
		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)		
			{
			if (*image == 0) continue;

			// add contribution to normal equations
			// for all image point observations
			if ((*image)->normal_eqns(_obj_points, _atpa, _atpl, _ltpl,
				num_obs) != OK)
				{
				os << "ERROR: Setup of normal equations failed.\n";
				return ERROR;
				}
			}

		// add weighted constraints
		_list.add_constraints(_atpa, _atpl, _ltpl);

		// stochastic weighting
		// all unknown parameters are treated as weighted direct observations
		_list.stoch_weighting(_atpa, _atpl, _ltpl);

		// solve normal equations by Cholesky
		//atpa.print_symb(os);
		//os << endl;
		timer2.start();
		
		// 1. factorization using Cholesky decomposition
		//if (chol_dec(_atpa, _dec_vec) != true)
		//	{
		//	//sta.status = ERRNUM;
		//	os << "*** ATTENTION: Solving of normal equations failed.\n";
		//	return ERROR;
		//	}

		// 2. solving by backsubstitution
		//chol_sol(_atpa, _dec_vec, _atpl, _dx);
		
		if (solve() != true)
			{
			os << "*** ERROR: Solving of normal equations failed.\n";
			return ERROR;
			}
			
		/*
		{
		Timer timer;
		int n = _atpa.rows();
		int p = upari.num_obptc();
		int r = n - p;
		cout << "n=" << n << ", p=" << p << ", r=" << r << endl;

		// inversion of N11
		timer.start();
		SubMatrix<double> Nii(_atpa, 0, 0, 3, 3);
		Vector<double> d(3);
		for (int i = 0; i < p; i += 3)
			{
			Nii.newpos(i, i);

			if (chol_dec(Nii, d) != true)
				{
				os << "ERROR: Cholesky decomposition failed: " << i << endl;
				return ERROR;
				}
			chol_inv(Nii, d);
			Nii.make_upper_symmetric();
			}
		
		SubMatrix<double> Q11(_atpa, 0, 0, p, p);
		Q11.make_upper_symmetric();

		os << "Q11: " << timer.lap() << endl;

		// reduced normal equations 
		SubMatrix<double> N21(_atpa, p, 0, r, p);
		SubMatrix<double> N12(_atpa, 0, p, p, r);
		SubMatrix<double> N22(_atpa, p, p, r, r);
		SubVector<double> n1(_atpl, 0, p);
		SubVector<double> n2(_atpl, p, r);

		Matrix<double> T(r, p, 0.);
		//Matrix<double> T = N21 * Q11;
		
		SubMatrix<double> N21ij(N21, 0, 0, 3, 3);
		SubMatrix<double> Q11jj(Q11, 0, 0, 3, 3);
		SubMatrix<double> Tij(T, 0, 0, 3, 3);
		for (i = 0; i < r; i += 3)
			{
			for (int j = 0; j < p; j += 3)
				{
				N21ij.newpos(i, j);
				Q11jj.newpos(j, j);
				Tij.newpos(i, j);
				mult(Tij, N21ij, Q11jj);
				}
			}
		
		//{
		//ofstream os("NpsQp.pbm");
		//NpsQp.print_pbm(os);
		//}

		Matrix<double> S = T * N12;
		Vector<double> s = T * n1;

		Matrix<double> Nr = N22 - S;
		Vector<double> nr = n2 - s;

		{
		ofstream os("Nr.pbm");
		Nr.make_upper_symmetric();
		Nr.print_pbm(os);
		}

		os << "Nr, nr: " << timer.lap() << endl;

		//Matrix<double> d;
		if (chol_dec(Nr, d) != true)
			{
			os << "ERROR: Cholesky decomposition failed.\n";
			return ERROR;
			}
		chol_inv(Nr, d);
		Nr.make_upper_symmetric();

		SubVector<double> dx1(dx, 0, p);
		SubVector<double> dx2(dx, p, r);
		mult(dx2, Nr, nr);

		//chol_sol(Nr, d, nr, dxs);

		Vector<double> u = N12 * dx2;
		Vector<double> v = n1 - u;
		mult(dx1, Q11, v);

		//os << "dx1:" << endl << dx1 << endl;
		//os << "dx2:" << endl << dx2 << endl;
		os << "dx: " << timer.lap() << endl;
		}
		*/
		//os << "dx:" << endl << _dx << endl;
		timer2.stop();

		// vtpv
		_vtpv = _ltpl - dot_prod(_dx, _atpl);
		//os << vtpv << endl;

		// output during iteration loop
		if (_pars.pr_obj_crds)
			{
			os << "Object point coordinates\n";
			os << "------------------------\n\n";
			//os << "  Param.     State"
			//	"       Approx.         Change       Adjusted\n\n";
			vector<ObjectPoint*>::iterator op;
			for (op = _obj_points.begin(); op != _obj_points.end(); ++op)		
				if (*op != 0) (*op)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_ext_ori)
			{
			os << "Exterior orientations\n";
			os << "---------------------\n\n";
			vector<ImageGeometry*>::iterator stat;
			for (stat = _stations.begin(); stat != _stations.end(); ++stat)		
				if (*stat != 0) (*stat)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_ap_sets && !is_empty(_ap_sets))
			{
			os << "Additional parameters\n";
			os << "---------------------\n\n";
			vector<APSet*>::iterator ap;
			for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)		
				if (*ap != 0) (*ap)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_pos_cor && !is_empty(_pos_corrs))
			{
			os << "Position corrections\n";
			os << "--------------------\n\n";
			vector<PosCorrection*>::iterator pcor;
			for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)	
				if (*pcor != 0) (*pcor)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_att_cor && !is_empty(_att_corrs))
			{
			os << "Attitude corrections\n";
			os << "--------------------\n\n";
			vector<AttCorrection*>::iterator acor;
			for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)	
				if (*acor != 0) (*acor)->print_changes(os, _dx);
			os << endl;
			}

		// average change of unknowns
		average_change_unknowns(os);

		// number of unknowns: observed and eliminated
		int nu = _list.num_unk();
		int nu_obs = _list.num(obs);
		int nu_eli = _list.num(eli);

		// number of contraints
		int num_con = _list.num_con();

		// redundancy
		_f = num_obs - nu + nu_obs + nu_eli + num_con;
		if (_f < 0)
			{
			os << "*** ERROR: Negative redundancy.\n";
			return ERROR;
			}

		// sigma of unit weight aposteriori
		// Questions:
		//  (a) What is the sigma node when f = 0?
		//  (b) What is the sigma node when f < 0?
		_sigma0 = (_f > 0) ? sqrt(_vtpv / double(_f)) : (0.);

		Form sci2(2); sci2.scientific();

		os << "Number of observations  :  " << num_obs << endl;
		os << "Number of unknowns      :  " << nu << " (thereof: "
			<< nu_obs << " obs. " << nu_eli << " elim.)\n";
		os << "Number of constraints   :  " << num_con << endl;
		os << "Redundancy              :  " << _f << endl;
		os << "vtpv                    :  " << sci2(_vtpv) << endl;
		os << "Sigma 0 a posteriori    :  ";
		if (_sigma0 == 0.)
			os << "undefined, no redundancy!\n";
		else
			os << fix2(_sigma0 * 1e6) << " um\n";
		os << endl;
		os.flush();

		// update unknown parameters
		update_unknowns(os);

		// check on convergence
		cv = _list.check_convergence(_dx);

		os << "Elapsed cpu time per iteration: " << fix2(timer1.stop())
			<< " s, " "solution only: " << fix2(timer2.time()) << " s\n\n";
		}
		// *** END OF ITERATION LOOP ***

	nit--;

	if (nit >= AdjObject::lspars.max_iter())
		{
		os << "\n*** WARNING: Maximum number of iterations "
			"has been reached.\n";
		}
	else
		{
		os << "\n*** Iteration criterion limit has been reached.\n";
		}

	os << "\nNumber of iterations: " << nit << endl;
	os << "Total cpu time for iterations : " << fix2(timer.stop()) << " s\n\n";

	// DON'T make the normal matrix symmetric! The inversion will fail
	// because the Cholesky factor is stored in the lower triangle!
	//_atpa.make_upper_symmetric();

	// write structure of normal matrix in PBM-format
	if (_pars.fname_normal_matrix_struct.compare("none") != 0)
		{
		ofstream os(_pars.fname_normal_matrix_struct.c_str());
		if (!os)
			cout << "ERROR: Cannot write structure of normal matrix file.\n";
		else
			write_normal_matrix_struct(os);
		}

	// write elements of normal matrix in ASCII-format
	if (_pars.fname_normal_matrix.compare("none") != 0)
		{
		ofstream os(_pars.fname_normal_matrix.c_str());
		if (!os)
			cout << "ERROR: Cannot write normal matrix file.\n";
		else
			write_normal_matrix(os);
		}
				
	// write elements of ride hand side vector in ASCII-format
	if (_pars.fname_rhs_vector.compare("none") != 0)
		{
		ofstream os(_pars.fname_rhs_vector.c_str());
		if (!os)
			cout << "ERROR: Cannot write right side vector file.\n";
		else
			write_rhs_vector(os);
		}
				
	// write elements of solution vector in ASCII-format
	if (_pars.fname_solution_vector.compare("none") != 0)
		{
		ofstream os(_pars.fname_solution_vector.c_str());
		if (!os)
			cout << "ERROR: Cannot write  solution vector file.\n";
		else
			write_solution_vector(os);
		}
	
	// check: A'Pv = 0
	/* This is a test on the adjustment.
	 * The condition A'Pv = 0 can also be tested on A'Pl
	 * since l = -v after iterations.
	 * ATTENTION: For eliminated unknowns and constraints
	 * there might be a large number added!
	 */
	check_atpv(os);

	// Hauptprobe missing!!!!

	// print adjusted unknown parameters
	// and image coord. residuals outside!

	return OK;
}


// print unknown parameters

void BundleAdjustment::print_unknowns(ostream& os) const
{
	// print object points
	if (_obj_points.size() > 0) print_object_points(os);

	// estimated parameters with significance test
	if (_stations.size() > 0)
		{
		os << "Exterior orientations\n";
		os << "=====================\n\n";
		vector<ImageGeometry*>::const_iterator stat;
		for (stat = _stations.begin(); stat != _stations.end(); ++stat)		
			if (*stat != 0) (*stat)->print_unknowns(os);
		}

	if (_ap_sets.size() > 0)
		{
		os << "Additional parameter sets\n";
		os << "=========================\n\n";
		vector<APSet*>::const_iterator ap;
		for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)		
			if (*ap != 0) (*ap)->print_unknowns(os);
		}

	if (_pos_corrs.size() > 0)
		{
		os << "Position corrections\n";
		os << "====================\n\n";
		vector<PosCorrection*>::const_iterator pcor;
		for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)		
			if (*pcor != 0) (*pcor)->print_unknowns(os);
		}

	if (_att_corrs.size() > 0)
		{
		os << "Attitude corrections\n";
		vector<AttCorrection*>::const_iterator acor;
		for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)		
			if (*acor != 0) (*acor)->print_unknowns(os);
		}

	os << endl;
}


// check point analysis
// The practical precision estimates/empirical accuracy are computed
// form all check point coordinate differences.
// The accuracy will be derived ONLY from NEW object points.

void BundleAdjustment::check_point_analysis(ostream& os) const
{
	Form str8; str8.width(8);
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);
	Form fix12(3); fix12.fixed().width(12);

	DPoint3 d, d_max, sum_d, sum_dd;
	int num_x = 0, num_y = 0, num_z = 0;

	os << "Check point analysis\n";
	os << "====================\n";
	os << endl;
	os << "Point ID    X-new/m      Y-new/m      Z-new/m "
		"       DX/m      DY/m      DZ/m\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)	
		{
		if (*op == 0) continue;

		// only for check points!!!
		if (!(*op)->is_check()) continue;

		os << str8((*op)->get_id()) << " "
			<< fix12((*op)->crds.x) << " "
			<< fix12((*op)->crds.y) << " "
			<< fix12((*op)->crds.z) << "  ";

		if ((*op)->is_chx())
			{
			//d.x = upari.get_obs(upari_x) - (*op)->crds.x;
			d.x = (*op)->ival(0) - (*op)->crds.x;
			sum_d.x += d.x;
			sum_dd.x += d.x * d.x;
			num_x++;
			if (fabs(d.x) > d_max.x) d_max.x = fabs(d.x); 
			os << " " << fix9(d.x);
			}
		else
			os << "     ---  ";

		if ((*op)->is_chy())
			{
			//d.y = upari.get_obs(upari_y) - (*op)->crds.y;
			d.y = (*op)->ival(1) - (*op)->crds.y;
			sum_d.y += d.y;
			sum_dd.y += d.y * d.y;
			num_y++;
			if (fabs(d.y) > d_max.y) d_max.y = fabs(d.y); 
			os << " " << fix9(d.y);
			}
		else
			os << "     ---  ";

		if ((*op)->is_chz())
			{
			//d.z = upari.get_obs(upari_z) - (*op)->crds.z;
			d.z = (*op)->ival(2) - (*op)->crds.z;
			sum_d.z += d.z;
			sum_dd.z += d.z * d.z;
			num_z++;
			if (fabs(d.z) > d_max.z) d_max.z = fabs(d.z); 
			os << " " << fix9(d.z);
			}
		else
			os << "     ---  ";

		os << endl;
		}
	os << endl;
	
	double mean_x = num_x == 0 ? 0. : sum_d.x / double(num_x);
	double mean_y = num_y == 0 ? 0. : sum_d.y / double(num_y);
	double mean_z = num_z == 0 ? 0. : sum_d.z / double(num_z);

	double rmse_x = num_x == 0 ? 0. : sqrt(sum_dd.x / double(num_x));
	double rmse_y = num_y == 0 ? 0. : sqrt(sum_dd.y / double(num_y));
	double rmse_z = num_z == 0 ? 0. : sqrt(sum_dd.z / double(num_z));

	os << "Summary of check point analysis\n";
	os << "-------------------------------\n";
	os << "Empirical accuracy in object space.\n\n";
	os << "                                  DX/m      DY/m      DZ/m\n\n";
	os << "Number of differences       :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	os << "Maximum of abs. differences :  "
		<< fix9(d_max.x) << " " << fix9(d_max.y) << " " << fix9(d_max.z)
		<< endl;
	os << "Average of differences      :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << "RMS of differences          :  "
		<< fix9(rmse_x) << " " << fix9(rmse_y) << " " << fix9(rmse_z) << endl;
	os << endl;
}


// control point residuals

void BundleAdjustment::ctrl_point_residuals(ostream& os) const
{
	Form str8; str8.width(8);
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);
	Form fix12(3); fix12.fixed().width(12);

	//ObjectPoint* op = 0;
	DPoint3 v, v_max, sum_v, sum_vv;
	//double sum_vxvx = 0., sum_vyvy = 0., sum_vzvz = 0.;
	//int num_vxvx = 0, num_vyvy = 0, num_vzvz = 0;
	int num_x = 0, num_y = 0, num_z = 0;

	//os << endl;
	os << "Control point residuals\n";
	os << "=======================\n\n";
	os << "Point ID        X/m          Y/m          Z/m"
		"      vX/m      vY/m      vZ/m\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		//assert(*op != 0);
		if(*op == 0 || !(*op)->is_ctrl()) continue;

		os << str8((*op)->get_id()) << " "
			<< fix12((*op)->crds.x) << " "
			<< fix12((*op)->crds.y) << " "
			<< fix12((*op)->crds.z);

		if (AdjObject::lspars.is_obs((*op)->sdev.x))
			{
			v.x = (*op)->crds.x - (*op)->ival(0);
			sum_v.x += v.x;
			sum_vv.x += v.x * v.x;
			num_x++;
			if (fabs(v.x) > v_max.x) v_max.x = fabs(v.x); 
			os << " " << fix9(v.x);
			}
		else
			os << "     ---  ";

		if (AdjObject::lspars.is_obs((*op)->sdev.y))
			{
			v.y = (*op)->crds.y - (*op)->ival(1);
			sum_v.y += v.y;
			sum_vv.y += v.y * v.y;
			num_y++;
			if (fabs(v.y) > v_max.y) v_max.y = fabs(v.y); 
			os << " " << fix9(v.y);
			}
		else
			os << "     ---  ";

		if (AdjObject::lspars.is_obs((*op)->sdev.z))
			{
			v.z = (*op)->crds.z - (*op)->ival(2);
			sum_v.z += v.z;
			sum_vv.z += v.z * v.z;
			num_z++;
			if (fabs(v.z) > v_max.z) v_max.z = fabs(v.z); 
			os << " " << fix9(v.z);
			}
		else
			os << "     ---  ";

		os << endl;
		}
	os << endl;

	double mean_x = num_x == 0 ? 0. : sum_v.x / double(num_x);
	double mean_y = num_y == 0 ? 0. : sum_v.y / double(num_y);
	double mean_z = num_z == 0 ? 0. : sum_v.z / double(num_z);

	double rmse_x = num_x == 0 ? 0. : sqrt(sum_vv.x / double(num_x));
	double rmse_y = num_y == 0 ? 0. : sqrt(sum_vv.y / double(num_y));
	double rmse_z = num_z == 0 ? 0. : sqrt(sum_vv.z / double(num_z));
	
	os << "Std. dev. from control point residuals\n";
	os << "--------------------------------------\n";
	os << "                                 vX/m      vY/m      vZ/m\n\n";
	os << "Number of residuals        :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	os << "Maximum of abs. residuals  :  "
		<< fix9(v_max.x) << " " << fix9(v_max.y) << " " << fix9(v_max.z)
		<< endl;
	os << "Average of residuals       :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << "RMS of residuals           :  "
		<< fix9(rmse_x) << " " << fix9(rmse_y) << " " << fix9(rmse_z) << endl;
	os << endl;
}


// calculate precision

int BundleAdjustment::precision(ostream* os)
{
	if (os != 0)
		{
		(*os) << "*** Matrix inversion ... ";
		(*os).flush();
		}

	Form fix2(2); fix2.fixed();
	Timer timer; timer.start();

	if (inversion() != true)
		{
		cout << "*** ERROR: Inversion of normal matrix failed.\n";
		return ERROR;
		}

	if (os != 0)
		{
		(*os) << "Done\n\n";
		(*os) << "Elapsed cpu time for inversion: " << fix2(timer.stop())
			<< " s\n\n";	
		}

	// update std. devs.
	//cout << _list.update_sdev(_sigma0, _qxx) << endl;
	_list.update_sdev(_sigma0, _qxx);

	return OK;
}


// precision for estimated parameters (partially with significance test)

void BundleAdjustment::precision(ostream& os) const
{
	os << "*** Precision from Qxx and Sigma 0 a posteriori ***" << endl;
	os << endl;

	// print object points
	if (_obj_points.size() > 0) print_object_points(os);

	// average precision of object coords.
	average_precision(os);
	os << endl;

	// print EOs
	if (_stations.size() > 0) print_ext_orientations(os);

	// precision for estimated parameters with significance test
	if (_ap_sets.size() > 0)
		{
		os << "Additional parameter sets\n";
		os << "=========================\n\n";
		//os << "  Param.     State        Value       Std.dev.    Ratio\n\n";
		vector<APSet*>::const_iterator ap;
		for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)		
			if (*ap != 0) (*ap)->print_precision(os);
		}

	if (_pos_corrs.size() > 0)
		{
		os << "Position corrections\n";
		os << "====================\n\n";
		vector<PosCorrection*>::const_iterator pcor;
		for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)		
			if (*pcor != 0) (*pcor)->print_precision(os);
		}

	if (_att_corrs.size() > 0)
		{
		os << "Attitude corrections\n";
		os << "====================\n\n";
		vector<AttCorrection*>::const_iterator acor;
		for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)		
			if (*acor != 0) (*acor)->print_precision(os);
		}

	os << endl;
}


// correlations
// print all correlations larger than the output limit

void BundleAdjustment::correlations(ostream& os) const
{
	Form fix(2); fix.fixed();
	double ccoef_lim = AdjObject::lspars.ccoef_lim();

	os << "Correlations\n";
	os << "============\n";
	os << "Output limit of correlation coeffs.:  " << fix(ccoef_lim) << "\n\n";
	os << "      i      j    Name      ID        Name      ID          rij\n\n";
	_list.print_correlations(os, _qxx, ccoef_lim);
	os << endl;
}


// correlation matrices between specific adjustment objects

void BundleAdjustment::correlations_spec(ostream& os) const
{
	Form fix(2); fix.fixed();
	double ccoef_lim = AdjObject::lspars.ccoef_lim();

	os << "Correlations\n";
	os << "============\n";
	os << "Correlation coeffs. larger " << fix(ccoef_lim)
		<< " marked with *\n\n";

	// correlation coefficient sub-matrices
	// ... of EO parameters
	vector<ImageGeometry*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		{
		if (*stat == 0) continue;
					
		os << "Station: " << (*stat)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **stat, **stat, ccoef_lim);
		}

	// ... of APs
	vector<APSet*>::const_iterator ap;
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		{
		if (*ap == 0) continue;
					
		os << "AP set: " << (*ap)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **ap, **ap, ccoef_lim);
		}

	// ... of position corrections
	vector<PosCorrection*>::const_iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		{
		if (*pcor == 0) continue;
				
		os << "Position corr.: " << (*pcor)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **pcor, **pcor, ccoef_lim);
		}

	// ... of attitude corrections
	vector<AttCorrection*>::const_iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		{
		if (*acor == 0) continue;
				
		os << "Attitude corr.: " << (*acor)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **acor, **acor, ccoef_lim);
		}

	// between APs and EOs
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		{
		if (*stat == 0 || *ap == 0) continue;
			
		os << "Between AP set " << (*ap)->get_id()
			 << " and station  " << (*stat)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **ap, **stat, ccoef_lim);
		}

	// between station and position corr.
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		for (stat = _stations.begin(); stat != _stations.end(); ++stat)
			{
			if (*stat == 0 || *pcor == 0) continue;
			
			os << "Between position corr. " << (*pcor)->get_id()
				 << " and station  " << (*stat)->get_id() << "\n\n";
			corr_matrix(os, _qxx, **pcor, **stat, ccoef_lim);
			}

	// between station and attitude corr.
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		for (stat = _stations.begin(); stat != _stations.end(); ++stat)
			{
			if (*stat == 0 || *acor == 0) continue;
			
			os << "Between attitude corr. " << (*acor)->get_id()
				 << " and station  " << (*stat)->get_id() << "\n\n";
			corr_matrix(os, _qxx, **acor, **stat, ccoef_lim);
			}

	// between APs and position corr.
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		{
		if (*ap == 0 || *pcor == 0) continue;
			
		os << "Between AP set " << (*ap)->get_id()
			 << " and position corr.  " << (*pcor)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **ap, **pcor, ccoef_lim);
		}

	// between APs and attitude corr.
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		{
		if (*ap == 0 || *acor == 0) continue;
			
		os << "Between AP set " << (*ap)->get_id()
			 << " and attitude corr.  " << (*acor)->get_id() << "\n\n";
		corr_matrix(os, _qxx, **ap, **acor, ccoef_lim);
		}

	// between position and attitude corr.
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
			{
			if (*pcor == 0 || *acor == 0) continue;
			
			os << "Between position corr. " << (*pcor)->get_id()
				 << " and attitude corr.  " << (*acor)->get_id() << "\n\n";
			corr_matrix(os, _qxx, **pcor, **acor, ccoef_lim);
			}
}


// print influence of APs on image points

void BundleAdjustment::aps_influence(ostream& os) const
{
	os << "Influence of APs on image points\n";
	os << "================================\n";
	os << "Without the influence of principle point changes.\n";
	os << "Number of grid points  :  nx = " << _pars.pr_aps_nx
		<< ",  ny = " << _pars.pr_aps_ny << endl;
	os << "(x,y)-corrections in um\n";
	os << endl;

	if (is_empty(_ap_sets))
		{
		os << "*** ATTENTION: List of AP sets is empty.\n";
		}
	else
		{
		vector<APSet*>::const_iterator aps;
		for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
			{
			if (*aps == 0) continue;
			(*aps)->print_aps_influence(os, _pars, _cameras);
			}
		}
	os << endl;
}


// remove image points of unused object points

void BundleAdjustment::clean_image_points(ostream *os)
{
	int num = 0;

	if (os != 0)
		{
		(*os) << "Clean image points\n";
		(*os) << "==================\n";
		(*os) << "(Remove image coords. of unused object points.)\n\n";
		}

	// loop over all object points
	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if(*op == 0 || !(*op)->is_unused()) continue;

		num += remove_image_point((*op)->get_id(), os);
		}

	if (os != 0)
		{
		(*os) << endl;
		(*os) << "Number of removed image points: " << num << endl;
		(*os) << endl;
		}
}


// check approximations of object coords.
// Loop over all image points and create new object points
// when corresponding point don't exist in the object point list.

int BundleAdjustment::check_approx(ostream* os)
{
	int num = 0;

	if (os != 0)
		{
		(*os) << "Check object point approximations\n";
		(*os) << "=================================\n\n";
		}

	// loop over all images
	vector<Image*>::const_iterator img;
	for (img = _images.begin(); img != _images.end(); ++img)
		{
			
		// loop over all image points
		vector<ImagePoint*>::iterator ip;
		for (ip = (*img)->_image_points.begin();
			ip != (*img)->_image_points.end(); ++ip)
			{
			if (*ip == 0) continue;
			string id = (*ip)->get_id();

			// search object points
			ObjectPoint* op = find(_obj_points, id);
			
			if (op == 0)
				{
				if (os != 0)
					(*os) << "Point " << id << " has no approximations.\n";

				// create new object point.
				// default: intersection point without approximations.
    			ObjectPoint* op = new ObjectPoint;
				assert(op != 0);
				op->set_id(id);
				//op->sdev = op->smaxu();
				//op->sdev = AdjObject::lspars.smaxu();
				//op->type(ObjectPoint);
				_obj_points.push_back(op);

				num++;
				}
			}
        }

	if (os != 0)
		{
		(*os) << endl;
		(*os) << "Number of additional object points: " << num << endl;
		(*os) << endl;
		}
	
	return OK;
}


// check number of rays per object point

void BundleAdjustment::check_rays_per_point(ostream* os)
{
	int removed = 0;

	if (os != 0)
		{
		(*os) << "Check number of rays per object point\n";
		(*os) << "=====================================\n";
		(*os) << "(Count number of rays per object point)\n";
		(*os) << "Number of points to check : " << _obj_points.size() << endl;
		(*os) << "Single-ray option         : "
			<< (_pars.sray_opt ? "on" : "off") << "\n\n";
		(*os) << "<point ID>: <image ID> ... [<number of rays>]\n\n";
		}

	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		// for all object points
		if(*op == 0) continue;

		int rays = 0; // count number of rays
		string id = (*op)->get_id();

		if (os) (*os) << id << ":";

		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)
			{
			assert(*image != 0);

			if ((*image)->image_point_exist(id))
				{
				rays++;
				if (os) (*os) << " " << (*image)->get_id();
				}
			}

		if (os != 0) (*os) << " [" << rays << "]";
		(*op)->num_rays(rays); // save number of rays

		// check the number of rays
		// as a course measure of reliability
		if (rays < 2)
			{
			// single ray option for full control points only!!!
			//if (pr.sray_opt && rays == 1 && (*op)->is_ctrl()
			//	&& (*op)->is_full())
			//if (_pars.sray_opt && rays == 1 && (*op)->is_csx(_ls_pars)
			//	&& (*op)->is_csy(_ls_pars) && (*op)->is_csz(_ls_pars))
			if (_pars.sray_opt && rays == 1 && (*op)->is_csx()
				&& (*op)->is_csy() && (*op)->is_csz())
				{
				(*op)->inuse(true);
				if (os) (*os) << " -> single ray";
				}
			else
				{
				// delete it really
				delete *op;
				*op = 0;
				removed++;
				
				if (os) (*os) << " -> removed";

				// remove image point
				//num += remove_ip(id, os);
				vector<Image*>::const_iterator image;
				for (image = _images.begin(); image != _images.end(); ++image)
					{
					if (*image == 0) continue;
					string img_id = (*image)->get_id(); // save image ID

					if ((*image)->remove_image_point(id))
						{
						if (os)
							(*os) << "\nATTENTION: Image " << img_id
								<< " point " << id << " removed.\n";
						}
					}
				}
			}
		else
			{
			(*op)->inuse(true);
			//(*op)->num_rays(rays); // save number of rays
			}
		if (os) (*os) << endl;
		}

	if (os)
		(*os) << "Number of removed object points: " << removed << "\n\n";
}


// check number of rays per image

void BundleAdjustment::check_rays_per_image(ostream* os)
{
	int removed = 0;

	if (os != 0)
		{
		(*os) << "Check number of rays per image\n";
		(*os) << "==============================\n";
		(*os) << "(Count number of rays per image)\n";
		(*os) << "Number of images to check : " << _images.size() << endl;
		(*os) << "<image ID>: <number of rays>\n\n";
		}

	vector<Image*>::iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if(*image == 0) continue;

		int rays = (*image)->count_image_points();

		if (os) (*os) << (*image)->get_id() << ": " << rays;

		if (rays < 3)
			{
			delete *image;
			*image = 0;
			removed++;
			if (os) (*os) << " -> removed";
			}
		if (os) (*os) << endl;
		}

	if (os)
		(*os) << "Number of removed images: " << removed << "\n\n";
}


// calculate approximate image geometry

bool BundleAdjustment::calc_approx_image_geometry(ostream* os, bool dbg)
{
	//if (os != 0)
	//	{
	//	(*os) << "Calculate image geometry approximations\n";
	//	(*os) << "=======================================\n";
		//(*os) << "(By spatial resection)\n";
	//	(*os) << endl;
	//	}

	//vector<Image*>::const_iterator image;
	//for (image = _images.begin(); image != _images.end(); ++image)
	//	{
	//	if (*image == 0) continue;

	//	if ((*image)->calc_approx_image_geometry(_obj_points, os) != true)
	//		return false;
	//	}
	//if (os != 0) (*os) << endl;

	return true;
}


// calculate object point approximations by spatial intersection

int BundleAdjustment::calc_object_point_approx(ostream* os, bool dbg)
{
	if (os != 0)
		{
		(*os) << "Calculate object point approximations\n";
		(*os) << "=====================================\n";
		(*os) << "(By spatial intersection)\n";
		(*os) << endl;
		(*os) << "                             X/m            Y/m "
			"           Z/m         e/m\n";
		(*os) << endl;
		}

	Form	str12; str12.width(12);
	Form	fix9(3); fix9.fixed().width(9);
	Form	fix12(3); fix12.fixed().width(12);
	
	int		num_pts = 0;		// count number of intersection points
	double	tot_sum_e = 0.;		// total sum of e square

	// loop over all object points
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if(*op == 0 || (*op)->has_approx()) continue;

		// calculate approximations for initial points
		vector<Image*> set;		// images with this point
		
		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)
			{
			//assert(*image != 0);
			if (*image == 0) continue;
			if ((*image)->image_point_exist((*op)->get_id()))
				set.push_back(*image);
			}

		if (os != 0)
			{
			(*os) << "Point: " << (*op)->get_id() << " by ";
			for (size_t i = 0; i < set.size(); i++)
				(*os) << (set[i])->get_id() << " ";
			(*os) << "[" << set.size() << "]" << endl;
			}

		if (set.size() < 2)
			{
			cout << "ATTENTION: Too few rays to compute object coordinates."
				<< endl;
			//(*op)->type(ObjectPoint::unused); ???
			continue;
			}

		// loop over all combination of images
		DPoint3		sum;
		double		sum_e = 0.;
		int			n = 0;

		for (size_t a = 0; a < set.size(); a++)
			for (size_t b = a + 1; b < set.size(); b++)
				{
				Image* image_a = set[a];
				Image* image_b = set[b];

				// find image point coords.
				ImagePoint *ip_a = find(image_a->_image_points,
					(*op)->get_id());
				assert(ip_a != 0);

				ImagePoint *ip_b = find(image_b->_image_points,
					(*op)->get_id());
				assert(ip_b != 0);

				// intersection A-B
				DPoint3 obj_c;
				double e = 0.; // residual parallax

				//if (image_a.intersection(ip_a->crds,
				//	*image_b, ip_b->crds, obj_c, e) != true)
				if (intersection(*image_a, ip_a->crds,
					*image_b, ip_b->crds, obj_c, e) != true)
					{
					cout << "ERROR: Intersection failed." << endl;
					return ERROR;
					}

				sum += obj_c;
				sum_e += e;
				n++;

				if (os != 0 && dbg == true)
					{
					(*os) << "  " 
						<< str12(image_a->get_id()) << " - " 
						<< str12(image_b->get_id()) << endl;
					(*os) << "                      "
						<< fix12(obj_c.x) << "   " 
						<< fix12(obj_c.y) << "   "
						<< fix12(obj_c.z) << "   "
						<< fix9(e) << endl;
					}
				}

			assert(n > 0);
			DPoint3 mean = sum / double(n);
			double mean_e = sum_e / double(n);
			if (os != 0)
				(*os) << "            average : "
					<< fix12(mean.x) << "   "
					<< fix12(mean.y) << "   "
					<< fix12(mean.z) << "   "
					<< fix9(mean_e) << "\n\n";

			(*op)->crds = mean;
			(*op)->approx(true);
			//(*op)->sdev = (*op)->smaxu();
			tot_sum_e += mean_e;
			num_pts++;
		}

	if (os != 0)
		{
		(*os) << "Number of intersection points : " << num_pts << endl;
		(*os) << "Total average distance of rays: e_mean = ";
		if (num_pts > 0)
			(*os) << fix9(tot_sum_e / double(num_pts));
		else
			(*os) << "   ---   ";
		(*os) << " m\n\n";
		}
	
	return OK;
}


// remove unused object points, EOs, Apsets, etc.
/*
			os << "ATTENTION: Image " << img->get_id()
				<< " without image points: image removed.\n";
*/

void BundleAdjustment::clean_up(ostream* os)
{
	if (os)
		{
		(*os) << "Clean up\n";
		(*os) << "--------\n";
		(*os) << "(Remove unused object points, EOs, Apsets, etc.)\n";
		(*os) << endl;
		}

	// remove images
	int num_imgs = 0;
	vector<Image*>::iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0 && (*image)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Image " << (*image)->get_id()
					<< " removed." << endl;
			delete *image;
			*image = 0;
			num_imgs++;
			}
		}

	// remove unused cameras
	int num_cams = 0;
	vector<Camera*>::iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam != 0 && (*cam)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Camera " << (*cam)->get_id()
					<< " removed." << endl;
			delete *cam;
			*cam = 0;
			num_cams++;
			}
		}

	// remove unused object points
	int num_opts = 0;
	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		
		if ((*op)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Object point " << (*op)->get_id()
					<< " removed." << endl;
			delete *op;
			*op = 0;
			num_opts++;
			}
		//else if ((*op)->is_approx())
		//	{
			// change type from approx to inter
		//	(*op)->type(ObjectPoint::inter);
		//	}
		}

	// remove unused EOs
	int num_eos = 0;
	vector<ImageGeometry*>::iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		{
		if (*stat != 0 && (*stat)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Station " << (*stat)->get_id()
					<< " removed." << endl;
			delete *stat;
			*stat = 0;
			num_eos++;
			}
		}

	// remove unused AP sets
	int num_aps = 0;
	vector<APSet*>::iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps != 0 && (*aps)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: AP set " << (*aps)->get_id()
					<< " removed." << endl;
			delete *aps;
			*aps = 0;
			num_aps++;
			}
		}

	// remove unused position corrections
	int num_pcor = 0;
	vector<PosCorrection*>::iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		{
		if (*pcor != 0 && (*pcor)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: position correction " << (*pcor)->get_id()
					<< " removed." << endl;
			delete *pcor;
			*pcor = 0;
			num_pcor++;
			}
		}

	// remove unused attitude corrections
	int num_acor = 0;
	vector<AttCorrection*>::iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		{
		if (*acor != 0 && (*acor)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: attitude correction " << (*acor)->get_id()
					<< " removed." << endl;
			delete *acor;
			*acor = 0;
			num_acor++;
			}
		}

/*
	// remove unused scanner EOs
	int num_seos = 0;
	vector<ScanEO*>::iterator seo;
	for (seo = scan_eos.begin(); seo != scan_eos.end(); ++seo)
		{
		if (*seo != 0 && (*seo)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Scanner EO #" << (*seo)->get_id()
					<< " removed." << endl;
			delete *seo;
			*seo = 0;
			num_seos++;
			}
		}

	// remove unused INS corr.
	int num_icor = 0;
	vector<INSCor*>::iterator icor;
	for (icor = ins_cors.begin(); icor != ins_cors.end(); ++icor)
		{
		if (*icor != 0 && (*icor)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: INS correction #" << (*icor)->get_id()
					<< " removed." << endl;
			delete *icor;
			*icor = 0;
			num_icor++;
			}
		}

	// remove unused CGI config.
	int num_cgic = 0;
	vector<CGICfg*>::iterator cgic;
	for (cgic = cgi_cfgs.begin(); cgic != cgi_cfgs.end(); ++cgic)
		{
		if (*cgic != 0 && (*cgic)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: CGI configuration #" << (*cgic)->get_id()
					<< " removed." << endl;
			delete *cgic;
			*cgic = 0;
			num_cgic++;
			}
		}
*/

	if (os)
		{
		(*os) << endl;
		(*os) << "Number of removed ..." << endl;
		(*os) << "  - images            : " << num_imgs << endl;
		(*os) << "  - cameras           : " << num_cams << endl;
		(*os) << "  - object points     : " << num_opts << endl;
		(*os) << "  - EOs               : " << num_eos << endl;
		(*os) << "  - AP sets           : " << num_aps << endl;
		(*os) << "  - Position corr.    : " << num_pcor << endl;
		(*os) << "  - Attitude corr.    : " << num_acor << endl;
		//(*os) << "  - Scanner EOs       : " << num_seos << endl;
		//(*os) << "  - INS correction    : " << num_icor << endl;
		//(*os) << "  - CGI configuration : " << num_cgic << endl;
		(*os) << endl;
		}
}


/* v-check is a test on the adjustment.
 * The condition A'Pv = 0 can also be tested on A'Pl
 * since l = -v after iterations.
 */

void BundleAdjustment::check_atpv(ostream& os) const
{
	Form sci3(1); sci3.scientific().width(6);

	os << "Check adjustment on A'Pv = 0" << endl;
	os << "----------------------------" << endl;
	os << "A'Pv limit:  " << sci3(AdjObject::lspars.atpv_lim()) << endl;

	int num_err = 0;
	for (int i = 0; i < _atpl.size(); i++)
		{
		if (AdjObject::lspars.chk_atpv(_atpl(i)))
			{
			os << "A'Pv[" << i << "] is large: " << sci3(_atpl(i)) << endl;
			num_err++;
			}
		}
	os << endl;

	os << "Number of A'Pv-values > limit : " << num_err << endl;
	os << endl;
}


// average change of unknowns

void BundleAdjustment::average_change_unknowns(ostream& os) const
{
	Form sci3(3); sci3.scientific().width(10);

	os << "Average change of unknown parameters\n";
	os << " - object point coords.             :  ";

	double av = 0.;
	av = _list.average_change_obj(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << " m\n";

	os << " - perspective centre coords.       :  ";
	av = _list.average_change_pcc(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << " m\n";

	os << " - rotation parameters              :  ";
	av = _list.average_change_rot(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << "\n";

	os << " - geometry parameters              :  ";
	av = _list.average_change_geo(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
	os << sci3(av) << "\n";

	os << endl;
}


// update unknowns and EO of the images

void BundleAdjustment::update_unknowns(ostream& os)
{
	// update all unknowns with the solution vector
	//cout << "Number of updated unknown parameters: "
	//	<< _list.update_unknowns(_dx) << endl;
	_list.update_unknowns(_dx);

	// calculate EO, including position and attitude correction
    vector<ImageGeometry*>::iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0) (*stat)->calc_orientation();
}


// average precision of object coords. derived from all
// standard deviations of coordinates of non-control points.

void BundleAdjustment::average_precision(ostream& os) const
{
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);

	DPoint3 sum_var, sd_max;
	int num_x = 0, num_y = 0, num_z = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		/*
		// check index
		if ((k = (*op)->index()) < 0)
			{
			cout << "ATTENTION: average_precision: Object point #" <<
				(*op)->get_id() << " has invalid index.\n";
			continue;
			}
		*/

		//if ((*op)->is_ctrl()) continue; // only for non-control points!!!
		
		if (!AdjObject::lspars.is_eli((*op)->sdev.x))
			{
			sum_var.x += (*op)->sdev.x * (*op)->sdev.x;
			num_x++;
			if ((*op)->sdev.x > sd_max.x) sd_max.x = (*op)->sdev.x;
			}

		if (!AdjObject::lspars.is_eli((*op)->sdev.y))
			{
			sum_var.y += (*op)->sdev.y * (*op)->sdev.y;
			num_y++;
			if ((*op)->sdev.y > sd_max.y) sd_max.y = (*op)->sdev.y;
			}

		if (!AdjObject::lspars.is_eli((*op)->sdev.z))
			{
			sum_var.z += (*op)->sdev.z * (*op)->sdev.z;
			num_z++;
			if ((*op)->sdev.z > sd_max.z) sd_max.z = (*op)->sdev.z;
			}
		}
	os << endl;

	double mean_x = num_x == 0 ? 0. : sqrt(sum_var.x / double(num_x));
	double mean_y = num_y == 0 ? 0. : sqrt(sum_var.y / double(num_y));
	double mean_z = num_z == 0 ? 0. : sqrt(sum_var.z / double(num_z));
	
	os << "Average precision of object coords.\n";
	os << "===================================\n";
	os << "Derived from std. dev. of coordinates of non-control points.\n\n";
	os << "                                   sX/m      sY/m      sZ/m\n\n";
	os << "Number of std. devs.         :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	//os << "Number of non-control points :  " << num << endl;
	os << "Maximum of std. devs.        :  "
		<< fix9(sd_max.x) << " " << fix9(sd_max.y) << " " << fix9(sd_max.z)
		<< endl;
	os << "Average of std. devs.        :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << endl;
}


// write residual image coords.

void BundleAdjustment::write_residuals(ostream& os) const
{
	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0) (*image)->write_residuals(_obj_points, os);
		}
	os << "stop-dep\n";
}


// print residual image coords.

void BundleAdjustment::print_residuals(ostream& os) const
{
	Form fix2(2); fix2.fixed();

    DPoint tot_v, tot_vv, tot_pvv;
    int tot_num = 0;
	//double vtpv = 0.;

	//os << endl;
	os << "Residuals of image coords.\n";
	os << "==========================\n";
	os << "Limit for printing residuals:  "
		<< fix2(AdjObject::lspars.res_lim() * 1e6) << " um\n";
	os << endl;

	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image == 0) continue;
	
		DPoint sum_v, sum_vv, sum_pvv;
    	int num = 0;
	
		(*image)->print_residuals(_obj_points,
			sum_v, sum_vv, sum_pvv, num, os);
		
		tot_v += sum_v;
		tot_vv += sum_vv;
		tot_pvv += sum_pvv;
		tot_num += num;
		}
	os << endl;

	fix2.width(10);

	os << "Summary for all image points of all images\n";
	os << "------------------------------------------\n";
	os << "                                vx/um        vy/um\n\n";
	os << "Total number of points :   " << tot_num << endl;
	os << "RMS of residuals       :   "
		<< fix2(sqrt(tot_vv.x / tot_num) * 1e6) << "   "
		<< fix2(sqrt(tot_vv.y / tot_num) * 1e6) << endl;
	os << "Average of residuals   :   "
		<< fix2(tot_v.x * 1e6 / tot_num) << "   "
		<< fix2(tot_v.y * 1e6 / tot_num) << endl;
	os << endl;

	fix2.width(0);
	Form sci3(3); sci3.scientific();

	double vtpv = tot_pvv.x + tot_pvv.y;
	double sigma0 = (_f > 0) ? sqrt(vtpv / double(_f)) : (0.);
	os << "Contribution to vtpv and sigma 0:   "
		<< sci3(vtpv) << " m^2   ";
	//	<< fix2(sqrt(vtpv / double(_f)) * 1e6) << " um\n";
	if (sigma0 == 0.)
		os << "undefined, no redundancy!\n";
	else
		os << fix2(sigma0 * 1e6) << " um\n";
	os << "Redundancy :   " << _f << endl;
	os << endl;
}


// update cameras

void BundleAdjustment::update_cameras(ostream* os)
{
	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam == 0) continue;

		bool rtn = (*cam)->update_io();
		if (os)
			{
			(*os) << "Camera '" << (*cam)->get_id() <<"'";
			if (rtn == true)
				(*os) << " updated\n";
			else
				(*os) << " no update!\n";
			}
		}
	if (os) (*os) << endl;
}


// update image coords.
/*
void BundleAdjustment::update_image_coords(ostream& os) const
{
	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)		
		{
		if (*image != 0) (*image)->update_image_coords(_obj_points, os);
		}
	os << "stop-dep\n";
}
*/

// write initial direction
/*
void BundleAdjustment::write_init_direction(ostream& os)
{
	if (scan_eos.empty())
		cout << "ATTENTION: List of scanner EOs is empty!" << endl;

	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->write_init_direction(os);
}


// write initial direction

void BundleAdjustment::print_init_direction(ostream& os)
{
	os << "Initial direction of scanner EO\n";
	os << "===============================\n";
	
	if (scan_eos.empty())
		os << "ATTENTION: List of scanner EOs is empty!" << endl;

	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->print_init_direction(os);

	os << endl;
}
*/


// write cameras

void BundleAdjustment::write_cameras(ostream& os) const
{
	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		if (*cam != 0) (*cam)->write(os);
	os << "stop-dep\n";
}


// write image coords.
/*
void BundleAdjustment::write_image_coords(ostream& os) const
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->write_image_coords(os);
	os << "stop-dep\n";
}
*/

// write photo coords.
/*
void BundleAdjustment::write_photo_coords(ostream& os) const
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->write_photo_coords(os);
	os << "stop-dep\n";
}
*/

// write object coords.

void BundleAdjustment::write_object_coords(ostream& os) const
{
	os << "indiv-sdev\n";
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) (*op)->write(os);
	os << "stop-dep\n";
}


// write exterior orientations

void BundleAdjustment::write_ext_orientation(ostream& os) const
{
	os << "indiv-type\n";
	vector<ImageGeometry*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0) (*stat)->write(os);
	os << "stop-dep\n";
}


// write AP sets

void BundleAdjustment::write_ap_sets(ostream& os) const
{
	vector<APSet*>::const_iterator ap;
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		if (*ap != 0) (*ap)->write(os);
	os << "stop-dep\n";
}


// write object point analysis

void BundleAdjustment::write_object_point_analysis(ostream& os) const
{
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		
		// point ID, object point coords.
		os << form("%12s %12.3f %12.3f %12.3f",
			(*op)->get_id().c_str(), (*op)->crds.x, (*op)->crds.y, (*op)->crds.z);
				
		// std. devs.
		if (AdjObject::lspars.is_rsdev((*op)->sdev.x))
			os << form(" %8.3f", (*op)->sdev.x);
		else
			os << form(" %8.1e", (*op)->sdev.x);
				
		if (AdjObject::lspars.is_rsdev((*op)->sdev.y))
			os << form(" %8.3f", (*op)->sdev.y);
		else
			os << form(" %8.1e", (*op)->sdev.y);
				
		if (AdjObject::lspars.is_rsdev((*op)->sdev.z))
			os << form(" %8.3f", (*op)->sdev.z);
		else
			os << form(" %8.1e", (*op)->sdev.z);
		
		// type
		os << form(" %1d", (*op)->get_type());
		
		// calculate control point residuals
		DPoint3 v(0., 0., 0.);
		char cr[] = "000";
		
		if ((*op)->state(0) == obs)
			{
			v.x = (*op)->crds.x - (*op)->ival(0);
			cr[0] = '1';
			}
		
		if ((*op)->state(1) == obs)
			{
			v.y = (*op)->crds.y - (*op)->ival(1);
			cr[1] = '1';
			}
		
		if ((*op)->state(2) == obs)
			{
			v.z = (*op)->crds.z - (*op)->ival(2);
			cr[2] = '1';
			}
		
		// calculate check point differences
		DPoint3 d(0., 0., 0.);
		char cs[] = "000";
		
		if ((*op)->is_chx())
			{
			d.x = (*op)->ival(0) - (*op)->crds.x;
			cs[0] = '1';
			}
		
		if ((*op)->is_chy())
			{
			d.y = (*op)->ival(1) - (*op)->crds.y;
			cs[1] = '1';
			}
		
		if ((*op)->is_chz())
			{
			d.z = (*op)->ival(2) - (*op)->crds.z;
			cs[2] = '1';
			}
		
		os << form(" %3s %8.3f %8.3f %8.3f", cr, v.x, v.y, v.z);
		os << form(" %3s %8.3f %8.3f %8.3f\n", cs, d.x, d.y, d.z);
		}
	os << "stop-dep\n";
}


// update scanner EO (and clear INS corr. and CGI config.)
/*
void BundleAdjustment::update_scan_eo(ostream* os)
{
	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->update(os);
	
	vector<INSCor*>::iterator icor;
	for (icor = ins_cors.begin(); icor != ins_cors.end(); ++icor)
		if (*icor != 0) (*icor)->clear();
	
	vector<CGICfg*>::iterator cgic;
	for (cgic = cgi_cfgs.begin(); cgic != cgi_cfgs.end(); ++cgic)
		if (*cgic != 0) (*cgic)->clear();
}
*/

// write scanner EO
/*
void BundleAdjustment::write_scan_eo(const char* stem, ostream* out)
{
	if (out)
		{
		(*out) << "Write scanner EO" << endl;
		(*out) << "================" << endl;
		(*out) << "(No replace! Fail if file exists)" << endl;
		(*out) << "stem=" << stem << endl;
		(*out) << endl;
		}
	
	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0)
			{
			char fname[DEP_MAXNAME];
			sprintf(fname, "%s_%d.eo", stem, (*eo)->get_id());
			if (out)
				{
				(*out) << "Strip " << (*eo)->get_id() << " -> " << fname << endl;
				(*out).flush();
				}

			ofstream os(fname, ios::out|ios::noreplace);
			if (!os)
				cout << "ATTENTION: Cannot create output file: "
					<< fname << endl;
			else
				(*eo)->write(os);
			}
}
*/


// find time interval

int BundleAdjustment::time_interval(double& t_min, double& t_max) const
{
	int num = 0;

	vector<ImageGeometry*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0)
			{
			t_min = t_max = (*stat)->get_time();
			num++;
			break;
			}

	for (; stat != _stations.end(); ++stat)
		if (*stat != 0)
			{
			double t = (*stat)->get_time();
			if (t < t_min) t_min = t;
			if (t > t_max) t_max = t;
			num++;
			}
	
	return num;
}


// set reference time for position and attitude correction

void BundleAdjustment::set_reference_time(double t0)
{
	vector<PosCorrection*>::iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		if (*pcor != 0) (*pcor)->set_reference_time(t0);

	vector<AttCorrection*>::iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		if (*acor != 0) (*acor)->set_reference_time(t0);
}


// remove image point in image list

int BundleAdjustment::remove_image_point(const string& id, ostream* os)
{
	int removed = 0;

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0 && (*image)->remove_image_point(id))
			{
			if (os)
				(*os) << "ATTENTION: Image '" << (*image)->get_id()
					<< "' point '" << id << "' removed.\n";
			removed++;
			}
		}

	return removed;
}


// count number of all object points

int BundleAdjustment::num_obpts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && !(*op)->is_unused()) n++;
		}

	return n;
}


// count number of control points

int BundleAdjustment::num_copts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && (*op)->is_ctrl()) n++;
		}

	return n;
}


// count number of check points

int BundleAdjustment::num_chpts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && (*op)->is_check()) n++;
		}

	return n;
}


// solve normal equation system

bool BundleAdjustment::solve()
{
	assert(_adj_interface != 0);
	
	int info = _adj_interface->solve(_atpa, _atpl, _dx, _dec_vec);
	assert(info >= 0);
	
	if (info > 0)
		{
		cout << "ATTENTION: Matrix is >>> not <<< positive definite at row/column "
			<< (info - 1) << endl;
		return false;
		}
	
	return true;
}


// inversion of normal matrix

bool BundleAdjustment::inversion()
{
	assert(_adj_interface != 0);
	
	int info = _adj_interface->inversion(_atpa, _qxx, _dec_vec);
	assert(info == 0);
	
	return true;
}


// write structure of normal matrix in PBM-format

void BundleAdjustment::write_normal_matrix_struct(ostream& os) const
{
	assert(_adj_interface != 0);
	_adj_interface->write_matrix_struct(os, _atpa);
}


// write elements of normal matrix in ASCII-format

void BundleAdjustment::write_normal_matrix(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _atpa.rows() << " " << _atpa.cols() << endl;
	for (int i = 0; i < _atpa.rows(); i++)
		{
		for (int j = 0; j < _atpa.cols(); j++)
			os << _atpa(i,j) << " ";
			//os << sci(_atpa(i,j)) << " ";
		os << endl;
		}
}


// write elements of right side vector in ASCII-format

void BundleAdjustment::write_rhs_vector(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _atpl.size() << endl;
	for (int i = 0; i < _atpl.size(); i++)
		os << _atpl(i) << endl;
		//os << sci(_atpl(i)) << endl;
}


// write elements of solution vector in ASCII-format

void BundleAdjustment::write_solution_vector(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _dx.size() << endl;
	for (int i = 0; i < _dx.size(); i++)
		os << _dx(i) << endl;
		//os << sci(_dx(i)) << endl;
}

} // namespace DEP
