/* $Id: bunadj.h,v 1.8.2.1.2.4 2007/03/07 17:32:26 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_BUNADJ_INCLUDED
#define DEP_BUNADJ_INCLUDED

#include "depdef.h"
#include "params.h"
#include "adjobj.h"
#include "aolist.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class AdjInterface;
class Camera;
class Image;
class ObjectPoint;
class ImageGeometry;
class APSet;
class PosCorrection;
class AttCorrection;
//class OCTransform;


/** Bundle Adjustment class.
 *  Block of images with inner and exterior orientation,
 *  plus the object point.
 *  @author Dirk Stallmann
 */
class BundleAdjustment
{
public:
	/// erase all images, cameras, etc.
	void clear();

	/// constructor
	BundleAdjustment();

	/// destructor
	~BundleAdjustment();

	/// read configuration
	int read_config(istream& cfg_in);

	/**@name print functions */
	//@{
		/// print parameters/options
		void print_params(ostream& os) const { _pars.print(os); }

		/// print least-squares parameters/options
		//void print_ls_params(ostream& os) const { _ls_pars.print(os); }
		void print_ls_params(ostream& os) const {
			AdjObject::lspars.print(os);
		}

		/// print cameras
		void print_cameras(ostream& os) const;
		//void print_calib(ostream& os);

		/// print images
		void print_images(ostream& os) const;

		/// print object points
		void print_object_points(ostream& os) const;

		//void print_ctrl_supp(ostream& os) const;

		/// print exterior orientations
		void print_ext_orientations(ostream& os) const;
		
		/// print AP sets
		void print_ap_sets(ostream& os) const;

		/// print position corrections
		void print_pos_corrs(ostream& os) const;

		/// print attitude corrections
		void print_att_corrs(ostream& os) const;

		//void print_ins_cor(ostream& os) const;
		//void print_scan_eo(ostream& os) const;
		//void print_cgi_cfg(ostream& os) const;

		/// print object coord. transformation
		//void print_oc_transf(ostream& os) const;
	//@}

	/**@name main functions */
	//@{
		/// initialize
		int initialize(ostream& os);

		/// perform adjustment
		int adjustment(ostream& os);

		/// print unknowns parameters
		void print_unknowns(ostream& os) const;

		/// control point residuals
		void ctrl_point_residuals(ostream& os) const;
	
		/** print check point analysis.
		 *  The practical precision estimates/empirical accuracy
		 *  are computed from all check point coordinate differences.
		 *  The accuracy will be derived ONLY from NEW object points.
		 */
		void check_point_analysis(ostream& os) const;

		/// calculate precision
		int precision(ostream* os=0);

		/** print precision for estimated parameters
		 *  (partially with significance test)
		 */
		void precision(ostream& os) const;

		/// print correlations
		void correlations(ostream& os) const;

		/// print correlation matrices between specific adjustment objects
		void correlations_spec(ostream& os) const;

		/// print influence of APs on image points
		void aps_influence(ostream& os) const;
	//@}

	/**@name support functions */
	//@{
		/// remove image points of unused object points
		void clean_image_points(ostream* os=0);

		/// check object point approx.
		int check_approx(ostream* os=0);

		/// check number of rays per object point
		void check_rays_per_point(ostream* os=0);

		/// check number of rays per image
		void check_rays_per_image(ostream* os=0);

		/// calculate approximate image geometry
		bool calc_approx_image_geometry(ostream* os=0, bool dbg=false);

		/// calculate object point approx.
		int calc_object_point_approx(ostream* os=0, bool dbg=false);

		/// removed unused objects
		void clean_up(ostream* os=0);

		/// v-check
		void check_atpv(ostream& os) const;

		/// average change of unknowns
		void average_change_unknowns(ostream& os) const;

		/// update unknowns
		void update_unknowns(ostream& os);

		/// average precision of object coords.
		void average_precision(ostream& os) const;

		/// write residual of image coords.
		void write_residuals(ostream& os) const;

		/// print residual of image coords.
		void print_residuals(ostream& os) const;

		/// update cameras by Ap sets
		void update_cameras(ostream* os=0);
	//@}

	/**@name write functions */
	//@{
		/// write cameras
		void write_cameras(ostream& os) const;

		/// write image coords.
		//void write_image_coords(ostream& os) const;

		/// write photo coords.
		//void write_photo_coords(ostream& os) const;

		/// write object coords.
		void write_object_coords(ostream& os) const;

		/// write exterior orientations
		void write_ext_orientation(ostream& os) const;

		/// write AP sets
		void write_ap_sets(ostream& os) const;

		/** write object point analysis
		 *  Output of all object points with std.devs.,
		 *  object point residuals and check point differences.
		 *
		 *  Output format:
		 *   <id> <X> <Y> <Z> <sX> <sY> <sZ> <type> <cr> <vX> <vY> <vZ> <cs> <dX> <dY> <dZ>
		 *    id            point ID
		 *    X, Y, Z       object point coords.
		 *    sX, sY, sZ    std. devs. of object point coords.
		 *    type          object point type
		 *    cr            control point residuals flags
		 *    vX, vZ, vZ    object point residuals or check point differences
		 *    cs            control support flags
		 *    dX, dY, dZ    object point residuals or check point differences
		 */
		void write_object_point_analysis(ostream& os) const;

		/// initial direction
		//void write_init_direction(ostream& os) const;
		//void print_init_direction(ostream& os) const;

		/// update scanner EO
		//void update_scan_eo(ostream* os=0);
		//void write_scan_eo(const char* stem, ostream* out=0);
	//@}

	/**@name reference time functions for position and attitude correction */
	//@{
		/// find time interval
		int time_interval(double& t_min, double& t_max) const;

		/// set reference time 
		void set_reference_time(double t0);
	//@}

	/// remove image point
	int remove_image_point(const string& id, ostream* os=0);

	/**@name count number of object points */
	//@{
		// count number of all object points
		int num_obpts() const;

		// count number of control points
		int num_copts() const;

		// count number of check points
		int num_chpts() const;

	//@}

private:
	/// get redundancy
	double f() const { return _f; }

	/**@name normal solution interface.
	 * Solving normal equations and calculate their inverse.
	 * Handler functions to use the desired adjustment interface.
	 */
	//@{
		/// solve normal equation system
		bool solve();

		/// inversion of normal matrix
		bool inversion();

		/// write structure of normal matrix in PBM-format
		void write_normal_matrix_struct(ostream& os) const;
				
		/// write elements of normal matrix in ASCII-format
		void write_normal_matrix(ostream& os) const;
				
		/// write elements of ride hand side vector in ASCII-format
		void write_rhs_vector(ostream& os) const;
				
		/// write elements of solution vector in ASCII-format
		void write_solution_vector(ostream& os) const;
	//@}

public:
	/// parameters/options
	Params			_pars;	

	// LS parameters/options
	//LSParams		_ls_pars;

	/// images
	vector<Image*>	_images;

	/// cameras
	vector<Camera*>	_cameras;

	/// object points
	vector<ObjectPoint*>	_obj_points;

	/// stations
	vector<ImageGeometry*>	_stations;
	
	/// AP sets
	vector<APSet*>			_ap_sets;

	/// position corrections
	vector<PosCorrection*>	_pos_corrs;

	/// attitude corrections
	vector<AttCorrection*>	_att_corrs;

	// object coordinate transformation.
	//  Only one transformation is defined. This feature is optional.
	//
  	//OCTransform*			_oc_transf;

   	// orbits
   	//vector<Orbit*>        _orbits;

private:
	/// list of adjustment objects
	AdjObjectList	_list;

	/// normal matrix, only the upper triangle is stored
	Matrix<double>	_atpa;

	/// right-hand side (rhs)
	Vector<double>	_atpl;

	/// decomposition vector
	Vector<double>	_dec_vec;

	/// solution vector
	Vector<double>	_dx;

	/// co-factor matrix
	Matrix<double>	_qxx;

	/// sum of square of l vector
	double		_ltpl;

	/// sum of square of residual vector
	double		_vtpv;

	/// sigma of unit weight a posteriori
	double		_sigma0;

	/// redundancy
	int			_f;

	/// adjustment interface
	AdjInterface*	_adj_interface;
};

} // namespace DEP

#endif
