/* $Id: camera.h,v 1.8.4.2 2007/03/06 08:47:28 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_CAMERA_INCLUDED
#define DEP_CAMERA_INCLUDED

#include "depdef.h"
#include "drvsub.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class Params;
class APSet;
class LinObsEqn;
class LinConditionEqn;
class Image;

/** Camera class.
 *  Common base class that implement a ''pure'' frame camera.
 *  This class contains only the three parameters of the inner
 *  orientation (IO): focal length and principal point coordinates. 
 *  The concept of IO follows the Manual of Photogrammetry
 *  (Slama et al., 1980).
 *
 *  This class define three types of coordinate systems:
 *  \begin{itemize}
 *  \item photo coordinates (3-D),
 *  \item image coordinates (2-D) and
 *  \item external (image) coordinates (2-D).
 *  \end{itemize}
 *  Unlike image coordinates photo coordinates are 3-D coordinates.
 *  The third dimension is defined by the focal length.
 *  We assume for this base class the image and external coordinates
 *  are identical.
 *  A new derived camera class, e.g. a CCD camera, have to provide
 *  the transformation between image and external coordinates.
 *
 *  The IO parameters are used here as constants, not as unknowns!
 *  Therefore this class is not dervied from the class AdjObject.
 *  But this class has a vector of pointers to Additional Parameters
 *  (AP) sets. The AP sets modelling systematic errors in image space.
 *  The option to connect the camera with many AP sets gives
 *  the possibility to combine different AP sets to model
 *  various sources of systematic errors.
 *  So the number of unknown parameters can vary with the type
 *  and the number of connected AP sets.
 *
 *  @author Dirk Stallmann
 */
class Camera
{
	friend class APSet;

public:
	/// constructor
	Camera();
	
	/// destructor
	virtual ~Camera() {}

	/// read
	virtual int read(istream& is);

	/// write
	virtual void write(ostream& os) const;

	/// print
	virtual void print(ostream& os) const;

	/// get camera type 
	virtual string type_str() const;

	/// get (external) coordinate system type
	virtual string csys_str() const;

	/**@name external/image coordinates transformation
	 *  To convert between external (image) and (internal) image coordinates.
 	 *  In this special case external and internal image coordinates
	 *  are identical!
	 *  Furthermore a derived class which implements e.g. a CCD camera will
	 *  override this functions to provide the pixel/image coordinate
	 *  transformation.
	 */
	//@{
	 	/// external-to-image coordinates transformation
		virtual void external_to_image(const DPoint2& ext_c,
			DPoint2& img_c) const {
			img_c = ext_c;
		}

		/// image-to-external-coords. transformation
		virtual void image_to_external(const DPoint2& img_c,
			DPoint2& ext_c) const {
			ext_c = img_c;
		}
	//@}

	/** transformation of standard deviations
	 *  from external to image coordinates
	 */
	virtual void sdev_external_to_image(const DPoint2& ext_s,
		DPoint2& img_s) const {
			img_s = ext_s;
	}

	/**@name read/write/ print image coordinates.
	 */
	//@{
		/// read image coords.
		virtual bool read_image_coords(istream& is, Image& image) const;

		/// write image coords.
		virtual void write_image_coords(ostream& os, const Image& image) const;

		/// print image coords.
		virtual void print_image_coords(ostream& os, const Image& image) const;
	//@}

	/**@name set/get member data
	 */
	//@{
		/// get camera ID
		const char* get_id() const { return _id.c_str(); }

		/// get focal length
		double get_focal_len() const { return _focal_len; }

		/// get focal length
		double ck() const { return _focal_len; }

		/// get principal point coords.
		const DPoint& get_pp() const { return _princ_pnt; }
	//@}

	/// set inuse flag
	void inuse(bool flg) { _inuse = flg; }

	/// test inuse flag
	bool is_unused() const { return _inuse == false; }

	/**@name image/photo coordinates transformation
	 *  To convert between image and photo coordinates.
 	 *  Correction of APs are included!
	 */
	//@{
	 	/// image-to-photo coordinates transformation
		void image_to_photo(const DPoint2& img_c, DPoint3& pho_c,
			bool corr_flg=true) const;

		/// photo-to-image-coords. transformation
		void photo_to_image(const DPoint3& pho_c, DPoint2& img_c,
			bool corr_flg=true) const;
	//@}

	/**@name AP related functions
	 */
	//@{
		/// attach AP set
		void attach(APSet* aps) { _ap_sets.push_back(aps); }

		/// check link to AP set
		bool has_ap_set() const { return !_ap_sets.empty(); }

		/// get AP set correction
		DPoint2 ap_set_corr(const DPoint3& pho_c) const;

		/// add subset of derivatives to linear observation equations
		void add(LinObsEqn& eqn) const;

		/// add subset of derivatives to linear condition equations
		void add_a(LinConditionEqn& eqn) const;

		// initialize partial derivatives of IO parameters
		//void init_derivs(DerivSub& derivs) const; 

		/// get number of AP sets
		size_t num_ap_sets() const { return _ap_sets.size(); }

		/// get partial derivatives of IO parameters
		void update_derivs(const DPoint3& pho_c);
		//void get_derivs(const DPoint2& img_c, vector<DerivSub>& derivs) const;

		/// get partial derivatives of IO parameters
		void get_derivs(const DPoint3& pho_c, vector<DerivSub>& derivs) const;
	
		/// update IO parameter using optional AP set
		bool update_io(); 
	//@}

	// attach calibration (placeholder only)
	//int attach(vector<Calib*>& calibs);

protected:
	/// camera ID
	string		_id;

	/// comment
	string		_comment;

	/// focal length
	double		_focal_len;

	/// principal point coordinates
	DPoint2		_princ_pnt;

	/// image size in x (width)
	double		_size_x;

	/// image size in y (height)
	double		_size_y;

	/// inuse flag
	bool		_inuse;

	/// AP set list (maybe empty)
	vector<APSet*>	_ap_sets;
};


// non-member functions ...

} // namespace DEP

#endif
