/* $Id: ccdcam.cpp,v 1.3.4.1 2005/12/12 12:27:29 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "ccdcam.h"
#include "image.h"
#include "imgpnt.h"

namespace DEP
{

// constructor

CameraCCD::CameraCCD() : Camera(),
	_nc(0), _nr(0),
	_sc(0.), _sr(0.),
	_c0(0.), _r0(0.) { }


// read camera data

int CameraCCD::read(istream& is)
{
	// identifier
	is >> _id;
	if (!is) return ERROR;

	// comment
	is >> _comment;
	if (!is) return ERROR;

	// focal length and principle point in mm
	is >> _focal_len >> _princ_pnt;
	if (!is) return ERROR;

	// convert from mm to m
	_princ_pnt *= 1e-3;
	_focal_len *= 1e-3;

	// sensor size and pixel spacing
	is >> _nc >> _nr >> _sc >> _sr;
	if (!is) return ERROR;

	// convert from um/pixel to m/pixel
	_sc *= 1e-6;
	_sr *= 1e-6;

	// calc. ICS location
	_c0 = (_nc - 1) / 2.;
	_r0 = (_nr - 1) / 2.;

	// calc. image format
	_size_x = _nc * _sc;
	_size_y = _nr * _sr;

	return OK;
}


// write camera data

void CameraCCD::write(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);
	Form fix6(6); fix6.fixed().width(12);

	os << "ccd-frame   " << _id << "   " << _comment << endl;
	os << fix6(_focal_len * 1e3) << " "
		<< fix6(_princ_pnt.x * 1e3) << " " << fix6(_princ_pnt.y * 1e3) << endl;
	os << fix6(_size_x * 1e3) << " " << fix6(_size_y * 1e3) << endl;
	os << _nc << " " << _nr << endl;
	os << fix3(_sc * 1e6) << " " << fix3(_sr * 1e6);
	os << endl;
}


// print camera data

void CameraCCD::print(ostream& os) const
{
	Form int10; int10.width(10);
	Form fix3(3); fix3.fixed().width(10);

	Camera::print(os);

	os << "Number of pixel    :  nc = " << int10(_nc) << " pix     "
		<< "nr = " << int10(_nr) << " pix\n";
	os << "ICS location       :  c0 = " << fix3(_c0) << " pix     "
		<< "r0 = " << fix3(_r0) << " pix\n";
	os << "Pixel spacing      :  sc = " << fix3(_sc * 1e6) << " um/pix  "
		<< "sr = " << fix3(_sr * 1e6) << " um/pix\n";
	os << endl;
}


// get camera type 

string CameraCCD::type_str() const
{
	ostringstream oss;
	oss << "CCD-frame";
	return oss.str();
}


// get coordinate system type 

string CameraCCD::csys_str() const
{
	ostringstream oss;
	oss << "pixel (top-left)";
	return oss.str();
}


// external-to-image-coords. transformation

void CameraCCD::external_to_image(const DPoint2& ext_c, DPoint2& img_c) const
{
	//cout << "CameraCCD::external_to_image" << endl;
	//img_c.x = (pix_c.x - _cp) * _sc;
	//img_c.y = (_rp - pix_c.y) * _sr;
	img_c.x =   (ext_c.x - _c0) * _sc;
	img_c.y = - (ext_c.y - _r0) * _sr;
}


// image-to-external-coords. transformation

void CameraCCD::image_to_external(const DPoint2& img_c, DPoint2& ext_c) const
{
	//cout << "CameraCCD::image_to_external" << endl;
	ext_c.x = _c0 + img_c.x / _sc;
	ext_c.y = _r0 - img_c.y / _sr;
}


// transformation of standard deviations from external to image coordinates

void CameraCCD::sdev_external_to_image(const DPoint2& ext_s,
DPoint2& img_s) const
{
	img_s.x = ext_s.x * _sc;
	img_s.y = ext_s.y * _sr;
}


// read image coords.
/*
bool CameraCCD::read_image_coords(istream& is, Image& image) const
{
	string id;

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0)
			break; // stop reading

		ImagePoint *ip = new ImagePoint;
		assert(ip != 0);
		ip->id(id);

		// read pixel coords. and std. dev.
		is >> ip->crds >> ip->sdev;
		if (!is) return false;

		// append or replace point
		image.add_image_point(ip);
		}

	return true;
}
*/


// print image coords.

void CameraCCD::print_image_coords(ostream& os, const Image& image) const
{
	Form str8; str8.width(8);
	Form fix8(4); fix8.fixed().width(8);
	Form fix10(4); fix10.fixed().width(10);

	os << "Number of points   :  " << image.count_image_points() << endl;
	os << "Coord. system type :  " << csys_str() << endl;
	os << endl;

	os << "      ID     x/pix      y/pix    sx/pix   sy/pix\n\n";
	vector<ImagePoint*>::const_iterator ip;
	for (ip = image._image_points.begin(); ip != image._image_points.end();
		++ip)
		{
		if (*ip == 0) continue;
		os << str8((*ip)->get_id())
			<< " " << fix10((*ip)->crds.x)
			<< " " << fix10((*ip)->crds.y)
			<< " " << fix8((*ip)->sdev.x)
			<< " " << fix8((*ip)->sdev.y) << endl;
		}
	os << endl;
}


// non-member functions ...

} // namespace DEP
