/* $Id: dlt.cpp,v 1.3.4.1 2005/12/12 12:27:29 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "dlt.h"

namespace DEP
{

// get type name

string DirectLinearTransf::type_str() const
{
	ostringstream oss;
	oss << "DLT";
	return oss.str();
}


// get key name

string DirectLinearTransf::key_str() const
{
	ostringstream oss;
	oss << "DLT";
	return oss.str();
}


// get parameter name

string DirectLinearTransf::uname(size_t i) const
{
	ostringstream oss;
	assert(i < _size);
	oss << "L" << (i + 1);
	return oss.str();
}


// object-to-photo coords. transformation

bool DirectLinearTransf::object_to_photo(const DPoint3& obj_c,
DPoint3& pho_c) const
{
	DPoint2 img_c;

	dlt_projection(_a, obj_c, img_c);

	pho_c.x = img_c.x;
	pho_c.y = img_c.y;

	return true;
}


// spatial intersection

bool DirectLinearTransf::intersection(const ImageGeometry& g2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	try {
		const DirectLinearTransf& dlt2 =
			dynamic_cast<const DirectLinearTransf&>(g2);

		return intersection(dlt2, pho_c1, pho_c2, obj_c, e);
	}
	catch (bad_cast) {
		cout << "DirectLinearTransf::intersection: "
			"Incompatible image geometries!\n";
		return false;
	}
}


// get partial derivatives

bool DirectLinearTransf::get_derivs(const DPoint3& pho_c, const DPoint3& obj_c,
DerivSub& deriv_oc, DerivSub& deriv_eo,
DerivSub& deriv_pc, DerivSub& deriv_ac, DerivSub& deriv_obs)
{
	// adjust sub-vectors
	deriv_oc.newsize(2,3);
	deriv_eo.newsize(2,11);

	// denominator
	double q = _a[8] * obj_c.x + _a[9] * obj_c.y + _a[10] * obj_c.z + 1.;
	if (fabs(q) == 0.) cerr << "DirectLinearTransf::get_derivs: "
		"zero denominator.\n";

	// nominators
	double m = _a[0] * obj_c.x + _a[1] * obj_c.y + _a[2] * obj_c.z + _a[3];
	double n = _a[4] * obj_c.x + _a[5] * obj_c.y + _a[6] * obj_c.z + _a[7];

	// photo coords. derived from unknowns.
	// Use these values not the observations !!!
	double x = m / q;
	double y = n / q;

	// derivatives
	deriv_eo(0, 0) = obj_c.x / q;
	deriv_eo(0, 1) = obj_c.y / q;
	deriv_eo(0, 2) = obj_c.z / q;
	deriv_eo(0, 3) = 1. / q;
	deriv_eo(0, 4) = 0.;
	deriv_eo(0, 5) = 0.;
	deriv_eo(0, 6) = 0.;
	deriv_eo(0, 7) = 0.;
	//deriv_eo(0, 8) = -pho_c.x * obj_c.x / q;
	//deriv_eo(0, 9) = -pho_c.x * obj_c.y / q;
	//deriv_eo(0,10) = -pho_c.x * obj_c.z / q;
	deriv_eo(0, 8) = -x * obj_c.x / q;
	deriv_eo(0, 9) = -x * obj_c.y / q;
	deriv_eo(0,10) = -x * obj_c.z / q;

	deriv_eo(1, 0) = 0.;
	deriv_eo(1, 1) = 0.;
	deriv_eo(1, 2) = 0.;
	deriv_eo(1, 3) = 0.;
	deriv_eo(1, 4) = obj_c.x / q;
	deriv_eo(1, 5) = obj_c.y / q;
	deriv_eo(1, 6) = obj_c.z / q;
	deriv_eo(1, 7) = 1. / q;
	//deriv_eo(1, 8) = -pho_c.y * obj_c.x / q;
	//deriv_eo(1, 9) = -pho_c.y * obj_c.y / q;
	//deriv_eo(1,10) = -pho_c.y * obj_c.z / q;
	deriv_eo(1, 8) = -y * obj_c.x / q;
	deriv_eo(1, 9) = -y * obj_c.y / q;
	deriv_eo(1,10) = -y * obj_c.z / q;

	//deriv_oc(0,0) = (_a[0] - pho_c.x * _a[8]) / q;
	//deriv_oc(0,1) = (_a[1] - pho_c.x * _a[9]) / q;
	//deriv_oc(0,2) = (_a[2] - pho_c.x * _a[10]) / q;
	deriv_oc(0,0) = (_a[0] - x * _a[8]) / q;
	deriv_oc(0,1) = (_a[1] - x * _a[9]) / q;
	deriv_oc(0,2) = (_a[2] - x * _a[10]) / q;

	//deriv_oc(1,0) = (_a[4] - pho_c.y * _a[8]) / q;
	//deriv_oc(1,1) = (_a[5] - pho_c.y * _a[9]) / q;
	//deriv_oc(1,2) = (_a[6] - pho_c.y * _a[10]) / q;
	deriv_oc(1,0) = (_a[4] - y * _a[8]) / q;
	deriv_oc(1,1) = (_a[5] - y * _a[9]) / q;
	deriv_oc(1,2) = (_a[6] - y * _a[10]) / q;
	//cout << "deriv_oc: " << deriv_oc << endl;
	
	return true;
}


// spatial intersection

bool DirectLinearTransf::intersection(const DirectLinearTransf& dlt2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	DPoint2 img_c1(pho_c1.x, pho_c1.y);
	DPoint2 img_c2(pho_c2.x, pho_c2.y);

	return dlt_intersection(this->_a, img_c1, dlt2._a, img_c2, obj_c);
}


// non-member functions ...

// project object point onto image plane using DLT

void dlt_projection(const double* l, const DPoint3& obj_c, DPoint2& img_c)
{
	assert(l != 0);

	// denominator
	double q = l[8] * obj_c.x + l[9] * obj_c.y + l[10] * obj_c.z + 1.;
	if (fabs(q) == 0.) cerr << "DEP::projection: zero denominator.\n";

	// nominators
	double m = l[0] * obj_c.x + l[1] * obj_c.y + l[2] * obj_c.z + l[3];
	double n = l[4] * obj_c.x + l[5] * obj_c.y + l[6] * obj_c.z + l[7];

	// image coords.
	img_c.x = m / q;
	img_c.y = n / q;
}


// intersection using DLT

bool dlt_intersection(const double* l1, const DPoint2& pho_c1, 
const double* l2, const DPoint2& pho_c2, DPoint3& obj_c)
{
	cout << "DirectLinearTransf::intersection: Not implemented yet!\n";
	assert(false);

	return false;
}

} // namespace DEP
