/* $Id: eopokrot.cpp,v 1.5.2.1.2.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "eopokrot.h"

namespace DEP
{

// read EO parameters
 
int EOAnglesPOKRot::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading ID failed.\n";
		return ERROR;
		}

	// unit of rotation angles and/or standard deviations
	string str;
	is >> str;
	if (!is)
		{
		cout << "ERROR: Reading unit of rotation angles failed.\n";
		return ERROR;
		}

	UnitAngle u_angle;
	if (!u_angle.set(str.c_str()))
		{
		cout << "ERROR: Unknown unit of rotation angles:" << str << endl;
		return ERROR;
		}
	double rho = u_angle.rho();

	// exposure time
	is >> _time;
	if (!is)
		{
		cout << "ERROR: Cannot read exposure time.\n";
		return ERROR;
		}

	// projection centre coords.
	is >> _a[0] >> _a[1] >> _a[2];
	if (!is)
		{
		cout << "ERROR: Reading projection centre coords. failed.\n";
		return ERROR;
		}

	// std. dev. projection centre coords.
	is >> _sa[0] >> _sa[1] >> _sa[2];
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of proj. centre coords. failed.\n";
		return ERROR;
		}

	// rotation parameters 
	is >> _a[3] >> _a[4] >> _a[5];
	if (!is)
		{
		cout << "ERROR: Reading rotation parameters failed.\n";
		return ERROR;
		}

	// conversion into radian
	_a[3] /= rho;
	_a[4] /= rho;
	_a[5] /= rho;

	// std. dev. of rotation parameters
	is >> _sa[3] >> _sa[4] >> _sa[5];
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of rotation angles failed.\n";
		return ERROR;
		}

	// only for real std. devs. conversion into radians
	if (is_rsdev(_sa[3])) _sa[3] /= rho;
	if (is_rsdev(_sa[4])) _sa[4] /= rho;
	if (is_rsdev(_sa[5])) _sa[5] /= rho;

	return OK;
}


// write EO parameters

void EOAnglesPOKRot::write(ostream& os) const
{
	double sc = AdjObject::lspars.scale_obj();
	double rho = AdjObject::lspars.rho();

	Form sci(6); sci.scientific().width(15);
	Form fix(1); fix.fixed().width(15);

	// type
	os << key_str() << " ";

	// ID
	os << _id << " ";

	// unit of rotation angles
	os << AdjObject::lspars.u_angle_cstr() << " ";

	// exposure time
	fix.precision(4);
	os << fix(_time) << endl;

	size_t i = 0;

	// projection centre
	fix.precision(4);
	for (i = 0; i < 3; i++)
		os << " " << fix(_a[i] * sc);
	os << endl;

	for (i = 0; i < 3; i++)
		{
		if (is_rsdev(_sa[i]))
			os << " " << fix(_sa[i] * sc);
		else
			os << " " << sci(_sa[i]);
		}
	os << endl;

	// rotation parameters
	for (i = 3; i < _size; i++)
		os << " " << sci(_a[i] * rho);
	os << endl;

	for (i = 3; i < _size; i++)
		os << " " << sci(_sa[i]);
	os << endl;
}


// print parameters

void EOAnglesPOKRot::print(ostream& os) const
{
	double sc = AdjObject::lspars.scale_obj();
	double rho = AdjObject::lspars.rho();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(4); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Type          :  " << type_str() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";

	size_t i = 0;

	fix.precision(3);
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(_sa[i])) << "   "
			<< fix(_a[i] * sc) << "   ";
		if (is_rsdev(_sa[i]))
			os << fix(_sa[i] * sc);
		else
			os << sci(_sa[i]);
		os << endl;
		}

	fix.precision(5);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(_sa[i])) << "   "
			<< fix(_a[i] * rho) << "   ";
		if (is_rsdev(_sa[i]))
			os << fix(_sa[i] * rho);
		else
			os << sci(_sa[i]);
		os << endl;
		}
	os << endl;

	vector<double> a;
	RotMatrix3<double> R;
	get_rot_params(a);
	calc_rot_matrix(R, a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// print changes of unknowns without header

void EOAnglesPOKRot::print_changes(ostream& os,
const Vector<double>& dx) const
{
	double sc = AdjObject::lspars.scale_obj();
	double rho = AdjObject::lspars.rho();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(6); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Type       :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;
   
	os << "  Name       State"
		"       Approx.         Change       Adjusted\n\n";
	
	size_t i = 0;
	int k = _index;

	for (; i < 3; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i) * sc) << "   "
			<< fix(dx(k) * sc) << "   "
			<< fix((par(i) + dx(k)) * sc) << endl;
		}

	fix.precision(5);
	for (; i < _size; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i) * rho) << "   "
			<< fix(dx(k) * rho) << "   "
			<< fix(((par(i) + dx(k)) * rho)) << endl;
		}
   	os << endl;
}


// print unknown parameters

void EOAnglesPOKRot::print_unknowns(ostream& os) const 
{
	double sc = AdjObject::lspars.scale_obj();
	double rho = AdjObject::lspars.rho();
		
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Type       :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";

	size_t i = 0;
	
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i) * sc) << "  "
			<< fix(par(i) * sc) << "  "
			<< fix(change(i) * sc) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * sc);
		else
			os << sci(sdev(i));
		os << endl;
		}
	
	fix.precision(5);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i) * rho) << "  "
			<< fix(par(i) * rho) << "  "
			<< fix(change(i) * rho) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * rho);
		else
			os << sci(sdev(i));
		os << endl;
		}
   	os << endl;

	vector<double> a;
	RotMatrix3<double> R;
	get_rot_params(a);
	calc_rot_matrix(R, a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// get type name

string EOAnglesPOKRot::type_str() const
{
	ostringstream oss;
	oss << "EO (phi-omega-kappa, rotated)";
	return oss.str();
}


// get key name

string EOAnglesPOKRot::key_str() const
{
	ostringstream oss;
	oss << "ext-ori-pok-rot";
	return oss.str();
}


// get parameter name

string EOAnglesPOKRot::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "Xo"; break;
		case 1: oss << "Yo"; break;
		case 2: oss << "Zo"; break;
		case 3: oss << "phi"; break;
		case 4: oss << "omega"; break;
		case 5: oss << "kappa"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// get partial derivatives of rotation parameters

void EOAnglesPOKRot::get_derivs_rot(DerivSub& deriv_eo) const
{
	// phi-derivatives
	deriv_eo(0,3) = _cn2 * (_n * (_rot_matrix(2,0) * _dx
		- _rot_matrix(0,0) * _dz)
		- _zx * (_rot_matrix(2,2) * _dx - _rot_matrix(0,2) * _dz));
	deriv_eo(1,3) = _cn2 * (_n * (_rot_matrix(2,1) * _dx
		- _rot_matrix(0,1) * _dz)
		- _zy * (_rot_matrix(2,2) * _dx - _rot_matrix(0,2) * _dz));
	
	double sin_p = sin(_rot_params[0]);
	double cos_p = cos(_rot_params[0]);
	double sin_o = sin(_rot_params[1]);
	double cos_o = cos(_rot_params[1]);
	double sin_k = sin(_rot_params[2]);
	double cos_k = cos(_rot_params[2]);

	// omega-derivatives
	deriv_eo(0,4) = _cn2 * (_n * _n * sin_k - _zx *
		(-sin_p * sin_o * _dx - cos_o * _dy - cos_p * sin_o * _dz));
	deriv_eo(1,4) = _cn2 * (_n * _n * cos_k - _zy *
		(-sin_p * sin_o * _dx - cos_o * _dy - cos_p * sin_o * _dz));

	// kappa-derivatives
	deriv_eo(0,5) =  _cn * _zy;
	deriv_eo(1,5) = -_cn * _zx;
}


// non-member functions ...

} // namespace DEP
