/* $Id: eoquaternion.cpp,v 1.5.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "eoquaternion.h"
#include "rotmatvec3.h"
#include "lobseqn.h"

namespace DEP
{

// get type name

string EOQuaternion::type_str() const
{
	ostringstream oss;
	oss << "EO (quaternion)";
	return oss.str();
}


// get key name

string EOQuaternion::key_str() const
{
	ostringstream oss;
	oss << "ext-ori-quaternion";
	return oss.str();
}


// get parameter name

string EOQuaternion::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "Xo"; break;
		case 1: oss << "Yo"; break;
		case 2: oss << "Zo"; break;
		case 3: oss << "q0"; break;
		case 4: oss << "q1"; break;
		case 5: oss << "q2"; break;
		case 6: oss << "q3"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// partial derivatives for q0

void EOQuaternion::dr_q0(RotMatrix3<double>& dR) const
{
	dR(0,0) =  2. * _a[3];
	dR(0,1) = -2. * _a[6];
	dR(0,2) =  2. * _a[5];

	dR(1,0) =  2. * _a[6];
	dR(1,1) =  2. * _a[3];
	dR(1,2) = -2. * _a[4];

	dR(2,0) = -2. * _a[5];
	dR(2,1) =  2. * _a[4];
	dR(2,2) =  2. * _a[3];
}


// partial derivatives for q1

void EOQuaternion::dr_q1(RotMatrix3<double>& dR) const
{
	dR(0,0) =  2. * _a[4];
	dR(0,1) =  2. * _a[5];
	dR(0,2) =  2. * _a[6];

	dR(1,0) =  2. * _a[5];
	dR(1,1) = -2. * _a[4];
	dR(1,2) = -2. * _a[3];

	dR(2,0) =  2. * _a[6];
	dR(2,1) =  2. * _a[3];
	dR(2,2) = -2. * _a[4];
}


// partial derivatives for q2

void EOQuaternion::dr_q2(RotMatrix3<double>& dR) const
{
	dR(0,0) = -2. * _a[5];
	dR(0,1) =  2. * _a[4];
	dR(0,2) =  2. * _a[3];

	dR(1,0) =  2. * _a[4];
	dR(1,1) =  2. * _a[5];
	dR(1,2) =  2. * _a[6];

	dR(2,0) = -2. * _a[3];
	dR(2,1) =  2. * _a[6];
	dR(2,2) = -2. * _a[5];
}


// partial derivatives for q3

void EOQuaternion::dr_q3(RotMatrix3<double>& dR) const
{
	dR(0,0) = -2. * _a[6];
	dR(0,1) = -2. * _a[3];
	dR(0,2) =  2. * _a[4];

	dR(1,0) =  2. * _a[3];
	dR(1,1) = -2. * _a[6];
	dR(1,2) =  2. * _a[5];

	dR(2,0) =  2. * _a[4];
	dR(2,1) =  2. * _a[5];
	dR(2,2) =  2. * _a[6];
}


// get derivatives of the quaternion

void EOQuaternion::get_derivs_rot(DerivSub& deriv_eo) const
{
	RotMatrix3<double> dR_q0;
	RotMatrix3<double> dR_q1;
	RotMatrix3<double> dR_q2;
	RotMatrix3<double> dR_q3;

	dr_q0(dR_q0);
	dr_q1(dR_q1);
	dr_q2(dR_q2);
	dr_q3(dR_q3);

	DPoint3 d(_dx, _dy, _dz);

	DPoint3 b_q0;
	DPoint3 b_q1;
	DPoint3 b_q2;
	DPoint3 b_q3;

	mult_inv(b_q0, dR_q0, d);
	mult_inv(b_q1, dR_q1, d);
	mult_inv(b_q2, dR_q2, d);
	mult_inv(b_q3, dR_q3, d);

	deriv_eo(0,3) = _cn2 * (_n * b_q0.x - _zx * b_q0.z);
	deriv_eo(1,3) = _cn2 * (_n * b_q0.y - _zy * b_q0.z);

	deriv_eo(0,4) = _cn2 * (_n * b_q1.x - _zx * b_q1.z);
	deriv_eo(1,4) = _cn2 * (_n * b_q1.y - _zy * b_q1.z);

	deriv_eo(0,5) = _cn2 * (_n * b_q2.x - _zx * b_q2.z);
	deriv_eo(1,5) = _cn2 * (_n * b_q2.y - _zy * b_q2.z);

	deriv_eo(0,6) = _cn2 * (_n * b_q3.x - _zx * b_q3.z);
	deriv_eo(1,6) = _cn2 * (_n * b_q3.y - _zy * b_q3.z);
}


// add weighted constraints

void EOQuaternion::add_constraints(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	//
	LinObsEqn eqn(1);

	// setup sub-matrices of derivatives
	DerivSub deriv(1,_size);

	// set index
	deriv.index(_index);

	// design matrix A
	deriv(0,0) = 0.;
	deriv(0,1) = 0.;
	deriv(0,2) = 0.;
	deriv(0,3) = 2. * _a[3];
	deriv(0,4) = 2. * _a[4];
	deriv(0,5) = 2. * _a[5];
	deriv(0,6) = 2. * _a[6];
	eqn.add(&deriv);

	// weight
	eqn.set_p(0, lspars.constr_wt());

	// misclosure vector
	double l = 1. - _a[3] * _a[3] - _a[4] * _a[4] - _a[5] * _a[5]
		- _a[6] * _a[6];
	eqn.set_l(0, l);

	// add contributions to normal equation
	eqn.add_contr(atpa, atpl, ltpl);
}


// non-member functions ...

} // namespace DEP
