/* $Id: extori.cpp,v 1.8.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "deperr.h"
#include "extori.h"
#include "poscor.h"

namespace DEP
{

// constructor

ExtOrientation::ExtOrientation(size_t sz) : ImageGeometry(sz+3),
	_per_centre(), _rot_params(sz), _rot_matrix(),
	_pos_cor(0)
{
	assert(3 <= sz && sz <= 9);
}


// read
 
int ExtOrientation::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading ID failed.\n";
		return ERROR;
		}

	// exposure time
	is >> _time;
	if (!is)
		{
		cout << "ERROR: Cannot read exposure time.\n";
		return ERROR;
		}

	// projection centre coords.
	is >> _a[0] >> _a[1] >> _a[2];
	if (!is)
		{
		cout << "ERROR: Reading projection centre coords. failed.\n";
		return ERROR;
		}

	// std. dev. projection centre coords.
	is >> _sa[0] >> _sa[1] >> _sa[2];
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of proj. centre coords. failed.\n";
		return ERROR;
		}

	// rotation parameters 
	size_t i = 0;
	for (i = 3; i < _size; i++)
		{
		is >> _a[i];
		if (!is)
			{
			cout << "ERROR: Reading rotation parameters failed.\n";
			return ERROR;
			}
		}

	// std. dev. of rotation parameters
	for (i = 3; i < _size; i++)
		{
		is >> _sa[i];
		if (!is)
			{
			cout << "ERROR: Reading std. dev. of rotation parameters failed.\n";
			return ERROR;
			}
		}

	return OK;
}
 

// write

void ExtOrientation::write(ostream& os) const
{
	double sc = AdjObject::lspars.scale_obj();

	Form sci(6); sci.scientific().width(15);
	Form fix(1); fix.fixed().width(15);

	// type
	os << key_str() << " ";

	// ID
	os << _id << " ";

	// exposure time
	fix.precision(4);
	os << fix(_time) << endl;

	size_t i = 0;

	// projection centre
	fix.precision(4);
	for (i = 0; i < 3; i++)
		os << " " << fix(_a[i] * sc);
	os << endl;

	for (i = 0; i < 3; i++)
		{
		if (is_rsdev(_sa[i]))
			os << " " << fix(_sa[i] * sc);
		else
			os << " " << sci(_sa[i]);
		}
	os << endl;

	// rotation parameters
	for (i = 3; i < _size; i++)
		os << " " << sci(_a[i]);
	os << endl;

	for (i = 3; i < _size; i++)
		os << " " << sci(_sa[i]);
	os << endl;
}


// print

void ExtOrientation::print(ostream& os) const
{
	double sc = AdjObject::lspars.scale_obj();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(4); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Type          :  " << type_str() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";

	size_t i = 0;

	fix.precision(3);
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(_sa[i])) << "   "
			<< fix(_a[i] * sc) << "   ";
		if (is_rsdev(_sa[i]))
			os << fix(_sa[i] * sc);
		else
			os << sci(_sa[i]);
		os << endl;
		}

	fix.precision(6);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(_sa[i])) << "   "
			<< fix(_a[i]) << "   ";
		if (is_rsdev(_sa[i]))
			os << fix(_sa[i]);
		else
			os << sci(_sa[i]);
		os << endl;
		}
	os << endl;

	vector<double> a;
	RotMatrix3<double> R;
	get_rot_params(a);
	calc_rot_matrix(R, a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// print changes of unknowns without header

void ExtOrientation::print_changes(ostream& os, const Vector<double>& dx) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(6); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Type       :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;
   
	os << "  Name       State"
		"       Approx.         Change       Adjusted\n\n";
	
	size_t i = 0;
	int k = _index;

	for (; i < 3; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i)) << "   "
			<< fix(dx(k)) << "   "
			<< fix(par(i) + dx(k)) << endl;
		}

	fix.precision(6);
	for (; i < _size; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i)) << "   "
			<< fix(dx(k)) << "   "
			<< fix((par(i) + dx(k))) << endl;
		}
   	os << endl;
}


// print unknown parameters

void ExtOrientation::print_unknowns(ostream& os) const 
{
	double sc = AdjObject::lspars.scale_obj();
		
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Type       :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";

	size_t i = 0;
	
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i)) << "  "
			<< fix(par(i)) << "  "
			<< fix(change(i)) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * sc);
		else
			os << sci(sdev(i));
		os << endl;
		}
	
	fix.precision(6);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i)) << "  "
			<< fix(par(i)) << "  "
			<< fix(change(i)) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i));
		else
			os << sci(sdev(i));
		os << endl;
		}
   	os << endl;

	vector<double> a;
	RotMatrix3<double> R;
	get_rot_params(a);
	calc_rot_matrix(R, a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// check on convergence

int ExtOrientation::check_convergence(const Vector<double>& dx) const
{
	if (lspars.chk_pcc())
		{
		int k = _index;
		for (size_t i = 0; i < 3; i++, k++)
			{
			if (fabs(dx(k)) > lspars.conv_pcc()) return i;
			}
		}

	if (lspars.chk_rot())
		{
		int k = _index + 3;
		for (size_t i = 3; i < _size; i++, k++)
			{
			if (fabs(dx(k)) > lspars.conv_rot()) return i;
			}
		}

	return -1;
}


// object-to-photo coords. transformation

bool ExtOrientation::object_to_photo(const DPoint3& obj_c,
DPoint3& pho_c) const
{
	projection(_per_centre, _rot_matrix, obj_c, pho_c);
	return true;
}


// spatial intersection

bool ExtOrientation::intersection(const ImageGeometry& g2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	// using method by Albertz/Kreiling
	//e = intersection3(this->_per_centre, this->_rot_matrix, pho_c1,
	//	eo2._per_centre, eo2._rot_matrix, pho_c2, obj_c);
	//return true;

	bool rtn = false;

	try {
		const ExtOrientation& eo2 = dynamic_cast<const ExtOrientation&>(g2);
		rtn = intersection(eo2, pho_c1, pho_c2, obj_c, e);
	}
	catch (bad_cast) {
		cerr << "ExtOrientation::intersection: "
			"Incompatible image geometries!\n";
		rtn = false;
	}

	return rtn;
}


// get partial derivatives

bool ExtOrientation::get_derivs(const DPoint3& pho_c, const DPoint3& obj_c,
DerivSub& deriv_oc, DerivSub& deriv_eo,
DerivSub& deriv_pc, DerivSub& deriv_ac, DerivSub& deriv_obs)
{
	// adjust sub-vectors
	deriv_oc.newsize(2,3);
	deriv_eo.newsize(2,_size);

	// The derivatives of the projection centre coordinates and
	// the object coordinates are always the same only the derivatives
	// of the rotation parameters are different!

	// object coords. reduced to the perspective centre
	_dx = obj_c.x - _per_centre.x;
	_dy = obj_c.y - _per_centre.y;
	_dz = obj_c.z - _per_centre.z;

	_n  = _rot_matrix(0,2) * _dx + _rot_matrix(1,2) * _dy
		+ _rot_matrix(2,2) * _dz;
	if (fabs(_n) == 0.)
		{
		cout << "ERROR: ExtOrientation::get_derivs: Zero denominator.\n";
		return false;
		}

	_zx = _rot_matrix(0,0) * _dx + _rot_matrix(1,0) * _dy
		+ _rot_matrix(2,0) * _dz;
	_zy = _rot_matrix(0,1) * _dx + _rot_matrix(1,1) * _dy
		+ _rot_matrix(2,1) * _dz;

	_cn = pho_c.z / _n;
	_cn2 = _cn / _n;

	// derivatives of the projection centre coords.

	// Xo-derivatives
	deriv_eo(0,0) = _cn2 * (_rot_matrix(0,2) * _zx - _rot_matrix(0,0) * _n);
	deriv_eo(1,0) = _cn2 * (_rot_matrix(0,2) * _zy - _rot_matrix(0,1) * _n);

	// Yo-derivatives
	deriv_eo(0,1) = _cn2 * (_rot_matrix(1,2) * _zx - _rot_matrix(1,0) * _n);
	deriv_eo(1,1) = _cn2 * (_rot_matrix(1,2) * _zy - _rot_matrix(1,1) * _n);

	// Zo-derivatives
	deriv_eo(0,2) = _cn2 * (_rot_matrix(2,2) * _zx - _rot_matrix(2,0) * _n);
	deriv_eo(1,2) = _cn2 * (_rot_matrix(2,2) * _zy - _rot_matrix(2,1) * _n);

	// get derivatives of position correction parameters
	if (_pos_cor != 0)
		_pos_cor->get_derivs(_time, deriv_eo, deriv_pc);

	// get derivatives of the rotation and attitude correction parameters
	// see comment of method!
	get_derivs_rot_acor(deriv_eo, deriv_ac);

	// The derivatives of the object coordinates are simply
	// the negative derivatives of the projection centre coordinates!

	// X-derivatives
	deriv_oc(0,0) = - deriv_eo(0,0);
	deriv_oc(1,0) = - deriv_eo(1,0);

	// Y-derivatives
	deriv_oc(0,1) = - deriv_eo(0,1);
	deriv_oc(1,1) = - deriv_eo(1,1);

	// Z-derivatives
	deriv_oc(0,2) = - deriv_eo(0,2);
	deriv_oc(1,2) = - deriv_eo(1,2);

	// derivatives of observations
	deriv_obs.newsize(2,2);
	deriv_obs.identity();

	return true;
}


// sum of absolute changes

void ExtOrientation::sum_changes_pcc(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index;

	for (size_t i = 0; i < 3; i++, k++)
		if (state(i) != eli) sum += fabs(dx(k)), num++;
}


// sum of absolute changes

void ExtOrientation::sum_changes_rot(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index + 3;

	for (size_t i = 3; i < _size; i++, k++)
		if (state(i) != eli) sum += fabs(dx(k)), num++;
}


//

void ExtOrientation::get_derivs_rot_acor(DerivSub& deriv_eo,
DerivSub& deriv_ac) const
{
	// get derivatives of the rotation parameters
	get_derivs_rot(deriv_eo);
}


// calculate EO, position and attitude correction included

void ExtOrientation::calc_orientation()
{
	// perspective centre coords.
	_per_centre.x = _a[0];
	_per_centre.y = _a[1];
	_per_centre.z = _a[2];

	// get position correction (optional)
	if (_pos_cor != 0)
		_per_centre += _pos_cor->calc_corr(_time);

	// rotation parameters
	for (size_t i = 3, j = 0; i < _size; i++, j++)
		_rot_params[j] = _a[i];

	// get attitude correction (optional)
	//if (_att_cor != 0)
	//	_rot_angles_cor += _att_cor->calc_corr(_time);
	
	// calculate rotation matrix
	calc_rot_matrix(_rot_matrix, _rot_params);
}


// spatial intersection

bool ExtOrientation::intersection(const ExtOrientation& eo2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	// using method by Albertz/Kreiling
	return DEP::intersection(this->_per_centre, this->_rot_matrix, pho_c1,
		eo2._per_centre, eo2._rot_matrix, pho_c2, obj_c, e);
}


// non-member functions ...

// project object point onto photo plane

void projection(const DPoint3& pc, const RotMatrix3<double>& R,
const DPoint3& obj_c, DPoint3& pho_c)
{
	// reduce coordinates
    double dx = obj_c.x - pc.x;
    double dy = obj_c.y - pc.y;
    double dz = obj_c.z - pc.z;
	//DPoint3 d = obj_c - pc;

	// rotate system
	//DPoint3 trf_c = R * d;
	//if (fabs(trf_c.z) == 0.) cerr << "DEP::projection: zero denominator.\n";

	// reciprocal scale factor
	//double reci_lam = pho_c.z / trf_c.z;

	// denominator
	double q = R(0,2) * dx + R(1,2) * dy + R(2,2) * dz;
	if (fabs(q) == 0.) cerr << "DEP::projection: zero denominator.\n";

	// nominators
	double m = R(0,0) * dx + R(1,0) * dy + R(2,0) * dz;
	double n = R(0,1) * dx + R(1,1) * dy + R(2,1) * dz;

	// photo coords.
	pho_c.x = pho_c.z * m / q;
	pho_c.y = pho_c.z * n / q;
	// scale into image plane
	//pho_c.x = trf_c.x * reci_lam;
	//pho_c.y = trf_c.y * reci_lam;
}


// Sarrus formular

static double sarrus(double a[3][3])
{
	double det = a[0][0] * a[1][1] * a[2][2];
	det += a[0][1] * a[1][2] * a[2][0];
	det += a[0][2] * a[1][0] * a[2][1];
	det -= a[0][2] * a[1][1] * a[2][0];
	det -= a[0][0] * a[1][2] * a[2][1];
	det -= a[0][1] * a[1][0] * a[2][2];
	return det;
}


// spacial intersection

bool intersection(
const DPoint3& pc1, const RotMatrix3<double>& rm1, const DPoint3& pho_c1,
const DPoint3& pc2, const RotMatrix3<double>& rm2, const DPoint3& pho_c2,
DPoint3& obj_c, double& e)
{
	//cout << "intersec3()\n";
	//cout << pc1 << endl;
	//cout << rm1 << endl;
	//cout << pho_c1 << endl;
	//cout << pc2 << endl;
	//cout << rm2 << endl;
	//cout << pho_c2 << endl;
	//cout << endl;

	// transform image vector into object space
	DPoint3 q1;
	q1.x = rm1(0,0) * pho_c1.x + rm1(0,1) * pho_c1.y + rm1(0,2) * pho_c1.z;
	q1.y = rm1(1,0) * pho_c1.x + rm1(1,1) * pho_c1.y + rm1(1,2) * pho_c1.z;
	q1.z = rm1(2,0) * pho_c1.x + rm1(2,1) * pho_c1.y + rm1(2,2) * pho_c1.z;
	q1 += pc1;

	DPoint3 q2;
	q2.x = rm2(0,0) * pho_c2.x + rm2(0,1) * pho_c2.y + rm2(0,2) * pho_c2.z;
	q2.y = rm2(1,0) * pho_c2.x + rm2(1,1) * pho_c2.y + rm2(1,2) * pho_c2.z;
	q2.z = rm2(2,0) * pho_c2.x + rm2(2,1) * pho_c2.y + rm2(2,2) * pho_c2.z;
	q2 += pc2;
	
	double dx = q1.x - pc1.x;
	double dy = q1.y - pc1.y;
	double dz = q1.z - pc1.z;
	double d1 = sqrt(dx * dx + dy * dy + dz * dz);

	double l1 = dx / d1;
	double m1 = dy / d1;
	double n1 = dz / d1;

	dx = q2.x - pc2.x;
	dy = q2.y - pc2.y;
	dz = q2.z - pc2.z;
	double d2 = sqrt(dx * dx + dy * dy + dz * dz);

	double l2 = dx / d2;
	double m2 = dy / d2;
	double n2 = dz / d2;

	double l = m1 * n2 - n1 * m2;
	double m = - (l1 * n2 - n1 * l2);
	double n = l1 * m2 - m1 * l2;

	double bx = pc2.x - pc1.x;
	double by = pc2.y - pc1.y;
	double bz = pc2.z - pc1.z;

	double D_lam[3][3];
	D_lam[0][0] = bx; D_lam[1][0] = l; D_lam[2][0] = l2;
	D_lam[0][1] = by; D_lam[1][1] = m; D_lam[2][1] = m2;
	D_lam[0][2] = bz; D_lam[1][2] = n; D_lam[2][2] = n2;

	double D_mue[3][3];
	D_mue[0][0] = bx; D_mue[1][0] = l; D_mue[2][0] = l1;
	D_mue[0][1] = by; D_mue[1][1] = m; D_mue[2][1] = m1;
	D_mue[0][2] = bz; D_mue[1][2] = n; D_mue[2][2] = n1;

	double D[3][3];
	D[0][0] = l1; D[1][0] = l2; D[2][0] = l;
	D[0][1] = m1; D[1][1] = m2; D[2][1] = m;
	D[0][2] = n1; D[1][2] = n2; D[2][2] = n;

	double det_lam = sarrus(D_lam);
	double det_mue = sarrus(D_mue);
	double det = sarrus(D);
	if (det == 0.)
		{
        cerr << "DEP::intersection: rays are parallel.\n";
		return false;
        }

	double lam = - det_lam / det;
	double mue = - det_mue / det;

	double x1 = pc1.x + lam * l1;
	double y1 = pc1.y + lam * m1;
	double z1 = pc1.z + lam * n1;

	double x2 = pc2.x + mue * l2;
	double y2 = pc2.y + mue * m2;
	double z2 = pc2.z + mue * n2;

	obj_c.x = (x1 + x2) / 2.;
	obj_c.y = (y1 + y2) / 2.;
	obj_c.z = (z1 + z2) / 2.;

	dx = x1 - x2;
	dy = y1 - y2;
	dz = z1 - z2;

	e = sqrt(dx * dx + dy * dy + dz * dz);

	return true;
}


// project image point onto Z-plane (ray-z-plane intersection)
/*
void intersection(const DPoint3& pc, const RotMatrix3<double>& R,
DPoint3& obj_c, const DPoint3& pho_c)
{
	// transformed photo coords.
	DPoint3 trf_c = R * pho_c;

	// scale factor
	double lam = (obj_c.z - pc.z) / trf_c.z;

	// get X,Y-object coords.
	obj_c.x = pc.x + lam * trf_c.x;
	obj_c.y = pc.y + lam * trf_c.y;
}
*/

} // namespace DEP
