/* $Id: extori.h,v 1.9.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_EXTORI_INCLUDED
#define DEP_EXTORI_INCLUDED

#include "imggeo.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class PosCorrection;
class DerivSub;

/** Exterior orientation class.
 *  Assuming a central perspective projection, the mapping from a 3D object
 *  coordinate system onto the 2D image coordinate system is accomplished
 *  via the collinearity equations.
 *  The spatial position and attitude of the camera
 *  station is defined by the Exterior Orientation (EO). 
 *  The spatial position, also called perspective projection centre, is
 *  defined by its coordinates Xo, Yo and Zo. The attitude is defined by
 *  the rotation matrix. In this base class no assumtion is made on 
 *  the parametrisation of the rotation matrix.
 *  Different parametrisations exist. Usual parametrisations are:
 *  three rotation angles (euler angles), an quaternion, Rodriguez matrix, 
 *  or direction cosines (the rotation matrix itself).
 *  This spezialisation is done by the drived classes,
 *  e.g. EOAnglesOPKRot, EOQuaternion.
 *
 *  As an option the position correction to model systematic errors
 *  in the perspective centre can be used.
 *  The position correction is time-dependent hence
 *  the correct (GPS-)time has to be read.
 *
 *  The attitude correction is not simply usable here
 *  because for more than 3 rotation parameters constraints
 *  between the correction polynoms exist.
 *
 *  @author Dirk Stallmann
 */
class ExtOrientation : public ImageGeometry
{
	friend double intersection(const ExtOrientation& eo1,
		const DPoint3& pho_c1, const ExtOrientation& eo2,
		const DPoint3& pho_c2, DPoint3& obj_c);

public:
	/** constructor
	 *  @param sz number of rotation parameters (3 <= sz <= 9)
	 */
	ExtOrientation(size_t sz=3);

	/**@name read/write/print EO parameters */
	//@{
		/// read
		virtual int read(istream& is);

		/// write
		virtual void write(ostream& os) const;

		/// print
		virtual void print(ostream& os) const;
	//@}

	/**@name overridden methods of base class AdjObject
	 */
	//@{
		/// print changes of unknowns
		virtual void print_changes(ostream& os, const Vector<double>& dx) const;

		/// print unknown parameters
		virtual void print_unknowns(ostream& os) const;

		/// check on convergence 
		int check_convergence(const Vector<double>& dx) const;
	//@}

	/**@name resolved pure virtual methods of class Image\-Geometry
	 */
	//@{
		/// object-to-photo coords. transformation
		bool object_to_photo(const DPoint3& obj_c,
			DPoint3& pho_c) const;

		/// spatial intersection
		bool intersection(const ImageGeometry& g2,
			const DPoint3& pho_c1, const DPoint3& pho_c2,
			DPoint3& obj_c, double& e) const;

		/// get partial derivatives
		bool get_derivs(const DPoint3& pho_c, const DPoint3& obj_c,
			DerivSub& deriv_oc, DerivSub& deriv_eo,
			DerivSub& deriv_pc, DerivSub& deriv_ac, DerivSub& deriv_obs);
	//@}

	/// set and get perspective centre coords.
	//@{
		/// set perspective centre coords.
		void set_pcc(const DPoint3& pcc) {
			_a[0] = pcc.x; _a[1] = pcc.y; _a[2] = pcc.z;
		}

		/// get perspective centre coords.
		void get_pcc(DPoint3& pcc) const {
			pcc.x = _a[0]; pcc.y = _a[1]; pcc.z = _a[2];
		}

		/// set std. devs.
		void set_sdev_pcc(const DPoint3& sdev) {
			_sa[0] = sdev.x; _sa[1] = sdev.y; _sa[2] = sdev.z;
		}
	//@}

	/// set and get rotation parameters
	//@{
		// get number of rotation parameters
		//size_t num_rot_params() const { return _size - 3; }

		/// set rotation parameters
		void set_rot_params(const vector<double>& a) {
			for (size_t i = 3, j = 0; i < _size; i++, j++)
				_a[i] = a[j];
		}

		/// get rotation parameters
		void get_rot_params(vector<double>& a) const {
			a.resize(_size - 3);
			for (size_t i = 3, j = 0; i < _size; i++, j++)
				a[j] = _a[i];
		}

		/// set std. devs.
		void set_sdev_rot(const vector<double>& sa) {
			for (size_t i = 3, j = 0; i < _size; i++, j++)
				_sa[i] = sa[j];
		}
	//@}

	/** Sum of absolute changes
	 *
	 *  Overwritten base class function. Simply do nothing. 
	 *  The summation is done separatly for the perspective centre coords.
	 *  and the rotation parameters.
	 */
	//@{
		/// sum of absolute changes.
		void sum_changes(const Vector<double>& dx,
			double& sum, size_t& num) const {}

		/// sum of absolute changes of perspective centre coords.
		void sum_changes_pcc(const Vector<double>& dx,
			double& sum, size_t& num) const;

		/// sum of absolute changes of rotation parameters
		void sum_changes_rot(const Vector<double>& dx,
			double& sum, size_t& num) const;
	//@}

	/** get partial derivatives of rotation parameters and
	 *  of the attitude correction parameters.
	 *  Only for parametrisations using 3 rotation parameters.
	 *  For parametrisations using more 3 rotation parameters
	 *  the 2nd argument is unused or when the attitude correction
	 *  is not defined!
	 */
	virtual void get_derivs_rot_acor(DerivSub& deriv_eo,
		DerivSub& deriv_ac) const;

	/** Pure virtual methods for rotation parameters
	 *
	 *  Methods to derive rotation parameters,
	 *  and rotation matrix and derivaties thereof.
	 *  Because no assumtion is made of the
	 *  type of rotation parameters.
	 */
	//@{
		/// calculate rotation parameters from rotation matrix
		virtual void calc_rot_params(vector<double>& a,
			const RotMatrix3<double>& R) const = 0;

		/// calculate rotation matrix from rotation parameters
		virtual void calc_rot_matrix(RotMatrix3<double>& R,
			const vector<double>& a) const = 0;

		/** get partial derivatives of rotation parameters
		 *  The type of parametrisation is undefined here!
		 */
		virtual void get_derivs_rot(DerivSub& deriv_eo) const = 0;
	//@}

	/// spatial intersection
	bool intersection(const ExtOrientation& eo2,
		const DPoint3& pho_c1, const DPoint3& pho_c2,
		DPoint3& obj_c, double& e) const;

	/// attach position correction
	void attach(PosCorrection* p) { _pos_cor = p; }

	/// calculate EO, position correction included
	virtual void calc_orientation();

	/// check link to position correction
	bool has_pos_cor() const { return _pos_cor != 0; }

protected:
	/// perspective centre coords, position correction included
	DPoint3		_per_centre;

	/// rotation parameters
	vector<double>	_rot_params;

	/// rotation matrix
	RotMatrix3<double>	_rot_matrix;

	/// position correction (optional)
	PosCorrection*		_pos_cor;

	/// coordinate difference in X direction
	double _dx;

	/// coordinate difference in Y direction
	double _dy;

	/// coordinate difference in Z direction
	double _dz;

	/// denominator
	double _n;

	/// quotient z/n 
	double _cn;

	/// quotient z/n squared
	double _cn2;

	/// x-nominator
	double _zx;

	/// y-nominator
	double _zy;
};


// non-member functions ...

/** Project object point onto photo plane.
 *  Basic collinearity equation to get photo from object coordinates.
 * 
 *   @param pc (IN) perspective centre coords.
 *   @param rm (IN) rotation matrix between object and photo coordinate system
 *   @param obj_c (IN) object point coords.
 *   @param pho_c photo coords.: pho_c.x, pho_c.y (OUT) x,y photo coords.,
 *        pho_c.z (IN) f focal length
 */
void projection(const DPoint3& pc, const RotMatrix3<double>& rm,
	const DPoint3& obj_c, DPoint3& pho_c);

/** Spacial intersection.
 *  Intersection by using the minimal distance e between two rays.
 *  Method published by Albertz/Kreiling.
 *
 *   @param pc1 (IN) perspective centre coords. of image 1
 *   @param rm1 (IN) rotation matrix of image 1
 *   @param pho_c1 (IN) photo coords. in image 1
 *   @param rm2 (IN) rotation matrix of image 2
 *   @param pc2 (IN) perspective centre coords. of image 2
 *   @param pho_c2 (IN) photo coords. in image 1
 *   @param obj_c (OUT) object point coords.
 *   @param e (OUT) minimal distance e
 */
bool intersection(
	const DPoint3& pc1, const RotMatrix3<double>& rm1, const DPoint3& pho_c1,
	const DPoint3& pc2, const RotMatrix3<double>& rm2, const DPoint3& pho_c2,
	DPoint3& obj_c, double& e);

/// project image point onto Z-plane (ray-z-plane intersection)
//void intersection(const DPoint3& pc, const RotMatrix3<double>& rm,
//	DPoint3& obj_c, const DPoint3& pho_c);

} // namespace DEP

#endif
