/* $Id: fiodep.h,v 1.6.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_FIODEP_INCLUDED
#define DEP_FIODEP_INCLUDED

#include "depdef.h"

namespace DEP
{
class LSParams;
class Camera;
class Image;
class ObjectPoint;
class ImageGeometry;
class APSet;
class PosCorrection;
class AttCorrection;

/** DEP file I/O.
 *  Some global functions for input and output of DEP-files.
 *	@name DEP file I/O
 *  @author Dirk Stallmann
 */
//@{
	/// File I/O types
	enum fio_type {
		/// unknown
		UNKNOWN,

		/// parameters/options
		PARAMS,

		/// LS parameters/options
		LS_PARAMS,

		/// photographic cameras
		CAMERA,
			/// frame camera
			CAMERA_FRAME,
			/// CCD frame camera
			CAMERA_CCD,
			/// CCD frame camera with top-right orientation
			CAMERA_CCD_TR,
			/// pushbroom camera
			CAMERA_PUSHBROOM,

		/// image types
		IMAGE,
			/// frame image
			IMAGE_FRAME,
			/// scanner image
			IMAGE_SCANNER,

		/// image geometry/exterior orientation
		IMAGE_GEOMETRY,
		EXT_ORI,
			EXT_ORI_OPK_ROT,
			EXT_ORI_POK_ROT,
			EXT_ORI_OPK_FIX,
			EXT_ORI_AUSTRALIS,
			EXT_ORI_QUATERNION,
			DLT,
			EXT_ORI_KRATKY,
			PMF_LINEAR,

		/// image coords.
		IMAGE_COORDS,

		/// object coords.
		OBJECT_COORDS,

		/// control support
		CONTROL_SUPP,

		/// additional parameter set
		APSET,
			/// implemented in GAP
			APSET_GAP,
			/// implemented in Australis
			APSET_AUSTRALIS,
			/// orthogonal polynom by Ebner
			APSET_P12,
			/// orthogonal polynom by Gruen
			APSET_P44,
			/// modelling IO
			APSET_INNER_OR,
			/// modelling radial distortion
			APSET_RADIAL,
			/// modelling decentering distortion
			APSET_DECENTERING,
			/// modelling in-plane distortion
			APSET_INPLANE,
			/// modelling out-of-plane distortion
			APSET_OUTOFPLANE,

		/// position correction
		POS_COR,

		/// attitude correction
		ATT_COR,

		/// PAT-B parameters/options
		PATB_PARAMS,

		/// PAT-B photographs
		PATB_PHOTOGRAPHS,

		/// PAT-B control points
		PATB_CONTROL_POINTS,

		/// PAT-B orientation parameters
		PATB_ORIENT_PARAMS,

		/// object coordinate transformation
		OC_TRANSFORM,

		/// stop-dep
		STOP_DEP
	};

	// type of rotation
	enum {
		ROT_MATRIX,		// rotation matrix
		ROT_ANGLES 		// angle of rotation
	};

	/// test keyword
	fio_type fio_test(const char* keyword);

	/// read cameras
	int read_cameras(istream& is, vector<Camera*>& cams);

	// read images
	int read_images(istream& is, const vector<Camera*>& cams,
		vector<Image*>& images);

	/// read exterior orientations
	int read_ext_orientation(istream& is, vector<ImageGeometry*>& eos);

	/// read exterior orientations have individual type and std. devs.
	int read_ext_orientation_indiv(istream& is, vector<ImageGeometry*>& eos);

	/// read stations have common type and std. devs.
	int read_ext_orientation_common(istream& is, vector<ImageGeometry*>& eos);

	/// read image coords.
	int read_image_coords(istream& is, vector<Image*>& images);

	/// read object point coords.
	int read_object_coords(istream& is, vector<ObjectPoint*>& obpts);

	/// read control support definition
	int read_ctrl_supp(istream& is, vector<ObjectPoint*>& obpts);

	// read AP sets
	int read_ap_sets(istream& is, vector<APSet*>& aps);

	// read position corrections
	int read_pos_correction(istream& is, vector<PosCorrection*>& pcors);

	// read attitude corrections
	int read_att_correction(istream& is, vector<AttCorrection*>& acors);
//@}

} // namespace DEP

#endif
