/* $Id: image.cpp,v 1.10.4.5 2007/05/20 13:57:24 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "depvec.h"
#include "image.h"
#include "lspars.h"
#include "drvsub.h"
#include "lconeqn.h"
#include "lobseqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "camera.h"
#include "imggeo.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"

namespace DEP
{

// constructor

Image::Image() :
	_image_id(),
	_station_id(),
	_eo(0),
	_camera_id(),
	_camera(0),
	_inuse(false)
{
	//empty
}


// read image data

int Image::read(istream& is, const vector<Camera*>& cams)
{
	// read image, station and camera ID
	is >> _image_id >> _station_id >> _camera_id;
	if (!is) return ERROR;

	// attach camera
	_camera = find(cams, _camera_id);
	if (_camera != 0)
		_camera->inuse(true);
	else
		{
		cout << "ERROR: Attach camera " << _camera_id
			<< " for image " << _image_id << " not failed.\n";
		return ERROR;
		}

	return OK;
}


// print image data

void Image::print(ostream& os, bool ic_out) const
{
	os << "Image ID           :  " << _image_id << endl;
	os << "Station ID         :  " << _station_id << endl;
	os << "Camera ID          :  " << _camera_id << endl;
	os << endl;

	// print image coords.
	if (ic_out)
		print_image_coords(os);
}


// get image type 

string Image::type_str() const
{
	ostringstream oss;
	oss << "frame";
	return oss.str();
}


// read image coords.
// assume photo coords. in meter, use image scale for conversion!

int Image::read_image_coords(istream& is)
{
	assert(_camera != 0);

	return _camera->read_image_coords(is, *this);
}


// write image coords.

void Image::write_image_coords(ostream& os) const
{
	assert(_camera != 0);

	_camera->write_image_coords(os, *this);
}


// print image coords.

void Image::print_image_coords(ostream& os) const
{
	assert(_camera !=0);

	_camera->print_image_coords(os, *this);
}


// add (append or replace) image point

bool Image::add_image_point(ImagePoint* ip)
{
	assert(ip != 0);
	bool rtn = false;

	int i = find_image_point(ip->_id);
	if (i < 0)
		{
		_image_points.push_back(ip);
		rtn = true;
		}
	else
		{
		ImagePoint* tmp = _image_points[i];
		_image_points[i] = ip; 
		delete tmp;
		cout << "ATTENTION: Image " << _image_id
			<< " point " << ip->_id << " was replaced.\n";
		}

    return rtn;
}


// count number of image points

int Image::count_image_points() const
{
	int n = 0;
	
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0) n++;
		}

	return n;
}


// test image point

bool Image::image_point_exist(const string& id) const
{
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id) return true;
		}

	return false;
}


// find image point by id, return index

int Image::find_image_point(const string& id) const
{
	int i = 0;

	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id) return i;
		i++;
		}

	return -1;
}


// get image point by id, return pointer

ImagePoint* Image::get_image_point(const string& id) const
{
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		if (*ip != 0 && (*ip)->get_id() == id) return *ip;

	return 0;
}


// remove image point
// erase by set element to 0!!!

bool Image::remove_image_point(const string& id)
{
	vector<ImagePoint*>::iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id)
			{
			delete *ip;
			*ip = 0;
			return true;
			}
		}

	return false;
}


// attach EO

bool Image::attach(const vector<ImageGeometry*>& eos)
{
	_eo = find(eos, _station_id);
	if (_eo != 0)
		_eo->inuse(true);
	else
		return false;
	
	return true;
}


// attach camera

bool Image::attach(const vector<Camera*>& cams)
{
	_camera = find(cams, _camera_id);
	if (_camera != 0)
		_camera->inuse(true);
	else
		return false;

	return true;
}


// object-to-image coords. transformation

int Image::object_to_image(const DPoint3& obj_c, DPoint2& img_c) const
{
	assert(_camera != 0);
	assert(_eo != 0);

	// set photo coords.
	// The z-component of the photo coordinates is initialized with z = -c!
	// The focal length is required for the central projection
	// (collinearity equations).
	// Other image geometries didn't need this information,
	// e.g. DLT. But a (pseudo) camera is still required
	// to define the external image coordinate system.
	DPoint3 pho_c(0., 0., -_camera->ck());

	// object-to-photo coords. transformation including position
	// and attitude correction
	_eo->object_to_photo(obj_c, pho_c);

	// photo-to-image coords. transformation including APs
	_camera->photo_to_image(pho_c, img_c);

	//cout << pho_c << endl;
	//cout << img_c << endl;

	return OK;
}


// image-to-object coords. transformation

int Image::image_to_object(const DPoint2& img_c, DPoint3& obj_c) const
{
	cout << "ERROR: image_to_object() not implemented yet!" << endl;
	assert(false);

	return ERROR;
}


// calculate approximate image geometry

bool Image::calc_approx_image_geometry(const vector<ObjectPoint*>& obpts,
ostream* os)
{
	assert(_eo != 0);
	// use image geometry dependend function!
	//if (_eo == 0)
	//	{
	//	cout << "ERROR: image geometry missing.\n";
	//	return false;
	//	}

	return _eo->calc_approx(obpts, os);
}

/*
// add contribution to normal equations for all image point observations
// using the Gauss-Markov model

int Image::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_eo != 0);
	assert(_camera != 0);

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_eo;			// EO, variable size
	vector<DerivSub> deriv_ap(_camera->num_ap_sets()); // APs, variable size
	DerivSub	deriv_pc;			// position corr., variable size
	DerivSub	deriv_ac;			// attitude corr., variable size

	// set index for EO (valid for the whole image)
	deriv_eo.index(_eo->get_index());

	// Always 2 condition equations for a single imaging ray.
	LinObsEqn obs_eqn(2);

	// design matrix A
	obs_eqn.add(&deriv_oc);	// always
	obs_eqn.add(&deriv_eo);	// always
	obs_eqn.add(deriv_ap);	// optional
	if (_eo->has_pos_cor()) obs_eqn.add(&deriv_pc);	//optional
	if (_eo->has_att_cor()) obs_eqn.add(&deriv_ac);	//optional

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _id << ":" << (*ip)->get_id() << ":\n";

		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert external to image std. dev.
		DPoint2 sdev;
		_camera->sdev_external_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// weights
		double px = AdjObject::lspars.sigma0() / sdev.x; px *= px;
		double py = AdjObject::lspars.sigma0() / sdev.y; py *= py;
		obs_eqn.set_p(0, px);
		obs_eqn.set_p(1, py);
		//cout << px << " " << py << endl;

		// convert external into internal image coords.
		DPoint2 img_c;
		_camera->external_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// convert image into photo coords.
		DPoint3 pho_c;
		_camera->image_to_photo(img_c, pho_c);
		//cout << "pho: " << pho_c << endl;
	
		// partial derivatives of
		// object point coords. and exterior orientation parameters,
		// position and attitude correction
		_eo->get_derivs(pho_c, op->crds, deriv_oc, deriv_eo,
			deriv_pc, deriv_ac);

		// partial derivatives of APs
		_camera->get_derivs(pho_c, deriv_ap);

		// image coords. derived from the unknown parameters
		DPoint img_c0;
		object_to_image(op->crds, img_c0);
		//cout << "img0: " << img_c0 << endl;

		// calculate shorten observations
		double lx = img_c.x - img_c0.x;
		double ly = img_c.y - img_c0.y;
		obs_eqn.set_l(0, lx);
		obs_eqn.set_l(1, ly);

		// add contributions to normal equation
		if (obs_eqn.add_contr(atpa, atpl, ltpl) != OK)
			return ERROR;

		num_obs += 2;
		}
	
	return OK;
}
*/

// add contribution to normal equations for all image point observations
// using the Gauss-Helmert model

int Image::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_eo != 0);
	assert(_camera != 0);

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_eo;			// EO, variable size
	vector<DerivSub> deriv_ap(_camera->num_ap_sets()); // APs, variable size
	DerivSub	deriv_pc;			// position corr., variable size
	DerivSub	deriv_ac;			// attitude corr., variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// set index for EO (valid for the whole image)
	deriv_eo.index(_eo->get_index());

	// Always 2 condition equations for a single imaging ray.
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(&deriv_eo);	// always
	con_eqn.add_a(deriv_ap);	// optional
	if (_eo->has_pos_cor()) con_eqn.add_a(&deriv_pc);	//optional
	if (_eo->has_att_cor()) con_eqn.add_a(&deriv_ac);	//optional

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _id << ":" << (*ip)->get_id() << ":\n";

		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert external to image std. dev.
		DPoint2 sdev;
		_camera->sdev_external_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);
		//cout << "Qll: " << Qll << endl;

		// convert external into internal image coords.
		DPoint2 img_c;
		_camera->external_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// convert image into photo coords.
		DPoint3 pho_c;
		_camera->image_to_photo(img_c, pho_c);
		//cout << "pho: " << pho_c << endl;
	
		// partial derivatives of
		// object point coords. and exterior orientation parameters,
		// position and attitude correction
		_eo->get_derivs(pho_c, op->crds, deriv_oc, deriv_eo,
			deriv_pc, deriv_ac, deriv_obs);

		// partial derivatives of APs
		// derivatives are computed with
		switch (AdjObject::lspars.ap_derivs())
			{
			case LSParams::IMAGE_COORDS:
				// image coords. without AP correction!
				_camera->image_to_photo(img_c, pho_c, false);
				break;
				
			case LSParams::COLLINEAR_EQUATION:
				// image coords. derived from collinear equation
				_eo->object_to_photo(op->crds, pho_c);
				break;
			
			case LSParams::IMAGE_COORDS_PLUS_APS:
				// image coords. plus AP correction
				// already calculated, see above
				break;
			
			default:
				assert(false);
			}
		_camera->get_derivs(pho_c, deriv_ap);

		// image coords. derived from the unknown parameters
		DPoint img_c0;
		object_to_image(op->crds, img_c0);
		//cout << "img0: " << img_c0 << endl;

		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> w(2);
		w(0) = img_c.x - img_c0.x;
		w(1) = img_c.y - img_c0.y;
		con_eqn.set_w(w);
		//cout << "w: " << w << endl;

		// design matrix B
		con_eqn.set_b(deriv_obs);
		//cout << "B: " << deriv_obs << endl;

		//con_eqn.print(cout); cout << endl;

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return ERROR;

		num_obs += 2;
		}
	
	return OK;
}


// print residual image coords.

void Image::print_residuals(const vector<ObjectPoint*>& obj_pnts,
DPoint2& sum_v, DPoint2& sum_vv, DPoint2& sum_pvv, int& num,
ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix10(2); fix10.fixed().width(10);

	DPoint2 img_c0;
	DPoint2 v, vv;
	DPoint2 p;
	double v_abs = 0.;

	// initialize accumulator
	sum_v.x = sum_v.y = 0.;
	sum_vv.x = sum_vv.y = 0;
	num = 0;
	
    os << "Image ID : " << _image_id << endl << endl;
	os << "  Point ID       vx/um      vy/um     |v|/um\n\n";

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		object_to_image(op->crds, img_c0);

		// convert external into image coords.
		DPoint2 img_c;
		_camera->external_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;
		v_abs = abs(v);
		vv = v * v;

		// convert external into image std. dev.
		DPoint2 sdev;
		_camera->sdev_external_to_image((*ip)->sdev, sdev);

		// weights
		p.x = AdjObject::lspars.sigma0() / sdev.x;
		p.y = AdjObject::lspars.sigma0() / sdev.y;
		p *= p;
			
		sum_v += v;
		sum_vv += vv;
		sum_pvv += vv * p;
		num++;

		// print residual
		if (AdjObject::lspars.is_small(v.x) || AdjObject::lspars.is_small(v.y))
			{
   			os << "  " << str8((*ip)->get_id()) << "  "
				<< fix10(v.x * 1e6) << " " << fix10(v.y * 1e6) << " "
				<< fix10(v_abs * 1e6);

			if (op->is_ctrl())
				{
				os << "   control point for";
				if (op->is_csx()) os << " X";
				if (op->is_csy()) os << " Y";
				if (op->is_csz()) os << " Z";
				}
			os << endl;
			}
		}
	os << endl;

	os << "Number of points       :   " << num << endl;
	os << "RMS of residuals       :   "
		<< fix10(sqrt(sum_vv.x / num) * 1e6) << "   "
		<< fix10(sqrt(sum_vv.y / num) * 1e6) << endl;
	os << "Average of residuals   :   "
		<< fix10(sum_v.x * 1e6 / num) << "   "
		<< fix10(sum_v.y * 1e6 / num) << endl;
	os << endl;
}


// write residuals

void Image::write_residuals(const vector<ObjectPoint*>& obj_pnts,
ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix10(3); fix10.fixed().width(10);
	Form fix12(3); fix12.fixed().width(12);

	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c0, img_c, v;

    os << _image_id << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		object_to_image(op->crds, img_c0);

		// convert external into image coords.
		_camera->external_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;

		os << "  " << str8((*ip)->get_id()) << "  "
			<< fix12(img_c.x * sc) << " " << fix12(img_c.y * sc) << " "
			<< fix10(v.x * sc) << " " << fix10(v.y * sc) << endl;
		}
	os << "stop-dep\n";
}


// write adjusted image coords.
/*
void Image::write_adjust_image_coords(ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix3(3); fix3.fixed().width(10);

	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c0, img_c;

    os << _image_id << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		object_to_image(op->crds, img_c0);

		// convert external into image coords.
		_camera->external_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;

   		os << "  " << str8((*ip)->get_id()) << "  "
			<< fix3(v.x * sc) << " " << fix3(v.y * sc) << endl;
		}
	os << "stop-dep\n";
}
*/

// write photo coords.
/*
void Image::write_photo_coords(ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix3(3); fix3.fixed().width(14);

	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c;
	DPoint3 pho_c;

	os << _image_id << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		_camera->external_to_image((*ip)->crds, img_c);
		_camera->image_to_photo(img_c, pho_c);

		//_camera->sdev_external_to_image((*ip)->sdev, sdev);

		os << str8((*ip)->get_id())
			<< " " << fix3(pho_c.x * sc) << " " << fix3(pho_c.y * sc)
			<< " " << fix3(pho_c.z * sc) << endl;
		}
	os << "stop-dep\n";
}
*/


// non-member functions ...

// spatial intersection

bool intersection(const Image& image1, const DPoint2& ext_c1,
const Image& image2, const DPoint2& ext_c2, DPoint3& obj_c, double& e)
{
	assert(image1._camera != 0 && image1._eo != 0);
	assert(image2._camera != 0 && image2._eo != 0);

	//if (typeid(*(image1._eo)) != typeid(*(image1._eo)))
	//	{
	//	cout << "ERROR: intersection: unequal image geometries." << endl;
	//	return false;
	//	}

	// external-to-image coords. transformation
	DPoint2 img_c1, img_c2;
	image1._camera->external_to_image(ext_c1, img_c1);
	image2._camera->external_to_image(ext_c2, img_c2);

	// image-to-photo coords. transformation
	DPoint3 pho_c1, pho_c2;
	image1._camera->image_to_photo(img_c1, pho_c1);
	image2._camera->image_to_photo(img_c2, pho_c2);

	image1._eo->intersection(*(image2._eo), pho_c1, pho_c2, obj_c, e);
	
	return true;
}

} // namespace DEP
