/* $Id: image.h,v 1.7.4.4 2007/03/07 17:32:08 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_IMAGE_INCLUDED
#define DEP_IMAGE_INCLUDED

#include "depdef.h"

using namespace GenLib2;

namespace DEP
{
class LSParams;
class DerivSub;
class ImagePoint;
class ObjectPoint;
class Camera;
class ImageGeometry;
class APSet;
class PosCorrection;
class AttCorrection;
class Block;

/** Image class.
 *  
 *  Ties everything together which belongs to the image:
 *  image points, image geometry and imaging camera. 
 *  @author Dirk Stallmann
 */
class Image
{
	friend bool intersection(const Image& image1, const DPoint2& img_c1,
		const Image& image2, const DPoint2& img_c2, DPoint3& obj_c, double& e);

public:
	/// constructor
	Image();

	/// virtual destructor
	virtual ~Image() {}

	/// read
	int read(istream& is, const vector<Camera*>& cams);

	/// print
	void print(ostream& os, bool ic_out) const;

	/// get image type 
	virtual string type_str() const;

	/**@name set and get image ID */
	//@{
		/// set image ID
		void get_id(const string& s) { _image_id = s; }

		/// get image ID
		const char* get_id() const { return _image_id.c_str(); }
	//@}
	
	/**@name inuse flag */
	//@{
		/// set inuse flag
		void inuse(bool flag) { _inuse = flag; }

		/// set inuse flag
		void set_inuse() { _inuse = true; }

		/// test inuse flag
		bool is_unused() const { return _inuse == false; }
	//@}

	/**@name 2-D image points.
	 *  external image coords.
	 */
	//@{
		/// image coordinates
		vector<ImagePoint*> _image_points;

		/// read image coords.
		int read_image_coords(istream& is);

		// write image coords.
		void write_image_coords(ostream& is) const;
	
		/// print image coords.
		void print_image_coords(ostream& os) const;

		/// append or replace image point
		bool add_image_point(ImagePoint* ip);

		/// count number of image points
		int count_image_points() const;

		/// check if image point exist
		bool image_point_exist(const string& id) const;

		/// find image point by id
		int find_image_point(const string& id) const;

		/// get image point by id
		ImagePoint* get_image_point(const string& id) const;

		/// remove image point
		bool remove_image_point(const string& id);
	//@}

	/**@name attach objects */
	//@{
		/// attach EO
		bool attach(const vector<ImageGeometry*>& eos);

		/// attach camera
		bool attach(const vector<Camera*>& cams);
	//@}

	/**@name coordinate transformations.
	 * CAVEATS: calculate exterior orientation before.
	 * Exterior orientation parameters are influenced by position and 
	 * attitude correction.
	 */
	//@{
		/** calculate exterior orientation:
		 * perspective centre coords. including position correction
		 * and rotation matrix including attitude correction
		 * do this before applying coordinate transformations!!!
		 */
		//void calc_orientation();

		/// object-to-image coords. transformation (collineare equation)
		virtual int object_to_image(const DPoint3& obj_c, DPoint2& img_c) const;

		/// image-to-object coords. transformation (ray-Z-plane intersection)
		virtual int image_to_object(const DPoint2& img_c, DPoint3& obj_c) const;
	//@}

	/// calculate approximate image geometry
	virtual bool calc_approx_image_geometry(const vector<ObjectPoint*>& obpts,
		ostream* os);
	
	/// add constribution to normal equations for all image point observations
	virtual int normal_eqns(const vector<ObjectPoint*>& obpts,
		Matrix<double>& atpa, Vector<double>& atpl, double& ltpl,
		int& num_obs) const;
	
	/// print residuals
	void print_residuals(const vector<ObjectPoint*>& obj_pnts,
		DPoint2& sum_v, DPoint2& sum_vv,
		DPoint2& sum_pvv, int& num, ostream& os) const;

	/// write residuals (in mue-meter)
	void write_residuals(const vector<ObjectPoint*>& obj_pnts,
		ostream& os) const;

	// write adjusted image coords.
	//void write_adjust_image_coords(ostream& os) const;

	// write photo coords. (in mue-meter)
	//void write_photo_coords(ostream& os) const;

protected:
	/// image ID
	string		_image_id;

	/// station ID
	string		_station_id;
				
	/// exterior orientation (always required)
	ImageGeometry*	_eo;

	/// camera ID
	string		_camera_id;

	/// imaging camera (always required)
	Camera*		_camera;
	
	/// inuse flag
	bool		_inuse;
};


// non-member functions ...
	
/// spatial intersection
bool intersection(const Image& image1, const DPoint2& ext_c1,
	const Image& image2, const DPoint2& ext_c2, DPoint3& obj_c, double& e);

} // namespace DEP

#endif
