/* $Id: imggeo.cpp,v 1.3.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "imggeo.h"

namespace DEP
{

// allocate memory for parameters and std. dev.

void ImageGeometry::allocate(size_t sz)
{
	_size = sz;
	_a = new double[_size];
	assert(_a != 0);
	_sa = new double[_size];
	assert(_sa != 0);
}


// free memory for parameters and std. dev.

void ImageGeometry::free()
{
	delete[] _a;
	delete[] _sa;
}


// new size for parameters and std. dev.

void ImageGeometry::newsize(size_t sz)
{
	if (_size == sz) return;
	free();
	allocate(sz);
}


// constructor

ImageGeometry::ImageGeometry(size_t sz) : AdjObject(), _time(0.)
{
	allocate(sz);
}


// destructor

ImageGeometry::~ImageGeometry()
{
	free();
}


// read
 
int ImageGeometry::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading station ID failed.\n";
		return ERROR;
		}

	if (!(is >> _time))
		{
		cout << "ERROR: Cannot read exposure time.\n";
		return ERROR;
		}

	// read values and std. devs.
	for (size_t i = 0; i < _size; i++)
		{
		is >> _a[i] >>  _sa[i];
		if (!is)
			{
			cout << "ERROR: Reading values and std. devs. failed.\n";
			return ERROR;
			}
		}

	return OK;
}
 

// write

void ImageGeometry::write(ostream& os) const
{
	Form fix(6); fix.fixed().width(12);
	Form sci(15); sci.scientific().width(23);

	os << key_str() << "   ";
	os << _id << "   ";
	os << fix(_time) << endl;
	
	for (size_t i = 0; i < _size; i++)
		os << sci(_a[i]) << "   " << sci(_sa[i]) << endl;
	os << endl;
}


// print

void ImageGeometry::print(ostream& os) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(6); fix.fixed().width(12);
	Form sci(5); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Type          :  " << type_str() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (size_t i = 0; i < _size; i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sa[i]))
			<< "   " << sci(_a[i]) << "   " << sci(_sa[i]) << endl;

	os << endl;
}


// get parameter name

string ImageGeometry::uname(size_t i) const
{
	assert(i < _size);

	ostringstream oss;
	oss << "a" << (i + 1);

	return oss.str();
}


// set upari

void ImageGeometry::set_upari()
{
	create(_size);

	for (size_t i = 0; i < _size; i++)
		set(i, &_a[i], &_sa[i]);
}


// check on convergence

int ImageGeometry::check_convergence(const Vector<double>& dx) const
{
	if (lspars.chk_geo())
		{
		int k = _index;
		for (size_t i = 0; i < _size; i++, k++)
			{
			if (fabs(dx(k)) > lspars.conv_geo())
			return i;
			}
		}

	return -1;
}


// non-member functions ...

} // namespace DEP
