/* $Id: imggeo.h,v 1.5.4.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_IMGGEO_INCLUDED
#define DEP_IMGGEO_INCLUDED

#include "adjobj.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class ObjectPoint;

/** Image geometry class.
 *  This class describes the imaging/mapping process between
 *  the 2/3-D image and the 3-D object coordinates.
 *  This class only contains the geometry parameters and thier std. devs.
 *  No assumption is made on the imaging camera/sensor
 *  and the mathematical model, e.g. the central projection or
 *  the Direct Linear Transformation (DLT).
 *
 *  @author Dirk Stallmann
 */
class ImageGeometry : public AdjObject
{
	friend double intersection(const ImageGeometry& geom1,
		const DPoint3& pho_c1, const ImageGeometry& geom2,
		const DPoint3& pho_c2, DPoint3& obj_c);

public:
	/// allocate memory for parameters and std. dev.
	void allocate(size_t sz);

	/// free memory for parameters and std. dev.
	void free();

	/// new size for parameters and std. dev.
	void newsize(size_t sz);

	/** constructor with size
	 *  @param sz number of parameters
	 */
	ImageGeometry(size_t sz);

	/// virtual destructor
	virtual ~ImageGeometry();

	/**@name read/write/print image geometry parameters */
	//@{
		/// read
		virtual int read(istream& is);

		/// write
		virtual void write(ostream& os) const;

		/// print
		virtual void print(ostream& os) const;
	//@}

	/**@name overwritten base class AdjObject members */
	//@{
		/// get parameter name
		virtual string uname(size_t i) const;
	
		/// set unknown parameter information
		virtual void set_upari();

		/// check on convergence 
		virtual int check_convergence(const Vector<double>& dx) const;
	//@}

	/// set and get exposure time
	//@{
		/// set exposure time
		void set_time(double t) { _time = t; }

		/// get exposure time
		double get_time() const { return _time; }
	//@}

	/// set geometry parameters
	//@{
		/// get number of parameters
		size_t size() const { return _size; }

		/// set parameters
		void set_params(double* a) {
			for (size_t i = 0; i < _size; i++)
				_a[i] = a[i];
		}

		/// set standard deviations
		void set_sdevs(double* s) {
			for (size_t i = 0; i < _size; i++)
				_sa[i] = s[i];
		}
	//@}

	/**@name pure virtual functions */
	//@{
		/// object-to-photo coords. transformation
		virtual bool object_to_photo(const DPoint3& obj_c,
			DPoint3& pho_c) const = 0;

		/// spatial resection
		//virtual bool resection(const vector<DPoint3>& pho_c1,
		//	const vector<DPoint3>& obj_c) const = 0;

		/// spatial intersection
		virtual bool intersection(const ImageGeometry& g2,
			const DPoint3& pho_c1, const DPoint3& pho_c2,
			DPoint3& obj_c, double& e) const = 0;

		/** get partial derivatives of image geometry parameters (eo),
		 *  object coords. (oc), position (pc) and attitude correction (ac)
		 *  parameters. The position and attitude correction parameters
		 *  are only defined/useful for parametrisation of the rotation
		 *  using three parameters. For other parametrisations these
		 *  arguments are unused.
		 *  Photo coords. are required to get focal length.
		 *  Derivatives are build from the adjusted (updated) parameters.
		 */
		virtual bool get_derivs(const DPoint3& pho_c, const DPoint3& obj_c,
			DerivSub& deriv_oc, DerivSub& deriv_eo,
			DerivSub& deriv_pc, DerivSub& deriv_ac, DerivSub& deriv_obs) = 0;
	//@}
	
	/** calculate approximate image geometry.
	 *  Per default do nothing and give an print a comment.
	 */
	virtual bool calc_approx(const vector<ObjectPoint*>& obpts, ostream* os) {
		cout << "ATTENTION: Calculate approximate image geometry"
			" not implemented.\n";
		return true;
	}

	/// calculate EO, position and attitude correction included ???
	virtual void calc_orientation() {}

	/// check link to position correction
	virtual bool has_pos_cor() const { return false; }

	/// check link to attitude correction
	virtual bool has_att_cor() const { return false; }

protected:
	/// exposure/reference time
	double	_time;

	/// number of geometry parameters
	size_t	_size;

	/// geometry parameter values
	double*	_a;

	/// std. dev. of geometry parameters
	double*	_sa;
};


// non-member functions ...

} // namespace DEP

#endif
