/* $Id: lobseqn.h,v 1.4.2.1.2.1 2005/12/12 12:27:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_LOBSEQN_INCLUDED
#define DEP_LOBSEQN_INCLUDED

#include "depdef.h"

namespace DEP
{
class DerivSub;

/** Linearised observation equations class.
 *  The linearised condition equation the functional model in the
 *  general least-squares adjustment.
 *  The linear condition equations has the form
 *
 *    $v = Ax - l$
 *
 *  where
 *    v       residual vector dim(v)=(n)
 *    A       design matrix, dim(A)=(n,u)
 *    x       unknown parameter vector, dim(x)=(u)
 *    l       constant vector, dim(u)=(r)
 *
 * Togehter with the stochatic model
 *
 *    Qll = sigma0**(-2) Vll
 *
 *  where
 *    Qll     cofactor matrix
 *    sigma0  standard deviation of unit weight
 *    Vll     covariance matrix, dim(Vll)=(n,n)
 *
 *  The linearised observation equations consist a vector of
 *  sub-matrices, the vector of shorten observation (l)
 *  and the weight vector (p).
 *  The sub-matrices holding the partial derivatives.
 *  Notice, the number of rows must be equal for all sub-matrices
 *  and vectors! 
 *  @author Dirk Stallmann
 */
class LinObsEqn
{
public:
	/// constructor
	LinObsEqn(int nr);

	/// get number of sub-matrices
	int num() const { return _a.size(); }

	/// set weights
	void set_p(int i, double p) { _p(i) = p; }

	/// set shorten observation
	void set_l(int i, double l) { _l(i) = l; }

	/// add sub-matrix
	void add(DerivSub* sub) { _a.push_back(sub); }

	/// add vector of design sub-matrix A
	void add(vector<DerivSub>& A) {
		vector<DerivSub>::iterator pa;
	    for (pa = A.begin(); pa != A.end(); ++pa)
			_a.push_back(&*pa);
	}

	/// print equations
	void print(ostream& os);

	/** add contributions of the lin. obs. eqns. to the normal equations.
	 *  By convention the normal matrix is handled in upper triangular form.
	 */
	bool add_contr(Matrix<double>& atpa, Vector<double>& atpl,
		double& ltpl) const;

private:
	/// pointer to sub-matrices of design matrix A
	vector<DerivSub*> _a;

	/// shorten observations
	Vector<double>	_l;

	/// weights
	Vector<double>	_p;
};

} // namespace DEP

#endif
