/* $Id: lspars.h,v 1.2.2.1.2.2 2007/03/06 08:51:07 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_LSPARS_INCLUDED
#define DEP_LSPARS_INCLUDED

#include "depdef.h"
#include "uangle.h"
#include "ulength.h"
#include "adjint.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{

/**@name default definition of standard deviation limits */
//@{
	const double SMIN		= 1e-30;
	const double SMAX		= 1e+30;
	const double SMIN_UNK	= 1e-31;
	const double SMAX_UNK	= 1e+31;
	const double UNKSUP_WT	= 1e+10;
	const double CONSTR_WT	= 1e+10;
//@}

/** Least-squares parameters and options class.
 *  @author Dirk Stallmann
 */
class LSParams
{
	// allows direct access for derived adjustment objects
	friend class AdjObject;

public:
	/// type of adjustment interface
	enum e_adj_interface
	{
		/// using native functions
		NATIVE,

		/// using LAPACK-functions
		LAPACK
	};
	
	/// type of partial derivatives of APs
	enum e_ap_derivs
	{
		/// image coords.
		IMAGE_COORDS,
		
		/// image coords. derived from collinear equation
		COLLINEAR_EQUATION,

		/// image coords. plus AP correction
		IMAGE_COORDS_PLUS_APS
	};

	/// constructor
	LSParams();

	/// virtual destructor
	virtual ~LSParams() {}

	/// read least-squares parameters and options
	int read(istream& is);

	/// print least-squares parameters and options
	void print(ostream& os) const;

	/**@name get member data */
	//@{
		/// get std. dev. of unit weight
		double sigma0() const { return _sigma0; }

		/// get maximum number of iterations
		int max_iter() const { return _max_iter; }

		/// get maximum number of iterations
		double atpv_lim() const { return _atpv_lim; }

		/// get output limit of correlation coeffs.
		double ccoef_lim() const { return _ccoef_lim; }
	
		/// get residual limit
		double res_lim() const { return _res_lim; }
	//@}

	/**@name get convergence criteria.
	 *  Unknowns used as iteration criterion:
	 */
	//@{
		/// - check object point coords.
		bool chk_obj() const { return _chk_obj; }

		/// - check object point coords.
		bool chk_pcc() const { return _chk_pcc; }

		/// - check rotation parameters
		bool chk_rot() const { return _chk_rot; }

		/// - check image geometry parameters
		bool chk_geo() const { return _chk_geo; }
	//@}
		
	/**@name get convergence limits.
	 *  Convergence limit of ...
	 */
	//@{
		/// - object point coords.
		double conv_obj() const { return _conv_obj; }

		/// - proj. centre coords.
		double conv_pcc() const { return _conv_pcc; }

		/// - rotation angles
		double conv_rot() const { return _conv_rot; }

		/// - image geometry parameters
		double conv_geo() const { return _conv_geo; }
	//@}

	/**@name get and test standard deviation limits */
	//@{
		/// get limit for eliminated parameter
		double smin() const { return _smin; }

		/// get limit for unknown parameter
		double smax() const { return _smax; }
		
		/// get standard deviation for eliminated parameter
		double sminu() const { return _sminu; }

		/// get standard deviation for unknown parameter
		double smaxu() const { return _smaxu; }

		/// get weight for suppression of unknowns
		double unksup_wt() const { return _unksup_wt; }

		/// get weight for constraints
		double constr_wt() const { return _constr_wt; }

		/// check on real std. dev.
		bool is_rsdev(double s) const { return (s >= _smin && s <= _smax); }

		/// check on observed parameter
		bool is_obs(double s) const { return (s >= _smin && s <= _smax); }

		/// check on eliminated parameter
		bool is_eli(double s) const { return (s < _smin); }

		/// check on unknown parameter
		bool is_unk(double s) const { return (s > _smax); }

		/// convert standard deviation (e.g. radian into degree)
		double conv_sdev(double s, double sc) const
		{
			return (is_rsdev(s) ? s * sc : s);
		}
	//@}
	
	/// A'Pv-check
	bool chk_atpv(double v) const { return (v > _atpv_lim); }

	/// check on significance
	bool chk_sign(double t) const { return (t > _t_quantil); }

	/// get conversion factor for object space coords.
	double scale_obj() const { return _unit_objc.scale(); }
	
	/// get conversion factor for rotation angles 
	double rho() const { return _unit_angle.rho(); }

	/// get symbol for unit of rotation angles 
	const char* u_angle_cstr() const { return _unit_angle.symbol(); }

	/// check limit for printing residuals
	bool is_small(double v) const
	{
		return ((fabs(v) < _res_lim) ? (false) : (true));
	}

	/// get type of adjustment interface
	e_adj_interface adj_interface() const { return _adj_interface; }
	
	/// get method to build derivatives of APs
	e_ap_derivs ap_derivs() const { return _ap_derivs; }
	
protected:
	/// std. dev. of unit weight
	double	_sigma0;

	/// maximum number of iterations
	int		_max_iter;

	/**@name convergence criteria.
	 *  Unknowns used as iteration criterion:
	 */
	//@{
		/// - check object point coords.
		bool	_chk_obj;

		/// - check object point coords.
		bool	_chk_pcc;

		/// - check rotation parameters
		bool	_chk_rot;

		/// - check image geometry parameters
		bool	_chk_geo;
	//@}
		
	/**@name convergence limits. */
	//@{
		/// convergence limit of object point coords.
		double	_conv_obj;

		/// convergence limit of proj. centre coords.
		double	_conv_pcc;

		/// convergence limit of rotation angles
		double	_conv_rot;

		/// convergence limit of image geometry parameters
		double	_conv_geo;
	//@}

	/**@name standard deviation limits. */
	//@{
		/// limit for eliminated parameter (fix unknown)
		double	_smin;

		/// limit for unknown parameter (free unknown)
		double	_smax;

		/// value to indicate that parameter is fixed
		double	_sminu;

		/// value to indicate that parameter is free
		double	_smaxu;
	//@}

	/// weight for suppression of unknowns
	double	_unksup_wt;	

	/// weight for the constraints
	double	_constr_wt;

	/// output limit of correlation coeffs.
	double	_ccoef_lim;	

	/// increment for num. derivs. of point coords.
	double	_incr_crd;

	/// increment for num. derivs. of rotation angles
	double	_incr_rot;

	/// quantil of t-distribution (significance test)
	double	_t_quantil;

	/// limit of A'Pv-check
	double	_atpv_lim;

	/// limit for printing residuals
	double	_res_lim;

	/**@name output parameters and units */
	//@{
		/// unit of object space coords.
		UnitLength	_unit_objc;

		/// unit of image space coords.
		UnitLength	_unit_imgc;

		/// unit of rotation angles
		UnitAngle	_unit_angle;
	//@}

	/// adjustment interface
	e_adj_interface	_adj_interface;
	
	/// method to build derivatives of APs
	e_ap_derivs		_ap_derivs;
};

} // namespace DEP

#endif
