/* $Id: objpnt.cpp,v 1.2.2.1.2.3 2007/04/29 19:40:26 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "deperr.h"
#include "objpnt.h"

namespace DEP
{

// constructor

ObjectPoint::ObjectPoint() :
	AdjObject(),
	crds(),					// initial coordinates (0., 0., 0.)
	sdev(lspars.smaxu()),	// initial std. devs., all coordinates are unknown 
	_type(inter),			// intersection point
	_approx_flag(false),	// no approximations exist
	_is_chx(false),
	_is_chy(false),
	_is_chz(false),			// no check point
	_num_rays(0)			// no rays
{
	// empty
}


// read object point coords. and std. dev.

int ObjectPoint::read(istream &is)
{
	is >> _id;
	if (!is) return ERROR;

	is >> crds >> sdev;
	if (!is) return ERROR;

	// test std. devs.
	//if (is_invalid()) return ERROR; // test std. devs.
	if (!is_valid(sdev.x) || !is_valid(sdev.y) || !is_valid(sdev.z))
		return ERROR;

	// set type from std. dev.
	if (is_unk(sdev.x) && is_unk(sdev.y) && is_unk(sdev.z))
		_type = inter;
	else
		_type = ctrl;

	_approx_flag =  true; // use read coords. as approximations
	_is_chx = _is_chy = _is_chz = false;

	return OK;
}


// write id, coords. and std.dev.

void ObjectPoint::write(ostream& os) const
{
	//double sc = AdjObject::lspars.scale_obj();
	double sc = 1.;

	Form str; str.width(15);
	Form fix9(5); fix9.fixed().width(9);
	Form fix15(5); fix15.fixed().width(15);
	Form sci1(1); sci1.scientific().width(9);

	os << str(_id.c_str())
		<< " " << fix15(crds.x) << " " << fix15(crds.y) << " " << fix15(crds.z)
		<< " ";
	if (is_rsdev(sdev.x))
		os << fix9(sdev.x * sc);
	else
		os << sci1(sdev.x);
	os << " ";
	if (is_rsdev(sdev.y))
		os << fix9(sdev.y * sc);
	else
		os << sci1(sdev.y);
	os << " ";
	if (is_rsdev(sdev.z))
		os << fix9(sdev.z * sc);
	else
		os << sci1(sdev.z);
	os << endl;

}


// print ID, coords., std.dev. and status

void ObjectPoint::print(ostream& os) const
{
	double sc = AdjObject::lspars.scale_obj();

	Form str8; str8.width(8);
	Form fix9(3); fix9.fixed().width(9);
	Form fix11(3); fix11.fixed().width(11);
	Form sci9(1); sci9.scientific().width(9);

	os << str8(_id.c_str()) << " "
		<< fix11(crds.x) << " " << fix11(crds.y) << " " << fix11(crds.z) << " ";

	if (is_rsdev(sdev.x))
		os << fix9(sdev.x * sc);
	else
		os << sci9(sdev.x);
	os << " ";

	if (is_rsdev(sdev.y))
		os << fix9(sdev.y * sc);
	else
		os << sci9(sdev.y);
	os << " ";

	if (is_rsdev(sdev.z))
		os << fix9(sdev.z * sc);
	else
		os << sci9(sdev.z);

	os << "   " << type_cstr() << endl;
}


// get type name

string ObjectPoint::type_str() const
{
	ostringstream oss;
	oss << "object point";
	return oss.str();
}


// get key name

string ObjectPoint::key_str() const
{
	ostringstream oss;
	oss << "obj-point";
	return oss.str();
}


// get parameter name

string ObjectPoint::uname(size_t i) const
{
	ostringstream oss;

	switch (i)
		{
		case 0: oss << "objc-X"; break;
		case 1: oss << "objc-Y"; break;
		case 2: oss << "objc-Z"; break;
		default: 
			assert(false);
			break;
		}

	return oss.str();
}


// set upari

void ObjectPoint::set_upari()
{
	create(3);

	set(0, &crds.x, &sdev.x);
	set(1, &crds.y, &sdev.y);
	set(2, &crds.z, &sdev.z);
}


// print changes of unknown parameters without header

void ObjectPoint::print_changes(ostream& os, const Vector<double>& dx) const
{
	os << "Point ID : " << _id << endl;
	os << "Index    : " << _index << endl;
	os << endl;

	os << "  Name       State"
		"       Approx.         Change       Adjusted\n\n";

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix3(3); fix3.fixed().width(12);

	int k = _index;
	for (size_t i = 0; i < _num; i++, k++)
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix3(par(i)) << "   " << fix3(dx(k)) << "   "
			<< fix3(par(i) + dx(k)) << endl;
	os << endl;
}


// check on convergence

int ObjectPoint::check_convergence(const Vector<double>& dx) const
{
	int k = _index;

	if (lspars.chk_obj())
		{
		if (fabs(dx(k+0)) > lspars.conv_obj()) return 0;
		if (fabs(dx(k+1)) > lspars.conv_obj()) return 1;
		if (fabs(dx(k+2)) > lspars.conv_obj()) return 2;
		}

	return -1;
}



// set point type from std. dev.

void ObjectPoint::set_type()
{
	_is_chx = _is_chy = _is_chz = false; // not a check point

	// set type from std. dev.
	//if (is_unk(sdev.x) && is_unk(sdev.y) && is_unk(sdev.z))
	if (sdev.x < lspars.smax() || sdev.y < lspars.smax() || sdev.z < lspars.smax())
		_type = ctrl;
	else
		_type = inter;
}


// get c-style string of point type

const char* ObjectPoint::type_cstr() const
{
	static char str[3];

	switch (_type)
		{
		case invalid:	strcpy(str, "nu"); break;
		case ctrl:		strcpy(str, "co"); break;
		case inter:		strcpy(str, "in"); break;
		case check:		strcpy(str, "ch"); break;
		default:
			assert(false);
			break;
		}

	return str;
}


// get c-style string of check point flags

const char* ObjectPoint::ch_cstr() const
{
	static char str[4];

	strcpy(str, "");
	if (_is_chx) strcat(str, "x"); else strcat(str, " ");
	if (_is_chy) strcat(str, "y"); else strcat(str, " ");
	if (_is_chz) strcat(str, "z"); else strcat(str, " ");

	return str;
}


// set type from std. dev.
/*
void ObjectPoint::set_type(const LSParams& lspars)
{
	// set type: when one CS indicate this coord. is
	// eliminated (fixed) or observed this point is a control point
	// otherwise assume the coords. are approximations
	// of a new (intersection) point.
	//ctrl_supp cs;
	//cs.is_x = upari.is_unk(sdev.x) ? 0 : 1;
	//cs.is_y = upari.is_unk(sdev.y) ? 0 : 1;
	//cs.is_z = upari.is_unk(sdev.z) ? 0 : 1;
	//cs.is_x = is_csx(upari);
	//cs.is_y = is_csy(upari);
	//cs.is_z = is_csz(upari);

	//if (cs.is_x || cs.is_y || cs.is_z)
	if (is_csx(lspars) || is_csy(lspars) || is_csz(lspars))
		_type = ctrl;
	else
		_type = approx;

	// set CS
	//if (upari.is_unk(sdev.x)) _cs.is_x = 0; else _cs.is_x = 1;
	//if (upari.is_unk(sdev.y)) _cs.is_y = 0; else _cs.is_y = 1;
	//if (upari.is_unk(sdev.z)) _cs.is_z = 0; else _cs.is_z = 1;
}
*/


// set std. dev. from type
/*
void ObjectPoint::set_sdev(const Upari& upari)
{
	if (_type == ctrl)
		{
		sdev.x = (_cs.is_x == 1) ? upari.sminu() : upari.smaxu();
		sdev.y = (_cs.is_y == 1) ? upari.sminu() : upari.smaxu();
		sdev.z = (_cs.is_z == 1) ? upari.sminu() : upari.smaxu();
		}
	else if (_type == approx || _type == chk || _type == inter) 
		sdev.x = sdev.y = sdev.z = upari.smaxu();
}
*/


// set std. dev. from CS

//void ObjectPoint::set_sdev(const ObjectPointCS& cs, const LSParams& lspars)
int ObjectPoint::set_sdev(const ObjectPointCS& cs)
{
	//cout << "ObjectPointCS::set_type: " << cs.id << " " << cs.type;
	//if (cs.cs.is_x) cout << " 1"; else cout << " 0";
	//if (cs.cs.is_y) cout << " 1"; else cout << " 0";
	//if (cs.cs.is_z) cout << " 1"; else cout << " 0";
	//cout << endl;

	switch (cs.type)
		{
		// control point: CS overwrite given std. devs.!
		case ctrl:
			sdev.x = cs.cs_x ? lspars.sminu() : lspars.smaxu();
			sdev.y = cs.cs_y ? lspars.sminu() : lspars.smaxu();
			sdev.z = cs.cs_z ? lspars.sminu() : lspars.smaxu();
			_is_chx = _is_chy = _is_chz = false;
			break;

		// check point (-> special intersection point)
		case check:
			sdev.x = sdev.y = sdev.z = lspars.smaxu();
			_is_chx = cs.cs_x;
			_is_chy = cs.cs_y;
			_is_chz = cs.cs_z;
			break;

		// intersection points
		case inter:
			sdev.x = sdev.y = sdev.z = lspars.smaxu();
			_is_chx = _is_chy = _is_chz = false;
			break;

		case invalid:
		default:
			cout << "ERROR: Invalid object point type: " << cs.type << endl;
			return ERROR;
		}

	_type = cs.type; // set type

	return OK;
}


// sum of absolute changes

void ObjectPoint::sum_changes(const Vector<double>& dx,
double& sum, int& num) const
{
	int k = _index;

	if (state(0) != eli) sum += fabs(dx(k++)), num++;
	if (state(1) != eli) sum += fabs(dx(k++)), num++;
	if (state(2) != eli) sum += fabs(dx(k  )), num++;
}


// constructor

ObjectPointCS::ObjectPointCS() :
	id(),
	type(ObjectPoint::invalid),
	cs_x(false),
	cs_y(false),
	cs_z(false)
{
	// empty
}


// read control support

int ObjectPointCS::read_ctrl_supp(istream &is)
{
	char c = 0;
	while (1)
		{
		is >> c;
		if (!is) return OK; // reach EOF

		// eat comments
		if (c == '#')
			while (is.get(c) && c != '\n');
		else
			{
			is.putback(c);
			break;
			}
		}

	// point ID
	is >> id;
	if (!is)
		{
		cout << "ERROR: Cannot read point ID.\n";
		return ERROR;
		}

	if (id.compare("stop-dep") == 0 || id.compare("-1") == 0)
		return OK;

	// point type
	string str;
	is >> str;
	if (!is)
		{
		cout << "ERROR: Cannot read object point type.\n";
		return ERROR;
		}

	if (str.compare("in") == 0)
		type = ObjectPoint::inter;
	else if (str.compare("co") == 0)
		type = ObjectPoint::ctrl;
	else if (str.compare("ch") == 0)
		type = ObjectPoint::check;
	else
		{
		cout << "ERROR: Invalid object point type: " << str << endl;
		return ERROR;
		}

	// for control or check points set control support
	cs_x = cs_y = cs_z  = false;
	if (type == ObjectPoint::ctrl || type == ObjectPoint::check)
		{
		//char* c = str;
		is >> str;
		if (!is) return ERROR;
		//cout << str << endl;

		if (str.find('x') != string::npos) cs_x = true;
		if (str.find('y') != string::npos) cs_y = true;
		if (str.find('z') != string::npos) cs_z = true;
		}

	return OK;
}


// non-member functions ...

} // namespace DEP
