/* $Id: objpnt.h,v 1.2.2.1.2.2 2007/03/06 08:52:45 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_OBJPNT_INCLUDED
#define DEP_OBJPNT_INCLUDED

#include "adjobj.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class ObjectPointCS;

/** Object point class.
 *  The object point class holds the information of a point in object space:
 *  the point identifier (ID), the coordinates and their standard deviations.
 *  The units of the object point coordinates and their standard deviations 
 *  are assumed in meter.
 *  Some additional information are required: the point type, the
 *  control support (CS) and the index to the normal equation system.
 *  The CS is only used by check points.
 *  
 *  @author Dirk Stallmann
 */
class ObjectPoint : public AdjObject
{
public:
	/// object point type
	enum op_type
	{
		/// invalid std. devs.
		invalid,
		
		/// control point
		ctrl,

		/// intersection/tie point (with or without approximations)
		inter,

		/// check point
		check
	};

	/** constructor.
	 *  Default: intersection point without approximations.
	 */
	ObjectPoint();

	/// read
	int read(istream& is);

	/// write
	void write(ostream& os) const;

	/// print
	void print(ostream& os) const;

	/**@name overridden base class members */
	//@{
		/// get type name
		virtual string type_str() const;

		/// get key name
		virtual string key_str() const;

		/// get parameter name
		virtual string uname(size_t i) const;

		/// print changes of unknowns
		virtual void print_changes(ostream& os, const Vector<double>& dx) const;

		/// check on convergence
		virtual int check_convergence(const Vector<double>& dx) const;
	//@}

	/**@name test control support */
	//@{
		/// test control support in X
		bool is_csx() const { return (!is_unk(sdev.x)); }

		/// test control support in Y
		bool is_csy() const { return (!is_unk(sdev.y)); }

		/// test control support in Z
		bool is_csz() const { return (!is_unk(sdev.z)); }

		// test control point
		//bool is_ctrl() const { return (is_csx() || is_csy() || is_csz()); }
	
		// test check point
		//bool is_check() const { return (_is_chx || _is_chy || _is_chz); }
	//@}

	/**@name set, get and test point type.
	 *  Types: control, check or intersection point.
	 *  We still assume the type is consistent with
	 *  the check points flags!
	 */
	//@{
		/// test control point
		bool is_ctrl() const { return (_type == ctrl); }

		/// test check point
		bool is_check() const { return (_type == check); }

		/// set point type from std.dev.
		void set_type();
	
		/// get point type
		op_type get_type() const { return _type; }

		/** get c-style string of point type
		 *  nu=invalid (null), co=control, ch=check, in=intersection.
		 */
		const char* type_cstr() const;
	//@}

	/**@name approximation flag. */
	//@{
		/// set approximation flag
		void approx(bool flag) { _approx_flag = flag; }

		/// check approximation flag
		bool has_approx() const { return (_approx_flag == true); }
	//@}

	/**@name get and test check point flags. */
	//@{
		/// test check point flags
		bool is_chx() const { return _is_chx; }
		bool is_chy() const { return _is_chy; }
		bool is_chz() const { return _is_chz; }

		/// get c-style string of check point flags
		const char* ch_cstr() const;
	//@}

	/// set number of rays
	void num_rays(int num) { _num_rays = num; }

	/// get number of rays
	int num_rays() const { return _num_rays; }

	// full control
	//bool is_full() { return (_cs.is_x && _cs.is_y && _cs.is_z); }
	// planimetric control
	//bool is_plan() { return (_cs.is_x && _cs.is_y); }
	// vertical control
	//bool is_vert() { return _cs.is_z; }

	/// set std. dev. from CS
	int set_sdev(const ObjectPointCS& cs);

	/// set upari
	void set_upari();

	/// sum of absolute changes
	void sum_changes(const Vector<double>& dx,
		double& sum, int& num) const;

public:
	/// 3-D object point coordinates
	DPoint3	 crds;

	/// standard deviations thereof
	DPoint3	 sdev;

private:
	/// point type
	op_type	_type;
	
	/// approximation flag
	bool	_approx_flag;
	
	/**@name check point flags */
	//@{
		/// check point in X
		bool	_is_chx;

		/// check point in Y
		bool	_is_chy;

		/// check point in Z
		bool	_is_chz;
	//@}

	/// number of rays
	int		_num_rays;
};


/** Object point control support class.
 *  
 *  @author Dirk Stallmann
 */
class ObjectPointCS
{
public:
	/// constructor
	ObjectPointCS();

	/// read CS
	int read_ctrl_supp(istream& is);

	/// point ID
	string id;

	/// point type
	ObjectPoint::op_type	type;

	/// control support in X
	bool	cs_x;

	/// control support in Y
	bool	cs_y;

	/// control support in Z
	bool	cs_z;
};


// non-member functions ...

} // namespace DEP

#endif
