/* $Id: pmflin.cpp,v 1.3.2.1.2.1 2005/12/12 12:28:01 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "pmflin.h"
#include "cholesky.h"

namespace DEP
{

// get type name

string PMFLinear::type_str() const
{
	ostringstream oss;
	oss << "PMF linear";
	return oss.str();
}


// get key name

string PMFLinear::key_str() const
{
	ostringstream oss;
	oss << "pmf-linear";
	return oss.str();
}


// get parameter name

string PMFLinear::uname(size_t i) const
{
	ostringstream oss;
	assert(i < _size);

	switch (i)
		{
		case 0: oss << "a0"; break;
		case 1: oss << "a1"; break;
		case 2: oss << "a2"; break;
		case 3: oss << "a3"; break;
		case 4: oss << "b0"; break;
		case 5: oss << "b1"; break;
		case 6: oss << "b2"; break;
		case 7: oss << "b3"; break;
		}

	return oss.str();
}


// object-to-photo coords. transformation

bool PMFLinear::object_to_photo(const DPoint3& obj_c,
DPoint3& pho_c) const
{
	DPoint2 img_c;

	affine_transform(_a, obj_c, img_c);

	pho_c.x = img_c.x;
	pho_c.y = img_c.y;

	return true;
}


// spatial intersection

bool PMFLinear::intersection(const ImageGeometry& g2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	bool rtn = false;

	try {
		const PMFLinear& pmf2 = dynamic_cast<const PMFLinear&>(g2);
		rtn = intersection(pmf2, pho_c1, pho_c2, obj_c, e);
	}
	catch (bad_cast) {
		cout << "PMFLinear::intersection: Incompatible image geometries!\n";
		rtn = false;
	}
	
	return rtn;
}


// get partial derivatives

bool PMFLinear::get_derivs(const DPoint3& pho_c, const DPoint3& obj_c,
DerivSub& deriv_oc, DerivSub& deriv_eo,
DerivSub& deriv_pc, DerivSub& deriv_ac, DerivSub& deriv_obs)
{
	// adjust sub-vectors
	deriv_oc.newsize(2,3);
	deriv_eo.newsize(2,_size);

	// derivatives
	deriv_eo(0, 0) = 1.;
	deriv_eo(0, 1) = obj_c.x;
	deriv_eo(0, 2) = obj_c.y;
	deriv_eo(0, 3) = obj_c.z;
	deriv_eo(0, 4) = 0.;
	deriv_eo(0, 5) = 0.;
	deriv_eo(0, 6) = 0.;
	deriv_eo(0, 7) = 0.;

	deriv_eo(1, 0) = 0.;
	deriv_eo(1, 1) = 0.;
	deriv_eo(1, 2) = 0.;
	deriv_eo(1, 3) = 0.;
	deriv_eo(1, 4) = 1.;
	deriv_eo(1, 5) = obj_c.x;
	deriv_eo(1, 6) = obj_c.y;
	deriv_eo(1, 7) = obj_c.z;

	deriv_oc(0,0) = _a[1];
	deriv_oc(0,1) = _a[2];
	deriv_oc(0,2) = _a[3];

	deriv_oc(1,0) = _a[5];
	deriv_oc(1,1) = _a[6];
	deriv_oc(1,2) = _a[7];

	// derivatives of observations
	deriv_obs.newsize(2,2);
	deriv_obs.identity();
	
	return true;
}


// spatial intersection

bool PMFLinear::intersection(const PMFLinear& pmf2,
const DPoint3& pho_c1, const DPoint3& pho_c2,
DPoint3& obj_c, double& e) const
{
	DPoint2 img_c1(pho_c1.x, pho_c1.y);
	DPoint2 img_c2(pho_c2.x, pho_c2.y);

	return affine_intersection(this->_a, img_c1, pmf2._a, img_c2, obj_c);
}


// non-member functions ...

// projection using affine transformation

void affine_transform(const double* a, const DPoint3& obj_c, DPoint2& img_c)
{
	assert(a != 0);

	// image coords.
	img_c.x = a[0] + a[1] * obj_c.x + a[2] * obj_c.y + a[3] * obj_c.z;
	img_c.y = a[4] + a[5] * obj_c.x + a[6] * obj_c.y + a[7] * obj_c.z;
}


// intersection using affine transformation

bool affine_intersection(const double* a1, const DPoint2& pho_c1, 
const double* a2, const DPoint2& pho_c2, DPoint3& obj_c)
{
	assert(a1 != 0);
	assert(a2 != 0);

	Matrix<double> A(4,3);
	Vector<double> l(4);

	A(0,0) = a1[1];
	A(0,1) = a1[2];
	A(0,2) = a1[3];

	A(1,0) = a1[5];
	A(1,1) = a1[6];
	A(1,2) = a1[7];

	A(2,0) = a2[1];
	A(2,1) = a2[2];
	A(2,2) = a2[3];

	A(3,0) = a2[5];
	A(3,1) = a2[6];
	A(3,2) = a2[7];

	l(0) = pho_c1.x - a1[0];
	l(1) = pho_c1.y - a1[4];

	l(2) = pho_c2.x - a2[0];
	l(3) = pho_c2.y - a2[4];

	Matrix<double> At = transpose(A);

	Matrix<double> N = At * A;
	Vector<double> n = At * l;

	Vector<double> dec;
	if (chol_dec(N, dec) != 0)
		{
		cout << "ERROR: Decomposition of normal matrix failed.\n";
		return false;
		}

	Vector<double> x;
	chol_sol(N, dec, n, x);

	obj_c.x = x(0);
	obj_c.y = x(1);
	obj_c.z = x(2);

	return true;
}

} // namespace DEP
